/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_client.h"
#include "nest_config.h"
#include "nest_protocol_manager.h"

enum ClientState {
    CLIENT_PASSIVE,
    CLIENT_ACTIVE
};

// here is how a nest client object is used
// when the client connects, a new object is formed and connection is assigned
// when the client makes a request, a request is assigned to the object
// if the request is a transfer, a transfer is assigned to the object
// then the transfer is started
// when finished with transfer, transfer is ended
// when finished with a simple request, request is ended

nestClient::nestClient( const char *initial_cwd,
						nestProtocol* connection,
						sockaddr_in *s_address )
{
    static int id           = 0;
    
    this->connection        = connection;
    this->storage           = NULL;
    this->otherStorage      = NULL;
    this->third_party       = NULL;
    this->req               = NULL;
    this->transfer          = NULL;
    this->last_transfer     = NULL;
    this->last_req          = NULL;
    this->last_proxy_req    = NULL;
    this->proxy_req         = NULL;     // a client request to a proxy
                                        // may create two requests
    this->proxy_address     = NULL;     // if a client registers as a proxy,
                                        // remember the registering address

    this->dirty_writes      = NULL;     // if the server is a proxy

    this->id                = ++id;
    this->state             = CLIENT_PASSIVE;
    this->buffer[0]         = '\0';
    this->cwd               = (char *) NestMalloc (NEST_MAXPATHLEN);
    this->receivedRequests  = 0;



    this->is_proxy          = false;
    this->waiting_to_close  = false;
    this->allocation        = 0;
    this->last_allocation   = 0;

    this->proxy_client      = NULL;
    this->proxy_fetch       = NULL;

    this->task_fp           = NULL;
    this->cleanup_fp        = NULL;
    this->ip_address        = NULL;
	this->host_name			= NULL;
	this->host_identity		= NULL;

    this->taskStatus        = NEST_NOT_YET_INITIALIZED;
    this->taskLength        = 0;

	this->user				= NULL;

    NestGettimespec( &this->connectTime );

    if ( s_address != NULL ) {
        this->address = s_address->sin_addr;
        this->port    = ntohs( s_address->sin_port );
        this->ip_address = NestGetIP( address );    // must free later
    } else {
        this->address.s_addr = 0;
        this->port = 0;
    }

    setCwd( initial_cwd );
    nest_debug( NEST_DEBUG_CLIENT, "new client %s (%d) cwd set to %s\n", 
        getUserIdentity(), id, cwd );
}    

nestClient::~nestClient( )
{
    nest_debug( NEST_DEBUG_CLIENT, "Deleting %s (%d) cwd set to %s\n", 
				getUserIdentity(), id, cwd );
    assert( state == CLIENT_PASSIVE );
    assert( transfer == NULL );

	if ( this->dirty_writes && this->dirty_writes->size() > 0 ) {
		nest_warn( __FILE__, __FUNCTION__, 
				   "Client %s exiting with %d unflushed dirty writes.\n",
				   getUserIdentity(), dirty_writes->size() );
		while ( dirty_writes->size() ) {
			nestRequest *req = dirty_writes->front();
			dirty_writes->pop();
			delete req;
		}
		delete dirty_writes;
	}

	if ( this->pending_reqs.size() > 0 ) {
		nest_warn( __FILE__, __FUNCTION__, 
				   "Client %s exiting with %d pending reqs.\n",
				   getUserIdentity(), pending_reqs.size() );
		while( pending_reqs.size() ) {
			nestRequest *req = pending_reqs.front();
			pending_reqs.pop();
			delete req;
		}
	}

	if ( this->cwd != NULL ) 		    NestFree( cwd );
    if ( this->req != NULL )           	delete req;
    if ( this->last_transfer != NULL ) 	delete last_transfer;
    if ( this->last_req != NULL )       delete last_req;
    if ( this->proxy_req != NULL )      delete proxy_req;
    if ( this->last_proxy_req != NULL ) delete last_proxy_req;
    if ( this->proxy_address != NULL )  NestFree( this->proxy_address );
    if ( this->ip_address != NULL )     NestFree( this->ip_address );
    if ( this->host_name != NULL )      NestFree( this->host_name );
    if ( this->host_identity != NULL )  NestFree( this->host_identity );
    if ( this->proxy_fetch != NULL )    delete proxy_fetch;
    if ( this->connection != NULL )     delete connection;
    if ( this->storage    != NULL )     delete storage;
    if ( this->third_party != NULL )    delete third_party;
    if ( this->otherStorage != NULL )   delete otherStorage;
	if ( this->user != NULL )			delete user;
}

int
nestClient::operator==( const nestClient &other )
{
    return ( id == other.id );
//    return false;     // what the hell is this shit?
}

void 
nestClient::setFlushing( bool flushing )
{
    this->flushing = flushing;
}

bool 
nestClient::isFlushing( )
{
    return flushing;
}

nestRequest *
nestClient::getDirtyWrite( )
{
    nestRequest *req = NULL;
    if ( dirty_writes != NULL && dirty_writes->size() ) {
        req = dirty_writes->front();
        dirty_writes->pop();
    }
    return req;
}

NestReplyStatus
nestClient::addDirtyWrite( nestRequest *req )
{
	if ( dirty_writes == NULL ) {
		dirty_writes = new queue < nestRequest * >;
	}

	dirty_writes->push( new nestRequest( req ) );
	nest_debug( NEST_DEBUG_PROXY, 
				"client %s now has %d dirty writes\n",
				getUserIdentity(), dirty_writes->size() );
	return NEST_SUCCESS;

}


void
nestClient::assignProxyRequest( nestRequest *req )
{
    if ( this->proxy_req != NULL ) {
        free( this->proxy_req );
    }
    this->proxy_req = req;
}

void
nestClient::assignTaskFunction( taskFunction *tfp, taskCleanupFunction *cfp )
{
    if ( tfp != NULL ) {
        assert( task_fp == NULL && cleanup_fp == NULL );
        setActive();
    } else {
        setPassive();
    }
    task_fp = tfp;
    cleanup_fp = cfp;
    taskStatus = NEST_NOT_YET_INITIALIZED;
    taskLength = 0;
}

taskFunction * 
nestClient::getTaskFunction()
{
    return task_fp;
}

taskCleanupFunction *
nestClient::getTaskCleanupFunction()
{
    return cleanup_fp;
}


void
nestClient::assignProxyClient( nestClient *proxy )
{
    if ( proxy != NULL ) {
        assert( proxy_client == NULL );
    }
    proxy_client = proxy;
}

nestClient *
nestClient::getProxyClient( )
{
    return proxy_client;
}

NestReplyStatus
nestClient::getTaskStatus( )
{
    return taskStatus;
}

void
nestClient::setTaskStatus( NestReplyStatus status )
{
    assert( taskStatus == NEST_NOT_YET_INITIALIZED );
    taskStatus = status;
}

nestProxyFetch *
nestClient::getProxyFetch()
{
    return proxy_fetch;
}

void
nestClient::assignProxyFetch ( nestProxyFetch *fetch )
{
    if ( fetch != NULL && this->proxy_fetch != NULL ) {
        nest_warn( __FILE__, __FUNCTION__,
				   "Overwriting a proxy fetch object in %s:%s\n",
				   __FUNCTION__, __FILE__ );
    }
    this->proxy_fetch = fetch;
}

nestRequest *
nestClient::getProxyRequest( )
{
    if ( proxy_req == NULL ) {
        return req;
    } else {
        return proxy_req;
    }
}

void
nestClient::setHostIp( const char *ip )
{
    if ( ip_address ) NestFree( ip_address );
    ip_address = NestStrdup( ip );
}

const char *
nestClient::getHostIp( )
{
    return ip_address;
}

// a proxy client represents a connection between a home server and a proxy
// a client can be a proxy for two reasons, either:
//  1) the server is a home server and the client is from the proxy, or
//  2) the server is a proxy and the client is to the home
void
nestClient::setProxy( bool whether )
{
    is_proxy = whether;
}

void
nestClient::setProxy( const char *proxylocation )
{
    proxy_address = NestStrdup( proxylocation );
    setProxy( true );
}

const char *
nestClient::proxyAddress( )
{
    return proxy_address;
}

bool
nestClient::isProxy( )
{
    return is_proxy;
}


NestReplyStatus 
nestClient::sendProxyIndex( DynamicLard *index )
{
    return connection->sendProxyIndex( index );
}

bool
nestClient::persistent()
{
    return connection->persistent();
}

short
nestClient::getClientPort( )
{
    return port; 
}

NestReplyStatus
nestClient::setCwd( const char *dir )
{
        // make new buffer if necessary
    nest_debug( NEST_DEBUG_CLIENT, "Setting client %d cwd: %s\n", id, dir );

    if (NestSnprintf( cwd, NEST_MAXPATHLEN, "%s", dir ) < 0) {
        free (cwd);
        cwd = NestMakeString( NEST_MAXPATHLEN, "%s", dir );
        if (!cwd) {
            return NEST_OUT_OF_RESOURCES;
        }
    }
    return NEST_SUCCESS;
}

bool
nestClient::sendAd( nestAd *ad )
{
    return connection->sendAd( ad );
}

bool 
nestClient::sendDirectoryList( NestFileInfo *list )
{
    return connection->sendDirectoryList( list );
}

bool
nestClient::sendPwd( const char *pwd)
{
    return connection->sendPwd( pwd );
}

bool
nestClient::sendVersion( const char *version )
{
    return connection->sendVersion( version );
}

bool
nestClient::sendDiskspace( nestLotSize availablebytes, nestLotSize totalbytes )
{
    return connection->sendDiskspace( availablebytes, totalbytes );
}

bool 
nestClient::sendErrorMessage( const char* filename,
							  const char* func_name,
							  int error )
{
        // in some case, the connection should close itself
        // if there is only one data and control path and the client
        // is expecting data then we must close the connection in order
        // to communicate failure - this decision must be made by the 
        // protocol
    return connection->sendErrorMessage( filename, func_name,  error );
}

bool 
nestClient::sendAck( )
{
    return connection->sendAck();
}

bool 
nestClient::sendFinalAck( )
{
        // POP workers can't call this function because their clients
        // don't have instantiated transfer objects
    return connection->sendFinalAck( req->getType() );
}
    
bool 
nestClient::sendFinalAck(  int transferType )
{
    return connection->sendFinalAck( transferType );
}

bool
nestClient::sendInitialAck( )
{
    return connection->sendInitialAck();
}

bool
nestClient::sendTransferInfo( )
{
    return connection->sendTransferInfo( this->req );
}

bool
nestClient::sendTimestamp( )
{
    return connection->sendTimestamp();
}

bool
nestClient::sendBandwidthStatistics( )
{
    return connection->sendBandwidthStatistics();
}

bool
nestClient::sendFilesize(  nestFileSize filesize )
{
    return connection->sendFilesize( filesize );
}

NestReplyStatus
nestClient::sendFileinfo( const NestFileInfo *info )
{
    return connection->sendFileinfo( info );
}

NestReplyStatus
nestClient::setFileattr( NestFileInfo *info,
						 const char *fname,
						 const void *fattr )
{
    return connection->setFileattr( info, fname, fattr ); 
}           

bool
nestClient::sendPhysName( const char *physname )
{
    return connection->sendPhysicalName( physname );
}

bool
nestClient::sendAcl(  const char *user,
                      int is_user,
                      nest_access_rights_t access)
{
    return connection->sendAcl( user, is_user, access );
}

bool
nestClient::sendAcl( const NestACL *Acl )
{
    return connection->sendAcl( Acl );
}

bool
nestClient::sendNestLot( NestLotInternal *r)
{
    return connection->sendNestLot(r);
}

bool
nestClient::sendLotHandle( NestLotHandle *h)
{
    return connection->sendLotHandle(h);
}

NestReplyStatus
nestClient::sendFilehandle( NestFileInfo *h)
{
    return connection->sendFilehandle(h);
}

int
nestClient::storeRequest( nestRequest *req )
{
    int size;
    pending_reqs.push( req );
    nest_debug( NEST_DEBUG_CLIENT, "Storing a request in client %d.\n", id );

    size = pending_reqs.size();
    /*
    if ( size > 1 ) {
        nest_warn( __FILE__, __FUNCTION__, "Multiple requests (%d) stored in client %d.\n", size, id );
    }
    */
    return size;
}

nestRequest *
nestClient::receiveRequest( void )
{

    nestRequest *req;

    if ( waiting_to_close ) {
            // the client has previously receive a close request and is
            // waiting for their transfer to finish, do not try to receive
            // any more requests at this point.  This makes life simpler
            // for the protocol receiveRequest() function which for 
            // non-persistent protocols can just return one CLOSE request
        if ( isActive() ) {
                // keep waiting for transfer to finish
            return NULL;
        } else {
                // transfer has now finished
            req = pending_reqs.front();
            pending_reqs.pop();
            nest_debug( NEST_DEBUG_CLIENT, 
						"Client %d (waiting_to_close) "
						"recving previously stored request.\n", id );
            if ( pending_reqs.size() == 0 ) {
                assert( req->getType() == CLOSE_CONNECTION_REQUEST );
            }
            goto DONE;
        }
        return NULL;
    }

        // if we make it here, then we are not in waiting_to_close state
        // look to see whether the connection has any new requests for us
        // also check to see whether this moves us into waiting_to_close
    req = connection->receiveRequest( );
    if ( req != NULL ) {
        waiting_to_close = ( req->getType() == CLOSE_CONNECTION_REQUEST ); 
    }

        // 8 possible states depending on active, req, and queue
        // whenever a real req is returned, it must also be assigned

    if ( isActive() ) {
            // whenever we're all ready active, return NULL
        if ( req != NULL ) {
                // queue the newly receive request, bec we can't deal w/ it now
            if ( req->getType( ) != CLOSE_CONNECTION_REQUEST ) {
                nest_warn( __FILE__, __FUNCTION__, 
						   "Queuing req %d (type %d) received"
						   " for busy client %s.\n", 
						   pending_reqs.size(), req->getType(), toString() );
            } else {
                waiting_to_close = true;
            }
            storeRequest( req );
        }
        return NULL;
    } 
    
        // if we make it here, we are not in wait_to_close state and we are
        // not active, so we will assign and return a req
    
    if ( pending_reqs.empty() ) {
            // here we're passive with an empty queue, return new req
        goto DONE;
    } else {
            // here we're passive with a non-empty queue, 
            // queue any newly received req and return head of req queue
        if ( req != NULL ) {
            if ( req->getType() != CLOSE_CONNECTION_REQUEST ) {
                nest_warn( __FILE__, __FUNCTION__, 
                    "Queuing request %d received for passive client %s.\n", 
                    pending_reqs.size(), toString() );
            }
            storeRequest( req );
        }
        req = (nestRequest *)pending_reqs.front();
        pending_reqs.pop();
        /*
           // this is actually the normal state of affairs for nfs
        if ( req->getType() != CLOSE_CONNECTION_REQUEST ) {
            nest_warn( __FILE__, __FUNCTION__, 
                "Deq'ing prev queued req for client %s.\n", toString());
        }
        */
        goto DONE;
    }

    DONE:
    if ( req != NULL ) {
        receivedRequests++;
        assignRequest( req );
    }
    return req;
}

bool
nestClient::hasSocket()
{
    if ( connection != NULL ) {
        return connection->hasSocket();
    } else {
        return false;
    }
}

const char *
nestClient::toString( )
{
    static char string[NEST_MESSAGESIZE];
    NestSnprintf( string,
                  NEST_MESSAGESIZE,
                  "%d, sock %d, prot %d, name %s, ip %s:%d ",
                  getId(),
                  connection->getSocket(),
                  connection->getType(),
                  getUserIdentity(),
                  getHostIp(),
                  getClientPort() );

    NestSnprintf( &(string[NestStrlen(string)]),
                  NEST_MESSAGESIZE - NestStrlen( string ),
                  "%s",
                  ( (state == CLIENT_PASSIVE) ? "(passive)" : "(active)" ) );
    return string;
}

void
nestClient::dump( )
{

    int socket = -1;
    int type = -1;

    if ( connection != NULL ) {
        socket = connection->getSocket();
        type   = connection->getType();
    }

	const char *state_str;
    if ( state == CLIENT_PASSIVE )
		state_str = "passive";
    else
		state_str = "active";

	char	req_str[128];
    if ( req == NULL )
		strcpy( req_str, "req is null" );
    else
		snprintf( req_str, sizeof( req_str ),
				  "req type is %d", req->getType() );

	const char *xfer_str = "";
    if (transfer == NULL)
		xfer_str = "(transfer is null)";

    nest_debug( NEST_DEBUG_CLIENT,
				"Dump client \"%s\" (%d), sock %d, prot %d, %s, %s %s\n", 
				getUserIdentity(), id, socket, type,
				state_str, req_str, xfer_str );
}

char *
nestClient::getBuffer(void)
{
    return buffer;
}

const char *
nestClient::getCwd(void) const
{
    return cwd;
}

int
nestClient::getId(void) const
{
    return id;
}

const char *
nestClient::getHostName( void ) const
{
	return host_name;
}

const char *
nestClient::getHostIdentity( void ) const
{
	return host_identity;
}

const char *
nestClient::getUserIdentity( void ) const
{
	if ( user ) {
		return user->getIdentity( );
	} else {
        const char *anonymous;
        if ( GetConfigValue( Config_AnonymousUser, &anonymous ) ) {
            return anonymous;
        } else {
            return NULL;
        }
	}
}

const char *
nestClient::getUserName( void ) const
{
	if ( user ) {
		return user->getName( );
	} else {
        const char *anonymous;
        if ( GetConfigValue( Config_AnonymousUser, &anonymous ) ) {
            return anonymous;
        } else {
            return NULL;
        }
	}
}

void
nestClient::setHostName( const char *name )
{
	if( this->host_name ) {
		free( (void*) this->host_name );
	}
	this->host_name = strdup( name );
}

void
nestClient::setHostIdentity( const char *identity )
{
	if( this->host_identity ) {
		free( (void*) this->host_identity );
	}
	this->host_identity = strdup( identity );
}

void
nestClient::setUser( nestUser *newuser )
{
	if ( this->user ) {
		delete user;
	}
	this->user = newuser;
}

void
nestClient::copyUser( const nestUser *newuser )
{
	if ( user ) {
		user->reset( );
		user->copyUser( newuser );
	} else {
		user = new nestUser( newuser );
	}
}

void
nestClient::assignRequest( nestRequest *req )
{
    nest_debug( NEST_DEBUG_CLIENT, "client %d: assignRequest \"%s\" (%d)\n", 
				id, req->toString(), req->getType() );
    dump();
    setActive( );

    assert( this->req == NULL );
    assert( this->transfer == NULL );
    if ( this->req      != NULL ) NestFree( this->req );
    if ( this->transfer != NULL ) NestFree( this->transfer );
    this->req = req;
}    

// this is called directly from outside the class
void
nestClient::endRequest( NestReplyStatus status )
{
    my_endRequest( status );
}

// this is called from within the class by either endRequest or endTransfer
void
nestClient::my_endRequest( NestReplyStatus status )
{

    const char *req_type = "special task";
    if ( req == NULL ) {        // we should be doing one or the other
        assert( proxy_fetch != NULL );
    } else {
        req_type = req->toString();
    }
    nest_debug( NEST_DEBUG_CLIENT,
				"client %s %d: endRequest \"%s\": %d '%s'\n", 
				getUserName(), 
				id,
				req_type,
				status,
				NestErrorString( status ) );
    dump();


        // clean up any aborted transfers (these probably only from 3pts)

    // turn off this behavior, sometimes users send files of length 0 . . . .
#if 0
	if ( req != NULL ) {
		if (req->getNestFile()!=NULL &&
			NestFilelength(req->getNestFile())== 0){
			nestStorageManager::deleteFile( getIdentity(),
											req->getNestFile() );
        }
    }
#endif
    
        // we don't delete this.
    if ( proxy_fetch != NULL )      proxy_fetch = NULL;

    if ( last_req != NULL )         delete( last_req );
    if ( last_proxy_req != NULL )   delete ( last_proxy_req );
    last_req = req;
    last_proxy_req = proxy_req;
    req = NULL;
    proxy_req = NULL;

    setPassive();
}    

NestReplyStatus
nestClient::assignStorageConnection( nestProtocol *storage_con )
{
    storage = storage_con;
    return NEST_SUCCESS;
}

NestReplyStatus
nestClient::assignOtherStorageConnection( nestProtocol *storage_con )
{
    otherStorage = storage_con;
    return NEST_SUCCESS;
}

void 
nestClient::assignThirdPartyConnection( nestProtocol *third )
{
    if ( third != NULL ) assert ( third_party == NULL );
    third_party = third;
}

nestProtocol *
nestClient::getThirdPartyConnection( )
{
    return third_party;
}



// this is called by the concurrency architectures
NestReplyStatus
nestClient::openThirdParty( nestProtocolType protocol )
{

    NestReplyStatus status   = NEST_SUCCESS;
    nestProtocol *other = NULL;
    nestProtocol *src, *dst;

    status = protocol_manager->initiateThirdPartyTransfer(
            this, &other, protocol );
    if ( status != NEST_SUCCESS ) {
        return status;
    } else {
        assignThirdPartyConnection( other );
    }

    switch( req->getType() ) {
    case THIRD_PARTY_GET:
        src = other;
        dst = storage;
        break;
    case THIRD_PARTY_PUT:
        dst = other;
        src = storage;
        break;
    default:
        assert( 0 );
        return NEST_UNKNOWN_REQUEST_TYPE;
    }
    
    this->transfer = new nestFiletransfer(src,dst, req->getLength(),
            buffer, FILEBUFFER, req->getType() );

    this->transfer->startTransfer();

    // special case for transfers that put data but don't prespecify 
    // a transfer length.  we can only allocate lot space for these
    // transfers incrementally.
    long filesize = req->getLength();
    if(filesize == ENTIRE_FILE) {
        this->transfer->setLimit((long)allocation);
    }

    return NEST_SUCCESS;
}

// this gets called by the main admin thread before the task is handed
// off to a concurrency
NestReplyStatus
nestClient::startTransfer( )
{
    assert( req != NULL );
    return startTransfer( req->getType(), req->getLength() );
}

// this might get called by any special concurrency tasks (e.g. proxies)
NestReplyStatus
nestClient::startTransfer( int type, long length )
{
	dump();

    nestProtocol *src, *dst;

    switch( type ) {
    case CLIENT_SEND_FILE: 
    case PROXY_GET:
        src = connection;
        dst = storage;
        break;
    case CLIENT_RECV_FILE: 
    case PROXY_PUT:
        src = storage;
        dst = connection;
        break;
    case TEST_CLIENT_TO_SERVER:
        src = connection;
        dst = NULL;
        break;
    case TEST_SERVER_TO_CLIENT:
        src = NULL;
        dst = connection;
        break;
    case THIRD_PARTY_GET:
    case THIRD_PARTY_PUT:
            // nothing to do here
        return NEST_SUCCESS; 
    case COPY_REQUEST:
        src = otherStorage;
        dst = storage;
        break;
    default:
        nest_die( "Can't do %s:%s with unknown request type %d\n",
                    __FILE__, __FUNCTION__, req->getType() );
        return NEST_UNKNOWN_REQUEST_TYPE;
    }

        // don't start 3pt here (although we used to)
        // because the connection process might be slow, each individual
        // concurrency architecture should do it

        // go ahead and create transfer object
    this->transfer = new nestFiletransfer(src, dst, length, 
										  buffer, FILEBUFFER, type );
    this->transfer->startTransfer();

    // special case for transfers that put data but don't prespecify 
    // a transfer length.  we can only allocate lot space for these
    // transfers incrementally.
    //long filesize = req->getEnd();
    if(length == ENTIRE_FILE) {
        assert( allocation > 0 );
        this->transfer->setLimit((long)allocation);
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestClient::endThirdPartyTransfer()
{

    nestProtocol *connection;
    switch( req->getType() ) {
    case THIRD_PARTY_PUT:
        connection = transfer->getDst();
        break;
    case THIRD_PARTY_GET:
        connection = transfer->getSrc();
        break;
    default:
        assert( 0 );
        return NEST_UNKNOWN_REQUEST_TYPE;
    }

    return ( connection->endThirdPartyTransfer( req->getType() ) 
            ? NEST_SUCCESS
            : NEST_NO_CONNECTION );
}

void
nestClient::endTransfer( NestReplyStatus status )
{
    nest_debug( NEST_DEBUG_CLIENT | NEST_DEBUG_TRANSFER, 
            "client %d: endTransfer %s\n", id, 
            NestErrorString( status ) ); 
    dump();

    // TODO: if transfer was a 3pt close the third party connection here
    // actually protocols should close it themselves when it is deleted

    /* 
    // in 3pt or proxy, we might not yet have instantiated
    //assert( transfer != NULL ); 
    */
    if ( transfer != NULL ) {
        transfer->endTransfer( status );
        if ( last_transfer ) delete( last_transfer );
        last_transfer = transfer;
        transfer = NULL;
    }

    my_endRequest( status );

    nest_debug( NEST_DEBUG_CLIENT | NEST_DEBUG_TRANSFER, 
            "client %d (sd %d): endTransfer\n", id, socket );
}

nestFiletransfer*
nestClient::getLastTransfer( )
{
    return last_transfer;
}

nestFiletransfer*
nestClient::getTransfer( )
{
    return transfer;
}

nestRequest*
nestClient::getRequest( )
{
    return req;
}

nestRequest*
nestClient::getLastRequest( )
{
    return last_req;
}

nestRequest*
nestClient::getLastProxyRequest( )
{
    return last_proxy_req;
}

nestProtocol *
nestClient::getConnection( )
{
    return connection;
}

void
nestClient::assignConnection( nestProtocol *prot )
{
    if ( prot != NULL ) assert( connection == NULL );
    connection = prot;
}

// for simple two-party transfers, data is exchanged between client and storage
nestProtocol *
nestClient::getStorageConnection( )
{
    return storage;
    /*
    return ( transfer->getSrc() != this->connection )
        ? transfer->getSrc() : transfer->getDst();
    */
}

nestProtocol *
nestClient::getOtherStorageConnection( )
{
    return otherStorage;
}

bool
nestClient::isActive( )
{
    return ( state == CLIENT_ACTIVE );
}

void
nestClient::setState( int newstate )
{
    if ( this->state == newstate ) 
        nest_warn( __FILE__, __FUNCTION__,"warning client state\n");
    this->state = newstate;
}

void
nestClient::setActive( )
{
    assert( state == CLIENT_PASSIVE );
    setState( CLIENT_ACTIVE );
}

void
nestClient::setPassive( )
{
    assert( state == CLIENT_ACTIVE );
    setState( CLIENT_PASSIVE );
}

void
nestClient::setLastAllocation(nestLotSize allocation)
{
    last_allocation = allocation;
}

nestLotSize
nestClient::getLastAllocation( )
{
    return last_allocation;
}

void
nestClient::setAllocation(nestLotSize allocation)
{
    this->last_allocation = this->allocation;
    this->allocation = allocation;
}

nestLotSize
nestClient::getAllocation( )
{
    return allocation;
}
