/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "errno.h"
#include "nest_dsi_finish.h"
#include "nest_dsi_session.h"
#include "nest_dsi_logger.h"
#include "nest_error.h"

// Generic constructor
nestDsiFinish::nestDsiFinish( globus_gfs_operation_t gfsOp )
{
	this->gfsOp = gfsOp;
	this->result_set = false;
}

nestDsiFinish::~nestDsiFinish( void )
{
}

globus_result_t
nestDsiFinish::genResult( const char *func,
						  int my_errno )
{
	(void) func;
    return globus_error_put (
		globus_error_construct_errno_error(
			GLOBUS_NULL, GLOBUS_NULL, my_errno ) );
}

const char *
nestDsiFinish::TranslateNestStatus( NestReplyStatus status )
{
	static const char *StatusStrings[] =	{
		0,				//NEST_SUCCESS
    	"Local file not found",
    	"Remote file not found",
    	"Insufficient space",
    	"No connection",
    	"Bad parameters",
    	"Not yet implemented",
    	"Unknown request type",
    	"Nonpriveledged user",
    	"Unknown user",
    	"Nonpriveledged server",
    	"Already exists",
    	"Not empty",
    	"Partial message",
    	"Unknown error",
    	"Temporarily unavailable",
    	"Unknown debug value",
    	"Invalid authentication",
    	"Could not authenticate",
    	"Out of resources",
    	"System file missing",
    	"File exists",
    	"File not found",
    	"Too many links",
    	"Filename too long",
    	"User exists",
    	"User not found",
    	"User limit reached",
    	"Group exists",
    	"Group not found",
    	"Group limit reached",
    	"Group active",
    	"Disk full",
    	"No quota enforcement",
    	"Quota enforcement error",
    	"Lot scheduled",
    	"Lot not found",
    	"Lot not enough",
    	"Lot not empty",
    	"Lot limit reached",
    	"Lot invalid",
    	"Lot disabled",
    	"Classad error",
    	"Stale fh",
    	"Intr",
    	"Flock",
    	"System file update failed",
    	"Lot update failed",
    	"Inconsistent state",
    	"Not yet initialized",
    	"Not a file",
		"Bad collection",
    	"Not directory",
    	0			// NEST_ERROR_MAX
	};
	static int		max = (sizeof( StatusStrings ) / sizeof( char * ))  - 1;
	if ( NEST_ERROR_MAX != max )
	{
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "Error map mismatch (Expect %d, found %d)\n",
						   NEST_ERROR_MAX, max );
		assert( 0 );
	}
	if ( ( status < 0 ) || ( status >= max ) )
	{
		return "Unknown";
	}
	else
	{
		return StatusStrings[(int)status];
	}
}

void
nestDsiFinish::setResult( NestReplyStatus status, const char *string )
{
	GlobusGFSName( nestDsiFinish::setResult );

	// Don't do it twice...
	if ( result_set ) {
		return;
	}

	// Default string
	if ( NULL == string ) {
		string = "Unknown";
	}

	// Look at what we've got...
	switch( status )
	{
	case NEST_SUCCESS:
		result = (globus_result_t) GLOBUS_SUCCESS;
		break;
	default:
		const char*	string = TranslateNestStatus( status );
		result = GlobusGFSErrorGeneric( string );

		_logger->logEvent( NEST_DSI_LOG_FINISH,
						   "setResult: status = %d, str = %s, result = %p\n",
						   status, string, result );
		break;
	}

	// Done
	result_set = true;
}

void
nestDsiFinish::setResult( int errno, const char *string )
{
	// Don't do it twice...
	if ( result_set ) {
		return;
	}

	// Default string
	if ( NULL == string ) {
		string = "Unknown";
	}

	if ( 0 == errno ) {
		result = (globus_result_t) GLOBUS_SUCCESS;
	} else {
		result = genResult( string, errno );
	}

	// Done
	result_set = true;
}

void
nestDsiFinish::setResult( globus_result_t	result )
{
	// Don't do it twice...
	if ( result_set ) {
		return;
	}

	this->result = result;
	this->result_set = true;

}


// ==================================================
// Base finish with "Info"
// ==================================================
nestDsiFinishInfo::nestDsiFinishInfo( globus_gfs_operation_t	gfsOp,
									  globus_gfs_operation_type_t type )
		: nestDsiFinish( gfsOp )
{
	info.type = type;
}

nestDsiFinishInfo::~nestDsiFinishInfo( void )
{
}

int
nestDsiFinishInfo::setCode( int code )
{
	info.code = code;
	return 0;
}

void
nestDsiFinishInfo::finished(bool valid, globus_result_t result )
{
	// Store the result if not already set
	if ( ! this->result_set ) {
		if ( ! valid ) {
			this->result = genResult( "Unknown", EINVAL );
		} else if ( GLOBUS_SUCCESS == result ) {
			this->result = GLOBUS_SUCCESS;
		} else {
			this->result = result;
		}
		this->result_set = true;
	}

	_logger->logEvent( NEST_DSI_LOG_FINISH,
					   "Finished: op=%d, result=%d, info=%p\n",
					   (int)gfsOp, this->result, &info );

    globus_gridftp_server_operation_finished(
		gfsOp,
		this->result,
		&info );
	_logger->logEvent( NEST_DSI_LOG_FINISH,
					   "Finish: back from finished\n" );

	return;
}


// ==================================================
// Session finished
// ==================================================
nestDsiFinishSession::nestDsiFinishSession( globus_gfs_operation_t	gfsOp,
											nestDsiSession			*session )
		: nestDsiFinishInfo( gfsOp, GLOBUS_GFS_OP_SESSION_START )
{
    // arg to pass back with each request
	info.info.session.session_arg = session;
	const char *name = session->getUsername( );
	if ( NULL == name ) {
		info.info.session.username = NULL;
	} else {
		info.info.session.username = strdup( name );
	}
	info.info.session.home_dir = NULL;
}

nestDsiFinishSession::~nestDsiFinishSession( void )
{
	if ( info.info.session.username ) {
		free( (void*) info.info.session.username );
	}
	if ( info.info.session.home_dir ) {
		free( (void*) info.info.session.home_dir );
	}
}


// ==================================================
// Command finished
// ==================================================
nestDsiFinishCmd::nestDsiFinishCmd(
	globus_gfs_operation_t			gfsOp,
	const globus_gfs_command_info_t	*cmd )
		: nestDsiFinishInfo( gfsOp, GLOBUS_GFS_OP_COMMAND )
{

    // arg to pass back with each request
	info.info.command.command = cmd->command;
	info.info.command.checksum = NULL;
	info.info.command.created_dir = NULL;
}

nestDsiFinishCmd::~nestDsiFinishCmd( void )
{
	if ( info.info.command.checksum ) {
		free( (void*) info.info.command.checksum );
	}
	if ( info.info.command.created_dir ) {
		free( (void*) info.info.command.created_dir );
	}
}

int
nestDsiFinishCmd::setCreatedDir( const char *dir )
{
	if ( info.info.command.created_dir ) {
		free( (void*) info.info.command.created_dir );
	}
	info.info.command.created_dir = strdup( dir );
	return( info.info.command.created_dir != NULL );
}

int
nestDsiFinishCmd::setChecksum( const char *cksum )
{
	if ( info.info.command.checksum ) {
		free( (void*) info.info.command.checksum );
	}
	info.info.command.checksum = strdup( cksum );
	return( info.info.command.checksum != NULL );
}



// ==================================================
// Stat finished
// ==================================================
nestDsiFinishStat::	nestDsiFinishStat( globus_gfs_operation_t gfsOp )
	: nestDsiFinish( gfsOp )
{
}

nestDsiFinishStat::~nestDsiFinishStat( void )
{
}

// Failed stat finish
void
nestDsiFinishStat::finished( void )
{
	globus_gridftp_server_finished_stat (
		gfsOp,
		this->result,
		NULL,
		0 );
}

// Successful stat finish
void
nestDsiFinishStat::finished( 
	globus_gfs_stat_t	*statarray,
	int					count )
{
	globus_gridftp_server_finished_stat (
		gfsOp,
		GLOBUS_SUCCESS,
		statarray,
		count );
}



// ==================================================
// Transfer finished
// ==================================================
nestDsiFinishTransfer::	nestDsiFinishTransfer( globus_gfs_operation_t gfsOp )
	: nestDsiFinish( gfsOp )
{
}

nestDsiFinishTransfer::~nestDsiFinishTransfer( void )
{
}

// Failed stat finish
void
nestDsiFinishTransfer::finished( void )
{
	if ( result_set ) {
		_logger->logEvent( NEST_DSI_LOG_FINISH,
						   "transfer finished: result=%0X\n", result );
		globus_gridftp_server_finished_transfer( gfsOp, result );
	}
	else {
		globus_gridftp_server_finished_transfer( gfsOp, GLOBUS_SUCCESS );
	}
}

// Successful stat finish
void
nestDsiFinishTransfer::finished( NestReplyStatus	status )
{
	setResult( status );
	globus_gridftp_server_finished_transfer( gfsOp, result );
}

