/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_config.h"
#include "nest_storage_manager.h"
#include "nest_io_manager.h"
#include "nest_scheduler.h"
#include "nest_request_handler.h"
#include "nest_main.h"

// Derive from nestConcurrencyBase so they have a comment ancestry
// for the callback
struct ThreadData
{
	// Thread specific stuff
	int					thread_id;
	pthread_t			thread;
	struct ThreadData	*next;			// Next thread
};

// the callback to notify the main server
void (*notifyServer) (nestClient *, NestReplyStatus );

    // private helper functions
void * workerLoop( void * );
NestReplyStatus workerRequestHandler( int thread_id,
									  nestClient *client );

nestIoManager::nestIoManager( void )
{

    this->pendingTasks = new nestClientPool(  true );
    this->doneTasks    = new nestClientPool(  true );

	// ** from POT **

        // set up a pipe to communicate to server
    if ( NestSpipe( server_pipe ) < 0 ) {
        nest_perror( "NestSpipe" );
        assert( 0 );
    }
    NestSetFileFlags( server_pipe[0], O_NDELAY );
	NestFdRegister( server_pipe[0], FD_READ );
    maxfd = server_pipe[1];

        // set up the args
    this->poolSize       = -1;

	admission_control = INT_MAX;
	task_queue = NULL;

	max_concurrency = 0;
	current_tasks = 0;
	ave_concurrency = 0;
	total_sizes = 0;

	unscheduled_tasks    = 0;
	fifoscheduled_tasks  = 0;
	scheduling_policy = NEST_FIFO;
	favoriteProtocol = NEST_PROTOCOL;

	bandwidth = -1;
	kbsIn = kbsOut = 0;
	transfers = errors = 0;

}

nestIoManager::~nestIoManager( void )
{
	// TODO:  This is crude, but will probably work unless the
	// threads are doing real work at the time...
	ThreadData	*tdata = this->first_thread;
	while( tdata ) {
		pthread_cancel( tdata->thread );
		tdata = tdata->next;
	}

	// Join them
	tdata = this->first_thread;
	while( tdata ) {
		ThreadData	*next;
		pthread_join( tdata->thread, NULL );

		// And, free up the memory
		next = tdata->next;
		NestFree( tdata );
		tdata = next;
	}
}

int
nestIoManager::init( void )
{
    int worker_count = -1;

    const char *value = NULL;

    if (GetConfigValue(Config_ThreadPoolSize, &value) && *value != 0 ) {
        worker_count = atoi( value );
    } else {
        nest_die( "Must specify size of thread pool.\n" );
    }

    if (GetConfigValue(Config_AdmissionControl, &value) && *value != 0 ) {
        admission_control = atoi(value);
    }

    if (GetConfigValue(Config_SchedulingPolicy, &value) && *value!=0) {
        if ( !strcmp(value, "fifo" ) ) {
            scheduling_policy = NEST_FIFO;
        } else if ( !strcmp(value, "sff" )  ) {
            scheduling_policy = NEST_SFF;
        } else {
            nest_die( "Bad arg (%s) for schedule command.\n", value );
        }
    }

    if (GetConfigValue(Config_PreferredProtocol, &value) && *value!=0){
        if ( !strcmp(value, "http" )  ) {
            favoriteProtocol = HTTP_PROTOCOL; 
        } else if ( !strcmp(value, "chirp" ) ) {
            favoriteProtocol = NEST_PROTOCOL;
        } else {
            nest_die( "Bad arg (%s) for protocol preference.\n", value );
        }
        nest_printf( "Preferring %s protocol.\n", value );
        scheduling_policy = NEST_PROTOCOL_PREFERENCE;
    }

        // seed random to be used for lottery concurrency selection
    srand( time(NULL) );

        // set up the classes we need for scheduling decisions
        // don't make task_queue thread safe, we'll do our own protection
    task_queue = new nestClientPool( false );

        // first start with equal strides
    for( int p = 0; p < PROTOCOL_COUNT; p++ ) {
        protocol_stats[p].kbs_consumed  = 0;
        protocol_stats[p].current_tasks = 0;
        protocol_stats[p].stride        = 1;
    }
    //protocol_stats[HTTP_PROTOCOL].stride = 4;

    this->poolSize       = worker_count;
    nest_pthread_mutex_init( &(this->mutex), NULL );

        // pass args to each newly created thread worker
	int			tid;
	this->first_thread = NULL;
	ThreadData	*prev = NULL;
    for ( tid = 0; tid < poolSize; tid++ )
    {
		// Allocate thread specific structure
		struct ThreadData   *tdata;
		tdata = (struct ThreadData *) NestMalloc( sizeof(struct ThreadData) );
		if ( ! tdata ) {
			nest_die( "Can't allocate thread space for thread %d\n", tid );
			assert( 0 );
		}

		// Fill it in
		tdata->thread_id = tid;		// The thread's "ID"
		tdata->next = NULL;			// Currently no next

		// Save off pointer
		if ( prev ) {
			prev->next = tdata;
		}
		prev = tdata;

		// Store off a pointer to the first one
		if ( ! this->first_thread ) {
			this->first_thread = tdata;
		}

		// And, finally, create the thread
        nest_pthread_create( &(tdata->thread),
							 NULL,
							 workerLoop,
							 (void *)tdata );
    }

    return 0;
}

// this function is used to compute the new weighted average 
// it tries to respect the new value without forgetting history too quickly
float
adjust_ave( float ave, float newval )
{
    
        // no bandwidth should ever be no-pos
    if ( newval <= 0 ) {
        nest_warn( __FILE__, __FUNCTION__,
				   "Ignoring weird new bandwidth %.6f.\n", newval );
        return ave;
    }
    
        // if first time
    if ( ave == -1 ) {
            return newval;
    } else {
            // slowly age the average
        float new_ave = ( 0.95 * ave + 0.05 * newval );
        if ( new_ave <= 0 ) {
            nest_warn( __FILE__, __FUNCTION__,
					   "New bandwidth %.6f is <= 0....?", new_ave );
            return 0.00001;
        } else {
            return new_ave;
        }
    }
}


// this function is called whenever a task is completed
// return responsibility to server to listen again to the client
void
nestIoManager::taskDone( nestClient *client, NestReplyStatus status )
{

        // look up the right model
    taskFunction *chore = client->getTaskFunction();


        // end the transfer here, but don't tell the client that it's ended
        // once the transfer has been ended, the client might return
        // a closeConnectionRequest which will result in the client getting
        // deleting
    nestFiletransfer *transfer = client->getTransfer();
    const char *file = NULL;
    if ( transfer != NULL ) transfer->endTransfer( status );

    lockMutex();
    current_tasks--;

        // adjust values only for successful transfers of positive size
        // ignore transfers of size 0, don't even count them
    long len = 0;
    if ( transfer != NULL ) {
        len = transfer->getBytesTransferred();
        if ( client->getRequest() != NULL ) {
            file = client->getRequest()->getNestFile();
        } else {
            assert( client->getProxyFetch() != NULL );
            file = client->getProxyFetch()->getProxyFile();
        }
    }

    if ( status == NEST_SUCCESS && len > 0 ) {
        double trans_bandwidth = transfer->getBandwidth();
        if ( trans_bandwidth < 0 ) {
            nest_warn( __FILE__, __FUNCTION__,
					   "Received negative bandwidth %.3f "
					   "for a succesful transfer.\n",
					   trans_bandwidth );
        }

            // collect bandwidth info for future selections may want
            // to keep seperate bandwidth info for short / long
            // transfers and maybe for different protocols as well
        bandwidth = adjust_ave( bandwidth, trans_bandwidth );
        total_bandwidth += trans_bandwidth;
        transfers++;
        nest_debug( NEST_DEBUG_CONCURRENCY, 
                "transfer (%s) ended: %s. "
					"%ld bytes transferred (%.3f MB/s)\n",
					protocol_manager->protocolName( client->getProtocol() ),
					NestErrorString( status ), len, trans_bandwidth );
        if ( transfer->isIncoming() ) {
            kbsIn += (len / 1024);
        } else {
            kbsOut += (len / 1024);
        }

        if ( scheduling_policy != NEST_PROTOCOL_PREFERENCE ) {
            protocol_stats[client->getProtocol()].kbs_consumed +=
				(len / 1024);
        }

    } else if ( status != NEST_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__,
				   "Task failed: %s\n", NestErrorString( status ) );
        /*
        if ( status == NEST_NO_CONNECTION ) {
            // TODO: maybe we should set a flag in the client so it
            // will next return a close_connection_request
            // I think this should happen anyway though
            //client->setAlive( false );
        }
        */
        errors++;
    }

            // stride scheduling decisions here (if this was normal transfer)
    if ( chore == NULL ) {
        if ( scheduling_policy == NEST_PROTOCOL_PREFERENCE ) {
            strideDone( client, len );
            makeStrides(); 
        } else {
                // check for ability to schedule next task 
            makeSchedule();
        }
    }
    

        // return responsibility to server 
    req_handler->taskDone( client, status ); 

    unlockMutex();
    return;
    
}

int
nestIoManager::makeSchedule( )
{
    int started = 0;
    while ( task_queue->size() > 0 && current_tasks < admission_control ) {
        nestClient *nextclient = selectQueuedTask();
        nest_debug( NEST_DEBUG_CONCURRENCY, 
                "starting queued task for client %d. Queue len %d\n", 
                nextclient->getId(), task_queue->size() + 1 );
        startTask( nextclient );
        started++;
    }
    return started;
}

void
nestIoManager::debugLevel( long level )
{
}

ssize_t
nestIoManager::writeAd( nestProtocol *ad )
{
    char buffer[NEST_MESSAGESIZE];

	const char *sched_names[] = { "FIFO", "Shortest File First", "Protocol" };
    NestSnprintf( &buffer[NestStrlen(buffer)], 
				  NEST_MESSAGESIZE - NestStrlen(buffer), 
				  "Scheduling = \"%s\"\n", sched_names[scheduling_policy] );

    if ( admission_control != INT_MAX || task_queue->size() ) {
        NestSnprintf( &buffer[NestStrlen(buffer)],
                      NEST_MESSAGESIZE - NestStrlen(buffer),
                      "Admission_Control = %d\n"
                      "Ave_Queue_Len  = %.4f\n"
                      "Max_Queue_Len  = %d\n"
                      "Cur_Queue_Len  = %d\n"
                      "Total_Q_Admits = %ld\n"
                      "Immediate_Scheduled = %ld\n"
                      "Fifo_Scheduled      = %ld\n",
                      admission_control,
                      task_queue->getAveLen(),
                      task_queue->getMaxLen(),
                      task_queue->size(),
                      task_queue->getAdmissions(),
                      unscheduled_tasks,
                      fifoscheduled_tasks );
    }

        // stride scheduling stuff
    for( int p = 0; p < PROTOCOL_COUNT; p++ ) {
        char protstring[25];
        NestSnprintf( protstring, sizeof(protstring), "%s_kbs",   
                      protocol_manager->protocolName( p ) );
        NestSnprintf( &buffer[NestStrlen(buffer)],
					  NEST_MESSAGESIZE - NestStrlen(buffer),
					  "%15s = %15ld\n",
					  protstring,
					  ( protocol_stats[p].kbs_consumed *
						protocol_stats[p].stride ) );
    }

    int ret = NestSnprintf( &buffer[NestStrlen(buffer)],
                            NEST_MESSAGESIZE - NestStrlen(buffer),
                            "Total_kilobytes_in = %ld\n"
                            "Total_kilobytes_out = %ld\n"
                            "Total_transfers = %ld\n"
                            "Total_errors = %ld\n"
                            "Average_transfer_bandwidth =%.4f\n"
                            "Current_concurrency = %d\n"
                            "Average_concurrency = %.4f\n"
                            "Max_concurrency = %d\n",
                            kbsIn, kbsOut, transfers, errors,
                            total_bandwidth / transfers,
                            current_tasks,
                            ( (double)ave_concurrency /
                              (transfers + errors + current_tasks) ),
                            max_concurrency );
    if ( ret < 0 ) {
        nest_warn( __FILE__, __FUNCTION__,
				   "buffer truncated in %s:%s.\n", __FILE__, __FUNCTION__ );
    }

    ssize_t s = ad->Sendn( buffer, NestStrlen( buffer ) );
    return s;
}

NestReplyStatus
nestIoManager::lockMutex( )
{
    return NEST_SUCCESS;
}

NestReplyStatus
nestIoManager::unlockMutex( )
{
    return NEST_SUCCESS;
}

void
nestIoManager::strideDone( nestClient *client, long stride )
{
    int prot = client->getProtocol();
    long adjust = 0;
    
    if ( stride != client->getTransfer()->getLength() ) {
            // need to adjust - we initially set it either to the length
            // if it was known or we guessed
        long init = client->getTransfer()->getLength();
        if ( init < 0 ) init = client->getTransfer()->getLengthGuess();
        adjust = ( stride - init / 1024 ) / protocol_stats[prot].stride; 
    }

    if ( adjust ) protocol_stats[prot].kbs_consumed += adjust; 
    protocol_stats[prot].current_tasks--;

    nest_debug( NEST_DEBUG_CONCURRENCY | NEST_DEBUG_PROTOCOL_SCHED,
            "strideDone: client %d prot %d,  %ld total kbs, %d current\n",
            client->getId(),
            prot, protocol_stats[prot].kbs_consumed,
            protocol_stats[prot].current_tasks );

    assert( protocol_stats[prot].kbs_consumed  >= 0 );
    assert( protocol_stats[prot].current_tasks >= 0 );
}

NestReplyStatus
nestIoManager::makeStrides( )
{

    nestClient *client = NULL;
    NestReplyStatus status = NEST_SUCCESS;

    while ( task_queue->size() > 0 && ( client = selectLowestStride() ) ) {
            // here is a new client to schedule
        status = startTask( client );
        assert( status == NEST_SUCCESS );
        
            // adjust stride info
        long len = client->getTransfer()->getLength();
        if ( len < 0  ) {
                // we don't know the transfer length, just guess
            nest_debug( NEST_DEBUG_PROTOCOL_SCHED,
                    "Guessing at transfer length unknown.\n" );
            len = 8 * 1024; 
            client->getTransfer()->setLengthGuess( len );
        }
        int prot = client->getProtocol();
        long stride = ( len / 1024 ) / protocol_stats[prot].stride;
        protocol_stats[prot].kbs_consumed += stride; 
        protocol_stats[prot].current_tasks++;
        nest_debug( NEST_DEBUG_CONCURRENCY | NEST_DEBUG_PROTOCOL_SCHED,
                "Starting new stride: "
                "client %d: prot %d, %ld total kbs, %d current\n",
                client->getId(), prot, protocol_stats[prot].kbs_consumed,
                protocol_stats[prot].current_tasks );
    }
    return status;
}

unsigned long
nestIoManager::findLowestStride( )
{
        // first find lowest consumption of all active protocols
    unsigned long lowest = ULONG_MAX;
    for( int p = 0; p < PROTOCOL_COUNT; p++ ) {
        if (   protocol_stats[p].current_tasks > 0 
            && protocol_stats[p].kbs_consumed < lowest ) 
        {
            lowest = protocol_stats[p].kbs_consumed;
        }
    }
    return lowest;
}

// the counterpart to this function is taskDone 
// however this function is always called by admin thread
// whereas taskDone can be called by thread workers (actually not anymore)
NestReplyStatus
nestIoManager::addTask( nestClient *client )
{ 

    NestReplyStatus status = NEST_SUCCESS;
    bool anticipatory = false;

        // take over responsibility for this client from the server
        // when task is done, return it
    if ( client->hasSocket() ) {
		NestFdUnregister( client->getConnection()->getSocket(), FD_READ );
    }

    lockMutex();

        // new stride scheduling stuff
    if ( scheduling_policy == NEST_PROTOCOL_PREFERENCE ) {
        nest_debug( NEST_DEBUG_CONCURRENCY | NEST_DEBUG_PROTOCOL_SCHED,
                "Queuing client %d\n", client->getId() );
        task_queue->addClient( client );
        status = makeStrides( );
    } else if ( anticipatory ) {
            // just queue it now, schedule later
        nest_debug( NEST_DEBUG_CONCURRENCY | NEST_DEBUG_PROTOCOL_SCHED,
                "Queuing client %d\n", client->getId() );
        task_queue->addClient( client );
    } else {
            // check for admission_control
        if ( current_tasks >= admission_control ) {
            assert( current_tasks == admission_control );
            task_queue->addClient( client );
            nest_debug( NEST_DEBUG_TRANSFER | NEST_DEBUG_CONCURRENCY, 
                "Task queued (len %d) for client %d. Queue len %d\n", 
                client->getRequest()->getLength(),
                client->getId(), task_queue->size() );
            status = NEST_SUCCESS;
        } else {
            status = startTask( client );
            unscheduled_tasks++;
        }
    }

    unlockMutex( );
    return status; 
}

NestReplyStatus
nestIoManager::startTask( nestClient *client )
{
    NestReplyStatus status = NEST_SUCCESS;

        
        // if we make it here, we're OK, go ahead and start the transfer
        // start the transfer here, end at taskDone
        // this will allow us to measure bandwidth
        // note that pop worker will need to start it's own transfer

        // only do this stuff for normal transfer functions
        // the startTransfer might also fail especially with lot 
        // unavailability. 
    if ( client->getTaskFunction() == NULL ) {
        status = client->startTransfer();
        NEST_REPLY_CHECK(status);
    }
    pendingTasks->addClient( client );

        // adjust all our accounting info
    current_tasks++;
    max_concurrency = NEST_MAX( max_concurrency, current_tasks );
    ave_concurrency += current_tasks;

    nest_debug( NEST_DEBUG_CONCURRENCY | NEST_DEBUG_TRANSFER, 
				"New task added; Concurrency is %d\n", 
				current_tasks );

    return status;
}

nestClient *
nestIoManager::selectFavoriteProtocol()
{
    CLIENT_ITERATOR itr = task_queue->begin();
    nestClient *cursor = NULL;

        // if our favorite protocol is found, return it
        // else just the head
    while( itr != task_queue->end() ) {
        cursor = *itr++;
        if ( cursor->getProtocol() == favoriteProtocol ) {
            task_queue->removeClient( cursor );
            nest_debug( NEST_DEBUG_PROTOCOL_SCHED,
                    "Scheduling a favorite protocol.\n" );
            return cursor; 
        }
    }
        // no favorite protocols found, just return head
    return task_queue->getFront();
}

nestClient *
nestIoManager::selectQueuedTask( )
{

    nestClient *client = NULL;
    
    switch( scheduling_policy ) {
    case NEST_PROTOCOL_PREFERENCE:
        client = selectFavoriteProtocol();
        break;
    case NEST_FIFO:
        assert( ! task_queue->empty() );
        client = task_queue->getFront();
        break;
    case NEST_SFF:
        client = selectShortestFile();
        break;
    default:
        nest_die( "Unknown scheduling policy\n" );
        return NULL;
    }
    return client;
}

nestClient *
nestIoManager::getFrontTask( void )
{
	return pendingTasks->getFront( true );

}

void
nestIoManager::addDoneTask( nestClient *client )
{
	doneTasks->addClient( client );
}

// this function is called to make sure that server breaks out of its main
// select loop after a task has been completed
// otherwise the server will not be listening to the client
void
nestIoManager::signalNest( int id )
{
    static char msg[] = "POT wakeup\n";
    NestWriten( server_pipe[1], msg, NestStrlen( msg ) );    
    nest_debug( NEST_DEBUG_CONCURRENCY, "Thread %d signaled dispatcher.\n", id );
}

nestClient *
nestIoManager::selectLowestStride( )
{
    unsigned long lowest = findLowestStride();

    CLIENT_ITERATOR itr = task_queue->begin();
    nestClient *cursor, *low = NULL;
    while( itr != task_queue->end() ) {
        cursor = *itr++;
        if ( protocol_stats[cursor->getProtocol()].kbs_consumed <= lowest ) {
            lowest = protocol_stats[cursor->getProtocol()].kbs_consumed;    
            low = cursor;
        }
    }

        // could be NULL if there are no queued clients for lowest strides
    if ( low != NULL ) task_queue->removeClient( low );
    return low;
}

nestClient *
nestIoManager::selectShortestFile( )
{
    CLIENT_ITERATOR itr = task_queue->begin();
    nestClient *cursor, *shortest = NULL;
    int shortestLength = INT_MAX; 

    while( itr != task_queue->end() ) {
        cursor = *itr++;
        if ( cursor->getRequest()->getLength() < shortestLength ) {
            shortest = cursor; 
            shortestLength = shortest->getRequest()->getLength();
        }
    }
    assert( shortest != NULL );
    task_queue->removeClient( shortest );
    return shortest;
}

void
nestIoManager::makeProgress( )
{
	// ** from POT **
    NestReplyStatus status = NEST_SUCCESS;

        // check to see if the server has started ignoring us.
        // why would that happen?
    if ( ! NestFdIsRegistered( server_pipe[0], FD_READ ) ) {
        nest_die( "Somehow the dispatcher has decided to ignore "
                  "the thread pipe.\n" );
    }

        // clear the server signal pipe here so we can signal later
    if ( NestFdIsSet( server_pipe[0], FD_READ ) ) {
        status = clearNestSignal();
        if ( status != NEST_SUCCESS ) {
            nest_warn( __FILE__, __FUNCTION__,
					   "Unable to clear pot signal: %s\n", 
					   NestErrorString( status ) );
        } 
    }

	// call taskDone only in admin thread
	while( doneTasks->size() ) {
        //for( int i = 0; i < thread_args->doneTasks->size(); i++ ) {
		nestClient *client = doneTasks->getFront();
		assert( client != NULL );
		NestReplyStatus status;
		taskFunction *fp = client->getTaskFunction();
		if ( fp != NULL ) {
			status = client->getTaskStatus();
		} else {
			assert( client->getRequest() != NULL );
			status = client->getRequest()->getStatus();
		}
		taskDone( client, status );
	}

}

// this function is called after the server breaks out of the select loop
// just clears the signal so that it can be used again
NestReplyStatus
nestIoManager::clearNestSignal( )
{
    static char msg[FILEBUFFER];
    int n = 0;
    int clear = 0;
    nest_debug( NEST_DEBUG_CONCURRENCY, "Clearing nest signals from pot\n" );

    nest_pthread_mutex_lock( &mutex );
    while (( n = NestRead( server_pipe[0], msg, FILEBUFFER )) > 0){
        nest_debug( NEST_DEBUG_CONCURRENCY, "Cleared %d bytes: %s\n", n, msg );
        clear++;
    }
    nest_pthread_mutex_unlock( &mutex );

        // if nothing was cleared return error
        // otherwise return success as long as errno is reasonable
    if ( clear == 0 ) return NEST_TEMPORARILY_UNAVAILABLE;
    return ( n == 0 || ( n == -1 && errno == EAGAIN ) )
            ? NEST_SUCCESS : ErrnoToNest( errno );
}

// For thread
NestReplyStatus
nestIoManager::workerRequestHandler(
	int worker_id,
	nestClient *client )
{
    NestReplyStatus status = NEST_SUCCESS;

        // special case if client provides their own concurrency function
    taskFunction *fp = client->getTaskFunction();
    if ( fp != NULL ) {
        NestReplyStatus status = fp( client );
        nest_debug( NEST_DEBUG_PROXY, 
            "Thread %d handled special request from client %s(%d): %s\n", 
            worker_id, client->getUserIdentity(), client->getId(),
            NestErrorString( status ) );
        client->setTaskStatus( status );
        return status;
    }

    // TODO if 3pt, start it here
    // any 3pt is initiated here
    switch( client->getRequest()->getType() ) {

	case CLIENT_SEND_FILE:
		// TODO check error value
		client->sendInitialAck( );
		break;

	case CLIENT_RECV_FILE:
		// TODO check error value
		client->sendTransferInfo( );
		break;

	case THIRD_PARTY_GET:
	case THIRD_PARTY_PUT:
		nest_debug( NEST_DEBUG_CONCURRENCY, "Thread %d doing 3pt\n", worker_id);
		status = client->openThirdParty(
			(nestProtocolType) client->getRequest()->getRemoteProtocol() );
            break;

	case PROXY_GET:
	case PROXY_PUT:
		// nothing to do
		nest_debug( NEST_DEBUG_PROXY, "Thread %d starting proxy op.\n",
                    worker_id );
		break;

	case COPY_REQUEST:
		// nothing to do
		break; 

	default:
		assert( 0 );
		break;

    }

    if ( status != NEST_SUCCESS ) {
        client->getRequest()->setStatus( status );
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        return status;
    }

        // now go ahead and do the actual transfer 
    nest_debug( NEST_DEBUG_CONCURRENCY, "Thread %d enter blockingTask\n", worker_id );
    status = client->getTransfer()->BlockingTransfer( );
    nest_debug( NEST_DEBUG_CONCURRENCY, "Thread %d Task complete: %s\n", 
                worker_id, NestErrorString( status ) );

        // clean up after a 3pt
    if (   client->getRequest()->getType() == THIRD_PARTY_PUT
        || client->getRequest()->getType() == THIRD_PARTY_GET ) 
    {
        client->endThirdPartyTransfer( );
    }
        
    if ( status != NEST_SUCCESS ) {
        if ( status != NEST_LOT_NOT_ENOUGH ) {
            if ( errno != EPIPE ) {
                nest_warn( __FILE__, __FUNCTION__,
						   "BlockingTask failure: %s.\n",
                        NestErrorString( status ) );
            }
            client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        }
    } else {
            // now end the transfer
        status =
			( client->sendFinalAck() ? NEST_SUCCESS : NEST_NO_CONNECTION );
    }

    client->getRequest()->setStatus( status );
    return status;
}


// this function is called when worker children die.
// they shouldn't die but they might 
// I used to replace the dead workers within this function but
// this is a bad idea because malloc is not re-entrant
// so now just set a flag and clean up later 
// but it turns out that when workers die there are still problems
// workers shouldn't die however . . . 
void
nestIoManager::sigChild( int signal, siginfo_t *info, void *context )
{

}

// this is the function where each worker thread lives
void *
workerLoop( void *void_args )
{

    struct ThreadData *data = ( struct ThreadData * ) void_args;
    NestReplyStatus status = NEST_SUCCESS;

    /*
    nest_debug( NEST_DEBUG_CONCURRENCY | NEST_DEBUG_ADMIN, 
            "Thread %d (pid %d) reporting\n", data->thread_id, NestGetpid() );
    */

        // get a request from the queue, deal with it, repeat
    nestClient *currentClient;
    while( true ) {
        currentClient = io_manager->getFrontTask( );
        assert( currentClient != NULL );

        const char *request_type = ( currentClient->getRequest() == NULL )
                                 ? "special task"
                                 : currentClient->getRequest()->toString();

        nest_debug( NEST_DEBUG_CONCURRENCY, 
                "Thread %d handling request %s from client %s (%d)\n", 
                data->thread_id, 
                request_type, 
                currentClient->getUserIdentity(),
                currentClient->getId() );
        status = io_manager->workerRequestHandler(
			data->thread_id, currentClient );
        nest_debug( NEST_DEBUG_CONCURRENCY, 
                "Thread %d handled request %s from client %s (%d): %s\n", 
                data->thread_id, 
                request_type,
                currentClient->getUserIdentity(),
                currentClient->getId(), 
                NestErrorString( status ) );

            // return responsibility to the server and signal it, unlock mutex
		io_manager->addDoneTask( currentClient );
        io_manager->signalNest( data->thread_id );
    }

}
