/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/
#include "nest_proxy.h"
#include "nest_io_manager.h"		// probably should be a nest_server.h
#include "nest_storage_manager.h"	// that contains all server stuff
#include "nest_request_handler.h"
#include "nest_config.h"
#include "nest_speak.h"
#include "nest_main.h"

#define MUST_BE_PROXY(name)	\
    if ( ! serverIsProxy() ) {\
         nest_warn( __FILE__, __FUNCTION__, \
                    "%s requested a proxy_operation, "\
                    "but I'm not a proxy\n",\
                    (name) ); \
         return BAD_REQUEST;\
    }


/*
   *
   *  This is where all the code responsible for managing 
   *  an individual proxy node goes.
   *
   *  Code for handling proxy connections at a home server
   *  goes in nest_server.C.
   *
*/

enum { PROXY_CONNECTION, COORD_CONNECTION, HOME_CONNECTION };
#if 0
static char *server_types[] = { "proxy", "coordinator", "home" };
#endif

typedef struct Proxy_Values {
    char *home;                 // who is the home server
    char *self;                 // who am I?

    nestClient *coordinator;    // connection to the proxy coordinator
    nestClient *home_con;       // connection to the home node
    NestReplyStatus coord_status;   // status of proxy coordinator

    nestClientPool *pending;    // clients waiting on already in progress xfers
    nestClientPool *waiting;    // clients waiting bec no avail connections

    nestProxyClientCache *connections;   // connections home & to other proxies

    DynamicLard *index;         // how to look up the other proxy servers
    struct nest_hash_table *coordinator_map;    // what the coordinator uses
                                                // to track proxies

    pthread_mutex_t mutex;      // for protecting internal state from
                                // asynchronous activity
    pthread_mutex_t auth_mutex;
};

// some parameters that affect how proxies behave

// does coordinator add self to index
static bool coordinator_also_serves = true;
// do proxies use each other
static bool cooperative_proxies     = true;


    // need to predeclare because registerCoordinatorDone might need to call
NestReplyStatus startRegisterCoordinator( );

    // collection of connections to the home node and other proxies
static Proxy_Values *proxys = NULL;

    // state from the main server 
static nestClientPool *client_pool = NULL;

    // don't think we need these locks anymore since we are more careful to do
    // all data manipulation in the dispatch thread and only allow the
    // IO threads to do very simple chores
void
lockAuthMutex( char *server )
{
    // seeing some weird failures from the gsi auth code.
	// maybe not threadsafe?
    nest_debug( NEST_DEBUG_AUTH, 
            "PID: %d Locking auth mutex to connect to %s\n", 
            NestGetpid(), server );
    nest_pthread_mutex_lock( &(proxys->auth_mutex) );
}

void
unlockAuthMutex( char *server, NestReplyStatus status )
{
    nest_pthread_mutex_unlock( &(proxys->auth_mutex) );
    nest_debug( NEST_DEBUG_AUTH, 
            "PID: %d Unlocking auth mutex after connecting to %s: %s\n", 
            NestGetpid(), server, NestErrorString(status) );
}

void
lockMutex()
{
    nest_pthread_mutex_lock( &(proxys->mutex) );
}

void
unlockMutex()
{
    nest_pthread_mutex_unlock( &(proxys->mutex) );

}

bool
checkTwoStrings( void *o, void *t )
{
    if ( o == NULL || t == NULL )
		return ( o == NULL && t == NULL );
    int stat = strcmp( (char*)o, (char*)t );
    return ( stat == 0 );
}

    // returns whether a nest node is running as a proxy server or as 
    // a "home" node
bool
serverIsProxy( )
{
    return ( proxys != NULL );
}

bool
serverIsProxyCoordinator( )
{
    return ( proxys->coord_status == NEST_SUCCESS 
             && proxys->coordinator == NULL );
}

bool
proxyIsDisconnected( )
{
    return ( proxys->home_con == NULL );
}

int
getProxyCount( )
{
    CHECK_PID;
    return proxys->connections->size();
}

// this add client socket to fdset but not to dispatch client pool
// (probably because already in there)
void
makeDispatcherListen( nestClient *client )
{
    CHECK_PID;
    if ( client->hasSocket() ) {
        int socket = client->getConnection()->getSocket();
		NestFdRegister( socket, FD_READ );
    }
}

// this adds client socket to fdset and to dispatch client pool
void
makeDispatcherResponsible( nestClient *client)
{
    CHECK_PID;
    makeDispatcherListen( client );
    client_pool->addClient( client );
}

void
stopDispatcherListen( nestClient *client )
{
    CHECK_PID;
    if ( client->getConnection()->hasSocket() ) {
		NestFdUnregister( client->getConnection()->getSocket(), FD_READ );
    }    
}

struct ProxyListArgs
{
    bool first;
    char *coord;
};

    // a proxy registers with a home nest and receives a list of all of the
    // other proxies already registered
void
receiveProxyList( char *proxy, void *arg )
{
    struct ProxyListArgs *args = ( struct ProxyListArgs *) arg;
    nest_debug( NEST_DEBUG_PROXY, "%s is a registered proxy\n", proxy );
    if ( args->first ) {
        args->coord = NestStrdup( proxy );
    }
    args->first = false;
}

void
writeProxyAd( nestProtocol *ad )
{
    CHECK_PID;
    if ( ! serverIsProxy() ) return;
    char buffer[NEST_MESSAGESIZE];
    lockMutex();
    int map_size = -1;
    int map_servers = -1;
    if ( proxys->index != NULL ) {
        map_servers = proxys->index->numServers();
        map_size = proxys->index->getSize();
    }
    snprintf( buffer, NEST_MESSAGESIZE,
            "Home_server = \"%s\"\n"
            "Disconnected = %d\n"
            "Proxy_coordinator = \"%s\"\n"
            "Coordinator_status = \"%s\"\n"
            "Number_of_proxy_connections = %d\n"
            "Pending_requests_current = %d\n"
            "Pending_requests_ave = %.2f\n"
            "Pending_requests_max = %d\n"
            "Hash_map_size    = %4d\n"
            "Hash_num_servers = %4d\n",
            proxys->home,
            proxyIsDisconnected(),
            ( proxys->coordinator == NULL ) ? 
                "self" : proxys->coordinator->getHostIdentity(),
            NestErrorString( proxys->coord_status ),
            getProxyCount(),
            proxys->pending->size() + proxys->waiting->size(),
            proxys->pending->getAveLen() + proxys->waiting->getAveLen(),
            proxys->pending->getMaxLen() + proxys->waiting->getMaxLen(),
            map_size,
            map_servers
            );
    unlockMutex();
    ad->Sendn( buffer, NestStrlen( buffer ) );
}


char *
serverToString( void *server )
{
    return (char *)server;
}

void *
stringToServer( char *string, void *args )
{
    CHECK_PID;
        // TODO:  This will leak, actually if we always call clear first,
        // it might not...
    assert( args == NULL );
    return (void *)NestStrdup(string);
}


// this function flushes a proxy cache.  probably just be used for 
// repeatable measurements against a cold cache
NestReplyStatus 
FlushProxyCache( nestClient *client )
{
    CHECK_PID;
    MUST_BE_PROXY( client->getHostIp() );
    nest_debug( NEST_DEBUG_CLIENT, 
            "Flushing proxy cache at request of %s (%d) from %s.\n",
            client->getHostIdentity(), client->getId(), client->getHostIp() );
    NestReplyStatus s=storage_manager->formatDisk( 
            client->getUser(), NULL );
    nest_debug( NEST_DEBUG_CLIENT, 
            "Flushed proxy cache at request of %s (%d) from %s: %s.\n",
            client->getHostIdentity(), client->getId(), client->getHostIp(),
            NestErrorString(s) );
    return s;
}

const char *
findResponsibleServer(  const char *filepath, 
                        bool was_forwarded, 
                        const char *requestor ) 
{
    CHECK_PID;

    const char *server = NULL;

    lockMutex();
    if ( proxys->index != NULL ) {
        server = (const char *)proxys->index->getServer( filepath );
    } else {
        server = NULL;
    }
    unlockMutex();

    if ( server == NULL ) {
        nest_warn( __FILE__, __FUNCTION__, 
                "Unable to find a server in my index.  Will go home\n" );
        server = proxys->home;
    } else if ( ! cooperative_proxies ) {
        nest_debug( NEST_DEBUG_PROXY, 
                "%s is responsible proxy for %s, but not running coop mode,"
                " so will fetch from home node\n", server, filepath );
        server = proxys->home;
    } else if ( ! strcmp( server, proxys->self ) ) {
        nest_debug( NEST_DEBUG_PROXY,
                "I'm the responsible proxy for %s, will fetch from home\n",
                filepath );
        server = proxys->home;
    } else if ( was_forwarded ) {
        nest_warn( __FILE__, __FUNCTION__, 
				   "%s is responsible proxy for %s, but %s thinks I am, "
				   "so I'll fetch anyway to avoid loops.\n",
				   (char*)server, filepath, requestor );
        server = proxys->home;
    } else {
        nest_debug( NEST_DEBUG_PROXY, 
                "%s is responsible proxy for %s, fetching accordingly\n" ,
                (char*)server, filepath );
    }
    return server;
}

    // this is the function that receives a new mapping index from the coord
NestReplyStatus
ReceiveProxyIndex( nestClient *client )
{
    MUST_BE_PROXY( client->getHostIp() );
    CHECK_PID;

    if ( client != proxys->coordinator ) {
        nest_warn( __FILE__, __FUNCTION__,
				   "%s is not coordinator, but is sending proxy index.\n",
                client->getHostIp() );
    }

    nest_debug( NEST_DEBUG_PROXY, "About to receive proxy index from %s\n",
            client->getHostIdentity() );

    lockMutex();

        // maybe clear is dangerous.... what if instead of clear, we free
        // and make a new one?
    delete proxys->index;
    proxys->index = new DynamicLard( NULL );
    //proxys->index->clear( true );
    NestReplyStatus status = NestReceiveIndex( 
                             client->getConnection()->getSocket(),
                             proxys->index,
                             stringToServer,
                             NULL );
    unlockMutex();

    nest_debug( NEST_DEBUG_PROXY, "Received proxy index from %s: %s\n",
            client->getHostIdentity(), NestErrorString( status ) );

    return status;

}

// TODO:  This is a hack bec the naming got goofy in the new parrot
// parrot always passes absolute names and we assume relative paths
int
skipStorageDir( const char *path )
{
    char *storagedir = storage_manager->getInitialClientDir();
    assert( storagedir != NULL );
    int i = 0;
    while( ! strncmp( storagedir, &path[i], strlen(storagedir) ) ) {
        i += strlen(storagedir);
    }
    return i;
}

// first check to see if we have the whole file locally.
// otherwise query the home connection for it.
NestReplyStatus  
proxyFilelength( nestFileSize *filesize,
				 const char *filename,
				 const nestUser *user,
				 const char *cwd ) 
{
    CHECK_PID;

    NestReplyStatus status = NEST_NOT_YET_INITIALIZED;
    *filesize = 0;

        // we have a complete copy local
    if ( ( storage_manager->filelength( filesize, filename, user, cwd )) 
		 == NEST_SUCCESS )
    {
        nest_debug( NEST_DEBUG_PROXY, 
                "Request for filesize of %s handled locally: %ld\n",
                filename, *filesize );
        return NEST_SUCCESS;
    }

        // TODO
        // first to see if we've already cached the size for this file.

        // need to query the home connection for it.
    //if ( proxyIsDisconnected() ) 
    if ( proxys->home_con == NULL ) return NEST_TEMPORARILY_UNAVAILABLE;

        // otherwise, go ahead and use the proxy connection to the home
        // node.  This is safe because this is done synchronously.
        // ignore any errors also communicating to the server.  The
        // dispatcher will notice these on the next select and inform us.
    assert( proxys->home_con != NULL );
    assert( proxys->home_con->hasSocket() );
    NestConnection home = proxys->home_con->getConnection()->getSocket();

       // get the real absolute path, then remove our storage dir
    string path;
    status = storage_manager->expand_path( path,
										   user,
										   cwd,
										   filename );
    const char *queryfile = path.c_str();
    int position = skipStorageDir( queryfile );
    queryfile = &queryfile[position];
    status = NestFilesize( *filesize, (char*)queryfile, home );   
    nest_debug( NEST_DEBUG_PROXY, 
            "Request for filesize of %s handled by querying home: %ld: %s\n",
            queryfile, *filesize, NestErrorString( status ) );
    return status;
}

// first check to see if we have the whole file locally.
// otherwise query the home connection for it.
NestReplyStatus  
proxyFileinfo(  NestFileInfo *info,
				const char *filename,
				const nestUser *user,
				const char *cwd ) 
{
    CHECK_PID;

    NestReplyStatus status = NEST_NOT_YET_INITIALIZED;
    info->name[0] = '\0';
    info->next = NULL;

        // we have a complete copy local
    if ( ( storage_manager->fileinfo( info,
									  filename,
									  user,
									  cwd )) 
                == NEST_SUCCESS )
    {
        nest_debug( NEST_DEBUG_PROXY, 
                "Request for fileinfo of %s handled locally\n", filename );
        return NEST_SUCCESS;
    }

        // TODO
        // first to see if we've already cached the size for this file.

        // need to query the home connection for it.
    //if ( proxyIsDisconnected() ) 
    if ( proxys->home_con == NULL ) return NEST_TEMPORARILY_UNAVAILABLE;

        // otherwise, go ahead and use the proxy connection to the home
        // node.  This is safe because this is done synchronously.
        // ignore any errors also communicating to the server.  The
        // dispatcher will notice these on the next select and inform us.
    assert( proxys->home_con != NULL );
    assert( proxys->home_con->hasSocket() );
    NestConnection home = proxys->home_con->getConnection()->getSocket();

       // get the real absolute path, then remove our storage dir
    string path;
    status = storage_manager->expand_path( path,
										   user,
										   cwd,
										   filename );
    const char *queryfile = path.c_str();
    int position = skipStorageDir( queryfile );
    queryfile = &queryfile[position];
    status = NestFileinfo( home, (char*)queryfile, info );   
    nest_debug( NEST_DEBUG_PROXY, 
            "Request for fileinfo of %s handled by querying home: %s\n",
            queryfile, NestErrorString( status ) );
    return status;
}


// returns the port that proxies communicate over
int
getProxyPort()
{
    const char *port_string;
    if ( GetConfigValue(Config_ProxyPort, &port_string) ) {
        return( atoi(port_string) );
    } else {
        nest_die( "To run a proxy, must specify proxy_port in config file\n" );
        return -1;
    }

}


#if 0
// create a connection to the home site, a coordinator or another proxy
NestReplyStatus
openProxyConnection( char *server, nestClient **connection, int type )
{

        // open socket
    int socket;
    NestReplyStatus status;
    lockAuthMutex( server );
    if ( type == PROXY_CONNECTION || type == COORD_CONNECTION ) {
        bool anonymous = false;
        bool user2user = true;
        status = myNestOpenConnection( 
            socket, server, anonymous, user2user, getProxyPort() );
    } else {
        status = NestOpenConnection( socket, server );
    }
    unlockAuthMutex( server, status );

        // make the connection
    if ( status == NEST_SUCCESS ) {
        if ( type == PROXY_CONNECTION || type == COORD_CONNECTION ) {
            *connection = new nestClient(
				storage_manager->getInitialClientDir(),
				(nestProtocol *) new nestProtocolProxy( socket ),
				NULL ); 
        } else {
            *connection = new nestClient(
				storage_manager->getInitialClientDir(),
				(nestProtocol *) new nestProtocolNest( socket ),
				NULL ); 
        }
        (*connection)->setName( server );
        (*connection)->setHostIp(   server );
    }
	int id = ( (status == NEST_SUCCESS) ? (*connection)->getId() : -1 );
    nest_debug( NEST_DEBUG_PROXY, 
            "PID: %d Made new %s connection to %s (%d): %s\n", 
            NestGetpid(), server_types[type], server, 
			id, NestErrorString(status) );
    return status;
}

NestReplyStatus
openHomeConnection( char *location )
{
    nestClient *connection;
    NestReplyStatus status = openProxyConnection( proxys->home, &connection, 
            HOME_CONNECTION );
    NEST_REPLY_CHECK(status);
    proxys->home_con = connection;
    return status;
}

// create a connection to a cooperating proxy
// TODO: Can this be done synchronously?  
// If so, all that needs to be done is make addProxyTransfer be asynch
NestReplyStatus
openCoordinatorConnection( char *location )
{
    nestClient *coord;
    nest_debug( NEST_DEBUG_PROXY, "Trying to open connection to proxy "
            "coordinator at %s.\n", location );
    NestReplyStatus status = openProxyConnection( location, &coord, 
            COORD_CONNECTION );
    NEST_REPLY_CHECK( status );

        // TODO:  we should register with coordinator to make sure
        // that coordinator is correct coordinator
    status = registerProxyConnection( coord );
    if ( status == NEST_SUCCESS ) {
            // OK, this is tricky.  The coordinator will push indices
            // so we need to listen to him so he must be added as a client
        proxys->coordinator = coord;
        client_pool->addClient( proxys->coordinator );
        makeDispatcherListen( proxys->coordinator );
    } else {
        nest_warn( __FILE__, __FUNCTION__,
				   "Couldn't register with coordinator %s: %s\n",
				   location, NestErrorString( status ) );
    }

    return status;
}
#endif

nestClient *
removeWaitingClient( nestClientPool *queue, const char *file )
{
    CHECK_PID;
    CLIENT_ITERATOR itr = queue->begin(); 
    while( itr != queue->end() ) {
        nestClient *queued = *itr;
        if ( strcmp( queued->getProxyRequest()->getNestFile(), file ) == 0 ) {
            nest_debug( NEST_DEBUG_PROXY, 
                    "Found client %s (%s) waiting for %s.\n",
                    queued->getHostIdentity(), queued->getHostIp(), file );
            itr = queue->erase( itr );
            return queued;
        } else {
            itr++;
        }
    }
    return NULL;
}

long proxyTimeout = 5000000;
void
proxyMaintenance( nestClientPool *queue )
{
    CHECK_PID;
    int count = 0;
    lockMutex();
    CLIENT_ITERATOR itr = queue->begin(); 
    while( itr != queue->end() ) {
        count++;
        nestClient *queued = *itr;
        *itr++;
        nestClient *proxy = queued->getProxyClient();
        const char *file = queued->getProxyRequest()->getNestFile(); 
        long millis = queued->getRequest()->getAge();
        const char *proxyname = ( ( proxy ) 
                ? proxy->getHostIdentity() : "not yet assigned" );
        if ( millis > proxyTimeout ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "Client %s from %s (id %d) has been waiting "
                    "for %s from %s for %.3f seconds.\n",
                    queued->getHostIdentity(),
                    queued->getHostIp(),
                    queued->getId(),
                    file,
                    proxyname,
                    (float)millis/MILLION );
        }
    }
    unlockMutex();
    if ( count > 0 ) {
        nest_debug( NEST_DEBUG_PROXY, 
                "%d clients still waiting on transfers.\n",
            count );
    }
}

void
proxyMaintenance()
{
    CHECK_PID;
    if ( ! serverIsProxy() ) return;
    proxyMaintenance( proxys->pending );
    proxyMaintenance( proxys->waiting );
}


// start a transfer (or send error) for everyone who was waiting for a file
int
checkProxyQueue( nestClientPool *queue,
				 const char *file,
				 NestReplyStatus stat)
{
    CHECK_PID;
    int started = 0;
    nestClient *queued;
    while( ( queued = removeWaitingClient( queue, file ) ) != NULL ) {
        if ( stat == NEST_SUCCESS ) {
            req_handler->TransferFile( queued, queued->getRequest(), true );
        } else {
            if ( queued->getStorageConnection() != NULL ) {
                storage_manager->close( queued->getStorageConnection() );
            }
            queued->sendErrorMessage( __FILE__, __FUNCTION__,  stat );
            if ( queued->getRequest() != NULL ) {
                queued->endRequest( stat );
            }
            makeDispatcherListen( queued );
        }
        started++;
    }
    return started;
}
    
// currently this is called from within the synch operation so the
// responibility to reply to the client becomes the proxy's.  Once all
// dirty writes are flushed, send the reply.
NestReplyStatus
flushDirtyWrites( nestClient *client )
{
    CHECK_PID;
    NestReplyStatus status = NEST_SUCCESS;

    nestRequest *dirty_write = client->getDirtyWrite( );
    if ( dirty_write ) {
        client->setFlushing( true );
        nest_debug( NEST_DEBUG_PROXY, "Getting set to flush a dirty write "
                "of %s for %s\n",
                dirty_write->getNestFile(),
                client->getHostIdentity() );
            // addProxyTransfer will add client to the pending queue
            // assigning it the proxy request will allow us to find this
            // client again once the proxy transfer completes
        client->assignProxyRequest( dirty_write );
        status = addProxyTransfer( client );    
        if ( status == NEST_SUCCESS ) {
            return NEST_SUCCESS;
        }
    } else {
        nest_debug( NEST_DEBUG_PROXY, 
                "Done flushing dirty writes for %s\n",
                client->getHostIdentity() );
            // the error checking is done in checkProxyTransfers
    }

        // if we make it here, either we had a flush error or there are no
        // more dirty writes
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck( );
    }
    client->endRequest( status );
    makeDispatcherListen( client );
    return status;
}

// convenience call which creates a shadow name for incoming files to 
// provide some degree of atomicity and consistency protection
char *
getShadowName( const char *file )
{
    CHECK_PID;
    char *shadow = (char *)NestCalloc( NestStrlen( file ) + 10 );
    NestSnprintf( shadow, NestStrlen( file ) + 10, "%s.proxy", file );
    return shadow;
}


bool
alreadyPendingProxyTransfer( nestClient *client, const char *file )
{
    CHECK_PID;
    for(    CLIENT_ITERATOR itr=proxys->pending->begin();
            itr!=proxys->pending->end(); 
            *itr++)
    {
        nestClient *queued = *itr;
        if ( strcmp( queued->getProxyRequest()->getNestFile(), file ) == 0 ) {
            return true;
        }
    }
    return false;
}

// for caching reads, create a shadow proxyreq
// this is because proxies caches blocks of data and we need to know
// which block to fetch as well as to remember the original parameters
// we actually also create a third request object which is what is 
// assigned to the proxy connection for the purpose of fetching the block 
// (safely synchronous - no network activity)
NestReplyStatus 
proxyTranslateRequest( nestClient *client )
{
    CHECK_PID;

    nestRequest *req = client->getRequest();
        
        // just accept writes as is, do we need a notion of sparse files?
    if ( req->isWrite() ) {
            // TODO
        // client->addDirty(); // actually do this when transfer completes
        nest_debug( NEST_DEBUG_PROXY, 
            "Client %s is writing %s %ld-%ld\n",
            client->getHostIdentity(),
            req->getNestFile(), req->getStart(), req->getEnd() );
        return NEST_SUCCESS;
    }

        // for third party puts and gets, don't do proxy stuff
    if ( req->isThirdParty() ) {
        return NEST_SUCCESS;
    }
        
    // for reads there are a couple of possibilities
    // 1) we have the whole file because it has been written here
    // 2) we are already fetching the needed block
    // 3) we already have the needed block
    // 4) we need to fetch the needed block from the home server
    // 5) we need to fetch the needed block from another proxy

    // 1) is easy and we just check for that here and leave without doing
    // anything it that is the case
    // 2-5) more complicated.  Make a proxy req asking for the chunk.
    // 2) if the chunk exists, it will be served normally.
    // 3-5) if the chunk doesn't exist, main server will call addProxyTransfer 

    // 1) we have the whole file because it has been written here
    nestFileSize filelength;
    if ( ( storage_manager->filelength( &filelength,
										req->getNestFile(),
										client->getUser(),
										client->getCwd() )) 
                == NEST_SUCCESS )
    {
        nest_debug( NEST_DEBUG_PROXY, 
                "Can satisfy req for %s with local whole file copy\n",
                req->getNestFile() );
        return NEST_SUCCESS;
    }
        
        // proxys don't keep whole input files, they only keep chunks
        // translate this file request into a chunk request

        // find the closest start of a chunk which is less than or equal
        // to this req start
    long chunk = 0;
    while( chunk+PROXY_BLOCK_SIZE <=req->getStart() ) {
		chunk += PROXY_BLOCK_SIZE;
	}
    char *newname = (char*)NestMalloc(NEST_MAXPATHLEN);
    NestSnprintf( newname, NEST_MAXPATHLEN,
				  "%s.%ld", req->getNestFile(),chunk);
    long newstart = req->getStart() - chunk;
    long newend   = req->getEnd() - chunk;

        // create a proxy req which is used to request the chunk from
        // either the home node or another proxy
    nest_debug( NEST_DEBUG_PROXY, 
				"Translating req for %s %ld-%ld to %s %ld-%ld\n",
				req->getNestFile(), req->getStart(), req->getEnd(),
				newname, newstart, newend );
    nestRequest *proxyreq = new nestRequest( req->getType() );
	proxyreq->setFile( newname );
	proxyreq->setStart( newstart );
	proxyreq->setEnd( newend );
    req->setChunkSize( chunk );
    proxyreq->setChunkSize( chunk );
    NestFree( newname );
    client->assignProxyRequest( proxyreq );
    

        // check whether the client has issued a req that will go beyond
        // the end of the current chunk.  If so, this could cause problems
        // for clients doing whole file gets.
    if (        proxyreq->getEnd() == ENTIRE_FILE 
            ||  proxyreq->getEnd() > chunk + PROXY_BLOCK_SIZE ) 
    {
        nest_warn( __FILE__, __FUNCTION__,
                "%s:%s client read will be truncated. Hope they try again\n",
                __FILE__, __FUNCTION__ );
    }
    return NEST_SUCCESS;
}


// this function can be used both to register with the proxy coordinator and
// the home server
NestReplyStatus
registerProxyConnection( nestClient *connection )
{

        // register as a proxy
    NestReplyStatus status = NestProxyRegister( 
                                    proxys->self,
                                    connection->getConnection()->getSocket() );
        
    nest_debug( NEST_DEBUG_PROXY, "Registered as proxy to %s: %s.\n",
            connection->getHostIdentity(), NestErrorString( status ) );
    if ( status != NEST_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__,
				   "Unable to register as proxy to %s: %s\n",
				   connection->getHostIdentity(), NestErrorString( status ) );
    }
    return status;
}


    // assign a thread the responsibility of opening a connection to home
    // server and registering there as a proxy
    // this one is easier than other asynchronous work because we can just
    // loop forever in this thread since the home node will never change
    // locations.  Later if it can, we need to change this.
NestReplyStatus
asynchronousRegisterHome( nestClient *home )
{
    NestReplyStatus status = NEST_NOT_YET_INITIALIZED;
    int socket;
    int attempts = 0;
    int sleep_max = 15;

    while( status != NEST_SUCCESS ) {
        if ( attempts++ > 0 ) {
            int sleep_time = ( attempts < sleep_max ) ? attempts : sleep_max;
            nest_warn( __FILE__, __FUNCTION__,
					   "Unable to register with home node %s: %s.  "
					   "Will sleep %d secs and try again.\n",
					   home->getHostName(),
					   NestErrorString( status ), sleep_time ); 
            sleep( sleep_time );
        }

            // first try to just open a socket
        status = NestOpenConnection( socket, (char *)home->getHostIdentity() );
        if ( status == NEST_SUCCESS ) {
            nestProtocolNest *npn = new nestProtocolNest( socket );
            home->assignConnection( npn );

                // then try to actually send a register command
            status = registerProxyConnection( home ); 
            if ( status != NEST_SUCCESS ) {
                home->assignConnection( NULL );
                delete npn;
            }
        }
    }

    return NEST_SUCCESS;
}

NestReplyStatus
registerHomeDone( nestClient *home, NestReplyStatus status )
{
    CHECK_PID;
    assert( status == NEST_SUCCESS );
    nest_debug( NEST_DEBUG_PROXY, 
				"Successfully registered with home %s.\n",
				home->getHostIdentity() );
    proxys->home_con = home;
    proxys->home_con->assignTaskFunction( NULL, NULL );
    makeDispatcherResponsible(   proxys->home_con );

        // special case start up.  Start the search for proxy coord now.
    if ( proxys->coord_status == NEST_NOT_YET_INITIALIZED ) {
        startRegisterCoordinator();
    }
    return NEST_SUCCESS;
}

NestReplyStatus
startRegisterHome( )
{
    CHECK_PID;
    nestClient *home = new nestClient(
		storage_manager->getInitialClientDir(),
		NULL,
		NULL );
    home->setHostName( proxys->home );
    home->setHostIdentity( proxys->home );
    home->assignTaskFunction(  asynchronousRegisterHome, 
                               registerHomeDone );
    return io_manager->addTask( home );
}

NestReplyStatus
registerCoordinatorDone( nestClient *coordinator, NestReplyStatus status )
{
    CHECK_PID;

        // always clear the task function pointers
    coordinator->assignTaskFunction( NULL, NULL );

    if ( status != NEST_SUCCESS ) {
            // need to try again
        delete coordinator;
        return startRegisterCoordinator();
    }

    nest_debug( NEST_DEBUG_PROXY, 
        "Successfully registered with coordinator %s.\n", 
        coordinator->getHostIdentity() );
    if ( ! strcmp( coordinator->getHostIdentity(), proxys->self ) ) {

        nest_debug( NEST_DEBUG_PROXY, "Hey, I'm the proxy coordinator!\n" );

        assert( proxys->coordinator == NULL );
        assert( proxys->coordinator_map == NULL );
        assert( proxys->index != NULL );

        proxys->coordinator_map = nest_hash_table_create( 1037, NULL );

            // this actually doesn't seem to be the problem....
            // maybe clear is dangerous.... what if instead of clear, 
            // we free  and make a new one?
        delete proxys->index;
        proxys->index = new DynamicLard( serverToString );
        proxys->index->setServerEqualityFunction( checkTwoStrings );
            /*
            proxys->index->clear( true );   // do we want to clear?
                                            // or keep using same index?
                                            // I think we have to clear...
            */

            // only coordinator needs to bother with maintaining the
            // message string.
        proxys->index->setServerToString( serverToString );
        proxys->index->setName( proxys->self );
        if ( coordinator_also_serves ) {
            proxys->index->addServer( (void*)NestStrdup(proxys->self) );
        }
            
    } else {

            // someone else is the coordinator
        proxys->coordinator = coordinator;
            
            // what if we only make the dispatcher listen, but not be
            // responsible?
        makeDispatcherResponsible( proxys->coordinator );
    }


    proxys->coord_status = status;
    return status;
}

NestReplyStatus
asynchronousRegisterCoordinator( nestClient *coordinator )
{

    NestReplyStatus status;

    struct ProxyListArgs args;
    args.first  = true; 
    args.coord  = NULL;

    
    /*
        // this is maybe a little goofy, but we stashed a duped
        // connection to the home node in this client
        // WATCH OUT: We assume it is a nestProtocolNest connection
    nestProtocolNest *home = (nestProtocolNest*)coordinator->getConnection();
    coordinator->assignConnection( NULL );
    */

        // actually all we have is the home_server name.  try to open 
        // connection so we can fetch proxy list.
        // it may be goofy, but we stashed the home name in the client.
    NestConnection home_con = -1;
    status = NestOpenConnection( home_con, proxys->home );
    if ( status != NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_PROXY, 
                "Unable to open home con, so can't discover coord: %s\n",
                NestErrorString( status ) );
    }
    
    if ( status == NEST_SUCCESS ) {

            // ask home who the coordinator is
        status = NestProxyQuery( home_con, receiveProxyList, &args  );
        if ( status != NEST_SUCCESS ) goto DONE;
        if ( args.coord == NULL ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "received empty proxy list.  this shouldn't happen.\n" );
            goto DONE;
        }
        coordinator->setHostName( args.coord );
            // need to set identity also to be used to open connection
        coordinator->setHostIdentity( args.coord );

        nest_debug( NEST_DEBUG_PROXY, "Home says coordinator is %s.\n",
                coordinator->getHostIdentity() );

            // hey, it's me
        if ( ! strcmp( args.coord, proxys->self ) ) {
            goto DONE;
        }

            // open a connection to whoever it is and register with them
        int socket;
        bool anonymous = false;
        bool user2user = true;
        int port = getProxyPort();

            // TODO: This whole next part of code should be shared
            // with asynchronousRegisterHome
        status = myNestOpenConnection(  socket,
                                        (char*)coordinator->getHostIdentity(),
                                        anonymous,
                                        user2user,
                                        port );
        if ( status != NEST_SUCCESS ) goto DONE;


        nestProtocolNest *npn = new nestProtocolNest( socket );
        coordinator->assignConnection( npn );
        status = registerProxyConnection( coordinator );
        if ( status != NEST_SUCCESS ) {
            coordinator->assignConnection( NULL );
            delete npn;
        }
    }


    DONE:
    if ( status != NEST_SUCCESS ) {
        int sleep_time = 4;
        nest_debug( NEST_DEBUG_PROXY, 
                "Unable to register with coordinator: %s.  "
                "Will try again in %d secs.\n",
                NestErrorString( status ),
                sleep_time );
        sleep( sleep_time );
    }
    if ( args.coord != NULL )   NestFree( args.coord );

    return status;
}

NestReplyStatus
startRegisterCoordinator( )
{
    CHECK_PID;
    nestClient *coordinator = new nestClient(
		storage_manager->getInitialClientDir(), NULL, NULL );

    nest_debug( NEST_DEBUG_PROXY, 
            "Will try to discover / register with coordinator now.\n" );

        // two options.
        // 1) query whether we're disconnected.  if so, add this to some
        //    pending queue when we get connected again.
        // 2) give this a Dup of the current home_con which may be invalid
        //    if this fails to use that Dup, then it sleeps and trys again
        //    and eventually the home_con will be good and so will the Dup
        // 3) actually 2) is no good, bec the dispatcher is listening on 
        //    the home_con and when we ask it to send traffic via the query
        //    proxys request, dispatcher may think it is a ready client and
        //    try to receiveRequest from it.  So, we should just make a new
        //    connection to home_node whenever we need to discover coord

    /*
        // lets do 2)
    if ( ! proxyIsDisconnected() ) {
        assert( proxys->home_con->getConnection()->getType() == NEST_PROTOCOL );

        int socket = NestDup( proxys->home_con->getConnection()->getSocket() );
        nestProtocolNest *npn = new nestProtocolNest( socket );
        coordinator->assignConnection( npn );
    }
    */

    coordinator->assignTaskFunction( asynchronousRegisterCoordinator,
                                     registerCoordinatorDone );
    return io_manager->addTask( coordinator );
}

bool
connectionFailure( NestReplyStatus status )
{
    if ( status == NEST_NO_CONNECTION )          return true;
    if ( status == NEST_COULD_NOT_AUTHENTICATE ) return true;
    return false;
}

NestReplyStatus
asynchronousProxyFetch( nestClient *proxy )
{

    NestReplyStatus status      = NEST_SUCCESS;
    nestProxyFetch *fetch       = proxy->getProxyFetch();
    nestProtocol   *connection  = proxy->getConnection();
    nestFileSize    length      = fetch->getHomeLength(); 
    const char     *file        = fetch->getHomeFile(); 
    nestFileSize    offset      = fetch->getHomeOffset(); 
    const char     *server      = fetch->getServer();

        // possible the proxy already has a connection
        // otherwise try to open a connection here
    if ( connection == NULL ) {
        int socket;
        int port       = NEST_SERVER_PORT;
        bool anonymous = false;
        bool user2user = false;
        
        if ( ! strcmp( server, proxys->home ) ) {   // home connection
            port = NEST_SERVER_PORT;
            user2user = false;
        } else {                                    // proxy connection
            user2user = true;
            port      = getProxyPort();
        }
        status = myNestOpenConnection(socket,server,anonymous,user2user,port);
        if ( status == NEST_SUCCESS ) {
            proxy->setHostName( server );
            proxy->setHostIp( server );
            connection = (nestProtocol*) new nestProtocolNest( socket );
            proxy->assignConnection( connection );
        }
        nest_debug( NEST_DEBUG_PROXY, "Opened connection to %s: %s\n",
                    server, NestErrorString( status ) );
    }
    
        // ask for permission to read this block
    if ( status == NEST_SUCCESS ) {
        assert( connection != NULL );
        int position = skipStorageDir( file );
        status = RequestReadFile(&file[position], offset, length, 
                connection->getSocket());
        if ( status != NEST_SUCCESS ) {
            nest_debug( NEST_DEBUG_PROXY,
                    "Unable to request read of %s (%d) from %s: %s\n",
                    file, offset, server, NestErrorString( status ) );
        }
    }

        // go ahead and read the block
        // TODO: this is where it would be nice to use IO manager to
        // actually schedule this transfer and get the accounting as well.
    if ( status == NEST_SUCCESS ) {
        proxy->startTransfer( PROXY_GET, length );
        status = proxy->getTransfer()->BlockingTransfer( );
    }

        // if no connection, need to return to dispatcher to rediscover
        // who is responsible for this block before we try again.
    if ( connectionFailure( status ) ) {
        // TODO: probably should not sleep here.  Add another 
        // chore type to just sleep.  Have proxy translator determine.
        // sleep here and hope things improve.  eventually need to die.
        fetch->incrementAttempts();
        long sleep_secs = 2 * fetch->getAttempts(); 
        nest_debug( NEST_DEBUG_PROXY,
					"Unable to connect to %s for %s (%ld)."
					" Try again in %d secs.\n",
					server, file, offset, sleep_secs );
        sleep( sleep_secs ); 
        nest_debug( NEST_DEBUG_PROXY,
					"Slept after failure for %s (%ld) from %s."
					" Should try again.\n",
					file, offset, server );
    }

    return status;
}

// if connectionFailure, clean up and try again
// otherwise, go ahead and start any pending clients waiting on this
NestReplyStatus
proxyFetchDone( nestClient *proxy, NestReplyStatus status )
{
    CHECK_PID;
    nestProxyFetch *fetch = proxy->getProxyFetch();
    assert( fetch != NULL );

    nest_debug( NEST_DEBUG_PROXY, 
            "Proxy fetch of %s done: %s\n",
            fetch->getProxyFile(),
            NestErrorString( status ) );

        // clean up the storage stuff regardless of status
    nestProtocol *storage = proxy->getStorageConnection();
    proxy->assignStorageConnection( NULL );
    storage_manager->close( storage, proxy );
    delete storage;

        // either rename or delete the temporary incoming name
    if ( status != NEST_SUCCESS ) {
        storage_manager->deleteFile( fetch->getClientUser(),
									 fetch->getClientCwd(), 
									 fetch->getShadowFile() );
    } else {
        NestReplyStatus renameStatus = storage_manager->renameFile( 
			fetch->getClientUser(), 
			fetch->getShadowFile(),
			fetch->getProxyFile(),
			fetch->getClientCwd(),
			false );
        nest_debug( NEST_DEBUG_PROXY, "Renamed proxyfile %s to %s: %s.\n",
					fetch->getShadowFile(),
					fetch->getProxyFile(),
					NestErrorString( renameStatus ) ); 
        if ( renameStatus != NEST_SUCCESS ) {
            nest_die( "Failed to rename proxyfile: %s\n",
                    NestErrorString( renameStatus ) );
        }
    }

        // see if we need to try again
    if ( connectionFailure( status ) ) {
            // try again
        proxy->endTransfer( status );
        delete proxy;
        nest_debug( NEST_DEBUG_PROXY,
                "Proxy fetch of %s failed %d times.  Will try again.\n",
                fetch->getProxyFile(),
                fetch->getAttempts() );
        NestReplyStatus status = startProxyFetch( fetch );
        assert( status == NEST_SUCCESS );
    } else {
        // if we make it here, either success or some probably permanent error
            // check to see if any of the pending clients were waiting for this
            // there'd better be at least one!
            // pending must be checked before waiting
            //  this ensures at least minimal forward progress
        int started = 0;
        lockMutex();
        const char *file = fetch->getProxyFile();
        started += checkProxyQueue( proxys->pending, file, status);
        started += checkProxyQueue( proxys->waiting, file, status);
        unlockMutex();
        assert( started != 0 );

            // this proxy is in good shape, let's cache it
        proxy->assignTaskFunction( NULL, NULL );
        proxy->assignProxyFetch( NULL );
        assert( ! strcmp( proxy->getHostName(), fetch->getServer() ) );
        proxys->connections->insertProxyClient( 
			(char*)proxy->getHostName(), proxy ); 
        delete fetch;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
startProxyFetch( nestProxyFetch *fetch )
{
        
        // 1) we need to open a storage connection for this object
        // 2) we need to create a proxy client to do the transfer

    NestReplyStatus status   = NEST_SUCCESS; 
    nestProtocol    *storage = NULL;
    nestClient      *proxy   = NULL;
    const char      *server  = NULL;
    int link_count           = 0;

        // figure out who the server is.
    server = findResponsibleServer( fetch->getProxyFile(),
                                    fetch->wasForwarded(),
                                    fetch->getClientHostName() ); 
    fetch->setServer( server );

        // actually get the storage connection
    status = storage_manager->open(  &storage,
									 fetch->getClientUser(),
									 fetch->getShadowFile(),
									 WRITE_FLAGS,
									 0,
									 fetch->getHomeLength(),
									 fetch->getClientCwd(),
									 link_count );
    if ( status != NEST_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__, 
                "Could not open an fd for %s: %s.\n",
                fetch->getShadowFile(), NestErrorString( status ) );
        return status;
    }
    assert( status == NEST_SUCCESS );

        // try to find a cached connection to this server
    proxy = proxys->connections->getCachedProxy( server );
    if ( proxy == NULL ) {
        nest_debug( NEST_DEBUG_PROXY,
                "Unable to find cached proxy to %s, creating new one.\n",
                server );
        proxy = new nestClient(
			storage_manager->getInitialClientDir(), NULL, NULL );
    }

    proxy->assignProxyFetch( fetch );
    proxy->assignStorageConnection( storage );
    proxy->assignTaskFunction(  asynchronousProxyFetch, 
                                proxyFetchDone );

    status = io_manager->addTask( proxy );
    return status;
}

    
    // are we here because 
    // 1) the file is missing and we need to fetch it
    // 1a) either from the home node
    // 1b) or from another proxy
    // 2) are we here because the file is already currently being fetched?
    // 3) are we here to flush a dirty write back to the home server?
    // -- in any event, don't do anything here that could block.
    // especially nothing that could block on another proxy bec of deadlock
NestReplyStatus
addProxyTransfer( nestClient *client )
{

    NestReplyStatus  status      = NEST_SUCCESS;
    nestProxyFetch   *fetch      = NULL;
    nestRequest      *proxyreq   = client->getProxyRequest();
    assert( proxyreq != client->getRequest() );

        // don't know how to do flush writes anymore.
    if( proxyreq->isWrite() ) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

        // 2) are we here because the file is already currently being fetched?
    if ( alreadyPendingProxyTransfer( client, proxyreq->getNestFile() ) ) {
       nest_debug( NEST_DEBUG_PROXY, "Adding another pending req for %s\n",
                proxyreq->getNestFile() );
        status = NEST_SUCCESS;
    } else {
            // need to go out and get it ourselves 
        fetch = NULL;
        const char *homefile  = client->getRequest()->getNestFile();
        const char *proxyfile = client->getProxyRequest()->getNestFile();
        char *shadowname      = getShadowName( proxyfile );
        long home_off         = client->getRequest()->getChunkSize();
        long length           = PROXY_BLOCK_SIZE;
        bool forwarded_req    = (client->getConnection()->getType()
                                 ==PROXY_PROTOCOL);
        nest_debug( NEST_DEBUG_PROXY, "Adding proxy fetch for user '%s'\n", 
					client->getUserIdentity() );
        fetch = new nestProxyFetch( homefile, 
                                    proxyfile, 
                                    (const char *)shadowname, 
                                    home_off,
                                    length, 
                                    client,
                                    forwarded_req );
        NestFree( shadowname );
        status = startProxyFetch( fetch );
    }

        // make sure to accept responsibility for this client so main server
        // won't accept on it anymore.  dispatcher turns back on at taskDone.
        // also make sure to remember that we're waiting on this file.
    if ( status == NEST_SUCCESS ) {
        stopDispatcherListen( client );
        proxys->pending->addClient( client );
    } else {
        if ( fetch != NULL ) delete fetch;
    } 
    return status;

# if 0
        // otherwise find an available connection, (or make one)
        // and try to open a storage connection for the proxy file
        // and request permission from the home server to start proxy transfer
    status = openProxyStorageConnection( client, proxyreq, &storage );
    if ( status != NEST_SUCCESS ) {
        return status;
    } 
    client->assignStorageConnection( storage );

        // need to remember here that this client is waiting for this
        // so that only one thread will request it
        // this does need to be done synchronously
        // however, this means that for failed transfers, we need to remove
        // client from the pending list
    lockMutex();
    proxys->pending->addClient( client );
    unlockMutex();

        // this is the code that needs to be put into a thread
    nest_debug( NEST_DEBUG_PROXY, "Assigning asycnInitProxyTransfer func ptr "
            "to client %s (%d) to start transfer of %s.\n",
            client->getHostIdentity(), client->getId(), proxyreq->getNestFile() );
    client->assignTaskFunction( 
            asynchronouslyInitiateProxyTransfer,
            asynchronousInitCleanup );
    status = io_manager->addTask( client );
    return status;
# endif
}

// TODO: This can segfault when it trys to send it to a client that has
// already crashed and been removed....
int
sendProxyIndexToOne( nestClient *client )
{
    CHECK_PID;
    assert( proxys->index != NULL );
    NestReplyStatus status = client->sendProxyIndex( proxys->index );
    nest_debug( NEST_DEBUG_PROXY, "Pushed revised index to %s: %s\n",
            client->getHostIp(), NestErrorString( status ) ); 
    if ( status != NEST_SUCCESS ) {
            // ok to ignore this here?
            // probably.  assume that only reason for failure is that
            // no connection, in which case the client is lost and soon
            // the dispatcher will tell us about it.
        if ( ! connectionFailure( status ) ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "Coordinator not dealing with proxy failure: %s\n",
                    NestErrorString( status ) );
        }
    }
    return 1;
}

int
sendProxyIndex( const char *key, void *item, void *arg )
{
    assert( arg == NULL );
    nestClient *client = (nestClient *)item;
    return sendProxyIndexToOne( client );
}

// TODO:  Does coordinator activity need to be asynchronous?
// I think not bec as long as there is only one coordinator, only that one
// coordinator will ever do active initiations to the other proxies as long
// as the other proxies active initiations are done asynchronously
//
// OK.  Admittedly there can be deadlock if two different proxies ever think
// they are each the coordinator...can this happen?
NestReplyStatus
sendProxyIndexToAll( )
{
    nest_debug( NEST_DEBUG_PROXY, "Sending proxy index to all\n" );
    nest_hash_table_iterate( proxys->coordinator_map,
            sendProxyIndex, NULL );
    return NEST_SUCCESS;
}


// this is where we check to see if we've lost any proxy connections that
// are important to us.
NestReplyStatus
removeClientFromProxy( nestClient *client )
{
    CHECK_PID;

    bool isProxy = ( client->getConnection()->getType() == PROXY_PROTOCOL ); 

    if ( ! isProxy )  {
        // check see if this normal client was was waiting on anything
        // however do we want to remove from the pending and waiting queues?
        // TODO remove real clients from waiting and pending lists removeClient
        // actually, if they stay in there, eventually the proxy transfer
        // will complete and we'll discover they down when we try to send
        // them the proxy file
		lockMutex();
		bool uh_oh = (bool)
			(proxys->pending->contains( client ) 
			 + proxys->waiting->contains(client));
		if ( uh_oh ) {
			nest_warn( __FILE__, __FUNCTION__, 
					   "Need to remove client %s (%d) from %s waiting "
					   "on %s (%d) from pending list\n",
					   client->getHostIdentity(), client->getId(),
					   client->getHostIp(), 
					   client->getRequest()->getNestFile(),
					   client->getRequest()->getChunkSize() );
		}
		unlockMutex();
    }

    if ( client == proxys->home_con ) {
            // lost home connection
        nest_warn( __FILE__, __FUNCTION__, 
                "Oh shit!  Lost our home connection %s at %s. "
                    "Will reconnect\n",
                    client->getHostIdentity(), client->getHostIp() );
        proxys->home_con = NULL;
        startRegisterHome();

    } 

    if ( client == proxys->coordinator ) {
            // lost coordinator connection
        assert( proxys->coordinator != NULL );
        nest_warn( __FILE__, __FUNCTION__,
				   "Oh shit!  Lost our coordinator %s at %s.\n",
				   client->getHostIdentity(), client->getHostIp() );
        proxys->coordinator = NULL;
        proxys->coord_status = NEST_NO_CONNECTION; 
        startRegisterCoordinator();
    } 
    
    if ( serverIsProxyCoordinator() && isProxy ) {
            // if we make it here, this proxy is the current coordinator
            // and this client is another proxy
        assert( proxys->coordinator_map != NULL );
        lockMutex();
        assert( proxys->index != NULL );
        nest_debug( NEST_DEBUG_PROXY, "Removing server %s from proxy map.\n",
                client->getHostIp() );
        proxys->index->removeServer( (void *)client->getHostIp(), true );
        nest_hash_table_remove( 
                proxys->coordinator_map, client->getHostIp() );
        sendProxyIndexToAll();
        unlockMutex();
    }

    return NEST_SUCCESS;
}


// proxies register themselves with who they think is the coordinator
NestReplyStatus
proxyRegisterProxyClient( nestClient *client, char *proxy )
{
    CHECK_PID;

    if ( ! serverIsProxyCoordinator() ) {
        nest_warn( __FILE__, __FUNCTION__,
				   "%s (%s) tried to register as proxy, but I'm not coord.\n",
				   client->getHostIdentity(), client->getHostIp() );
        client->sendErrorMessage( __FILE__, __FUNCTION__,
								  NEST_NOT_YET_INITIALIZED ); 
        return NEST_SUCCESS;
    } else {
            // make sure to send the ack before sending the index
        client->sendAck();
    }

        // if we make it here, this proxy is the current coordinator
        // and this client is another proxy
    //client->setName( client->getHostIp() );
    lockMutex();
    assert( proxys->coordinator_map != NULL );
    if (!nest_hash_table_lookup( proxys->coordinator_map,
								 client->getHostIp())){
        nest_debug( NEST_DEBUG_PROXY, 
                "New proxy (%s) from client %s (%d) registering\n",
                proxy, client->getHostIp(), client->getId() );
        proxys->index->addServer( (void*)(NestStrdup(client->getHostIp())) );
        nest_hash_table_insert( proxys->coordinator_map, 
                client->getHostIp(),
                client );
        sendProxyIndexToAll();
    } else {
        nest_debug( NEST_DEBUG_PROXY, "New proxy %s (%d) already registered\n",
                client->getHostIdentity(), client->getId() );
        sendProxyIndexToOne( client );
    }
    unlockMutex();
    return NEST_SUCCESS;
}

NestReplyStatus
addNewClientToProxy( nestClient *client )
{

        // nothing to do here now.
    return NEST_SUCCESS;
}

// find out who the coordinator is.  may need to do this if the 
// current coordinator dies as well.
// this currently just keeps looping until it correctly finds one
#if 0
NestReplyStatus
findCoordinator( nestClient *home )
{
    struct ProxyListArgs args;
    args.self  = proxys->self;
    NestReplyStatus status = NEST_UNKNOWN_ERROR;

    int attempts = 0;
    do {
        if ( attempts > 0 ) {
            long sleep_time = 500000000 * attempts;
            nest_warn( __FILE__, __FUNCTION__, "Unable to connect to %s: %s.  "
                       "Will sleep for %.3f and try again.\n", 
                       args.coord, ((float)sleep_time / BILLION ));
            NestNanosleep( sleep_time );
        }
        args.first = true;
        args.coord = NULL;

        if ( proxys->coordinator != NULL ) {
            nest_die( 
                "Trying to find new coordinator but old one still exists.\n");
        }

        status = NestProxyQuery( home->getConnection()->getSocket(),
                receiveProxyList, &args  );
        if ( args.coord == NULL ) {
            nest_die( "Couldn't discover proxy coordinator.\n" );
        }
        if ( status != NEST_SUCCESS ) break;
        
        if ( ! strcmp( args.coord, proxys->self ) ) {
            nest_debug( NEST_DEBUG_PROXY, "Hey!  I'm the proxy coordinator\n" );
            lockMutex();
            proxys->coordinator = NULL;
            if ( proxys->coordinator_map ) {
                nest_hash_table_delete( proxys->coordinator_map );
            }
            proxys->coordinator_map = nest_hash_table_create( 1037, NULL );
            proxys->index->clear( true );

            if ( coordinator_also_serves ) {
                proxys->index->addServer( (void*)NestStrdup(proxys->self) );
            }
            unlockMutex();
        } else {
            status = openCoordinatorConnection( args.coord );
        }
        
        if ( args.coord ) NestFree( args.coord );
    } while ( status != NEST_SUCCESS && attempts < 10 );

    return status;
}
#endif

char *
selfIpAddress()
{

    struct utsname utsname_buf;
    int ret = uname( &utsname_buf );
    if ( ret < 0 ) nest_die( "Couldn't discover self ip address.\n" );
//    printf( "Name is %s.\n", utsname_buf.nodename );
    struct hostent *host_info = gethostbyname( utsname_buf.nodename );
    char *ip = inet_ntoa( *(struct in_addr *)*host_info->h_addr_list );
//    printf( "ip is %s.\n", ip );

    return ip;
}


void
proxyHelp()
{
    nest_printf("--home home, to make a proxy server for a home node\n" );
    nest_printf("--standalone, to make proxies not cooperate\n" );
}


    // set up proxy structures and connect to the home server
int
initProxy( nestClientPool *clients )
{
    CHECK_PID;
    
    const char *home = NULL;

        // save the state passed from the main server
    client_pool = clients;

	// Read the proxy host from the config file...
    const char *value;
    if ( GetConfigValue( Config_ProxyHost, &value ) && *value != 0) {
        home = value;
    }
    nest_debug( NEST_DEBUG_PROXY, "Config for proxy host says home is %s\n",
            home );

    if ( GetConfigValue( Config_ProxysStandAlone, NULL ) ) {
        cooperative_proxies = false;
    }

    if ( home == NULL ) return 0;
        
        // set up the structure
    assert( proxys == NULL );
    proxys = (Proxy_Values *) NestCalloc( sizeof( *proxys ) );
    proxys->self = NestStrdup( selfIpAddress() );
    proxys->home = NestStrdup( home );
    proxys->home_con = NULL;
    proxys->pending = new nestClientPool( false );
    proxys->waiting = new nestClientPool( false );
    proxys->connections = new nestProxyClientCache( );
    proxys->coordinator = NULL;
    proxys->coordinator_map = NULL;
    proxys->coord_status = NEST_NOT_YET_INITIALIZED;
    proxys->index = NULL;
    proxys->index = new DynamicLard( NULL );
    assert( proxys->index != NULL );
    proxys->index->setServerEqualityFunction( checkTwoStrings );
    nest_pthread_mutex_init( &(proxys->mutex), NULL );
    nest_pthread_mutex_init( &(proxys->auth_mutex), NULL );
    nest_debug( NEST_DEBUG_PROXY, "Initing proxy data for proxy of %s.\n",
            proxys->home );
        

        // make an asynchronous task to register with the server
        // we _could_ do it synchronously without deadlock but this
        // would unnecessarily starve clients during any disconnection
    startRegisterHome();
    return 0;

# if 0
        // open connection to the home node and register with it
    status = openHomeConnection( proxys->home );
    if ( status == NEST_SUCCESS ) {
        status = registerProxyConnection( proxys->home_con );
    }
    if ( status != NEST_SUCCESS ) {
        nest_die( "Could not set up as proxy to %s: %s.\n",
                home, NestErrorString( status ) );
    }

        // open connection to the coordinator 
    status = findCoordinator( proxys->home_con );
    if ( status != NEST_SUCCESS ) {
        nest_die( "Could not find the proxy coordinator: %s\n",
                NestErrorString( status ) );
    }
    nest_debug( NEST_DEBUG_PROXY, "Set up as proxy of %s: %s\n",
            home, NestErrorString( status ) );
# endif
}
