/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_common.h"

#include "nest_io_manager.h"
#include "nest_collection_manager.h"
#include "nest_storage_manager.h"
#include "nest_client_pool.h"
#include "nest_protocol_manager.h"
#include "nest_advertiser.h"
#include "nest_protocol_advertise.h"
#include "nest_config.h"
#include "nest_hash_table.h"
#include "nest_user.h"
#include "nest_request_handler.h"
#include "nest_request.h"
#include "nest_proxy.h"
#include "nest_version.h"
#include "nest_server.h"
#include "nest_main.h"

    // client connection states
enum {  CLOSE_CONNECTION,       // client is no longer in the building
        ACTIVE_CONNECTION,      // client is being handled by io mgr
        PASSIVE_CONNECTION      // client should be monitored by server
};

nestRequestHandler::nestRequestHandler( void )
{
	registered_proxys = NULL;
	allowedAuthentication = NEST_AUTH_GSS;
}

nestRequestHandler::~nestRequestHandler( void )
{
}

void
nestRequestHandler::setAllowedAuth( int types )
{
	allowedAuthentication = types;
}

nestRequest *
nestRequestHandler::ReceiveRequest(
	nestClient *client )
{
    struct timeval start, end;
    nestRequest *req;

    NestGettimeofday( &start );
    req = client->receiveRequest( );
    NestGettimeofday( &end );
    if ( NestTimediff( &start, &end, __FUNCTION__ ) > 1000000 ) {
        nest_warn(
			__FILE__, __FUNCTION__,
			"Receive request took %.3f secs.\n", 
			(float)NestTimediff( &start, &end, __FUNCTION__ ) / 1000000 );
    }
	if ( req && (! req->isValid() ) ) {
		nest_warn( __FILE__, __FUNCTION__,
				   "Received invalidated request %d (%s) from client %s\n",
				   req->getType(), req->toString(), client->toString() );
	}
    return req;
}

int
nestRequestHandler::HandleRequests(
	nestClientPool *requests,
	fd_set *readfds )
{
    nestClient *client;
    CLIENT_ITERATOR itr = requests->begin();
    while ( itr != requests->end() ) {
        client = *itr;
        itr = requests->erase( itr );

        switch( HandleRequest( client ) ) {
        case CLOSE_CONNECTION:
            RemoveClient( client, readfds );
            break;
        case PASSIVE_CONNECTION:
        case ACTIVE_CONNECTION:
                // nothing to do, we used to clear the client fd from the
                // read_fd set we pass to select here for an active
                // connection.  But now this is done in the io
                // manager.
            break;
        default:
            assert( 0 );
            break;
        }
    } // end of client service loop
    return 0;
}

void 
nestRequestHandler::taskDone(
	nestClient *client,
	NestReplyStatus status )
{

        // special case for transfers that put data but don't prespecify 
        // a transfer length.  we can only allocate lot space for these
        // transfers incrementally.
    if ( status == NEST_LOT_NOT_ENOUGH ) {
        status = storage_manager->increaseAllocation( client );
        if ( status == NEST_SUCCESS ) {
            status = io_manager->addTask( client );
            return;
        } else {
            client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        }
    }


        // sometimes a proxy may add other special non-transfer requests
        // to the io manager.  Allow them to clear up after them.
    if ( client->getTaskFunction() != NULL ) {
        taskCleanupFunction *cleanup = client->getTaskCleanupFunction();
        assert( cleanup != NULL );
        //NestReplyStatus cleanup_status = 
        cleanup( client, status );
        return;
    }

        // this persistent thing is for a weird case with http clients and
        // the pot io.  It was possible for the taskDone to get
        // called after the client was removed.
        //
        // also however nfs clients are not persistent . . . 
        //
        // I think there still may be a race condition if the client object
        // is deleted while we're operating in taskDone
        // I think it is important to not call client->endTransfer until
        // ___after__ we've returned control to the server
        // because once client->endTransfer is called it is possible that
        // the client could be deleted.  So we can't reference it again.
        // 
    if ( client->hasSocket() && client->persistent() ){
		NestFdRegister( client->getConnection()->getSocket(), FD_READ );
        nest_debug( NEST_DEBUG_CONCURRENCY | NEST_DEBUG_FD, 
					"Task_done adding %d to readfds\n", 
					client->getConnection()->getSocket() );
    }



        // close the storage connections
    if ( client->getStorageConnection() != NULL ) {
        nestProtocol *storage_con = client->getStorageConnection();
        storage_manager->close( storage_con, client );
        delete( storage_con );
        client->assignStorageConnection( NULL ); 
    }

    if ( client->getOtherStorageConnection() != NULL ) {
        nestProtocol *storage_con = client->getOtherStorageConnection();
        storage_manager->close( storage_con ,client );
        delete( storage_con );
        client->assignOtherStorageConnection( NULL ); 
    }

        // TODO:  Be nice to maybe cache these.
    if ( client->getThirdPartyConnection() != NULL ) {
        nestProtocol *third = client->getThirdPartyConnection();
        delete( third );
        client->assignThirdPartyConnection( NULL );
    }
        // this seems weird to me (jb), why do we handle this one particular
        // error specially?
    if ( status == NEST_INSUFFICIENT_SPACE ) { 
        client->sendErrorMessage( __FILE__, __FUNCTION__, status);
    }
        // client endTransfer must be last thing to happen
        // otherwise, client might suddenly disappear
    client->endTransfer( status );

        // if we're running as proxy, start any who were waiting for this
        // and clean this one up
        // no longer needed, bec proxy taskDone is called directly 
        // above using the cleanup function pointer
    /*
    if ( serverIsProxy() ) {
        checkProxyTransfers( client, status ); 
    }
    */

}

int
nestRequestHandler::HandleRequest(
	nestClient *client )
{
	nestRequest	*request = client->getRequest( );
    switch( request->getType() ) {
    case CLIENT_RECV_FILE:
    case CLIENT_SEND_FILE:
    case THIRD_PARTY_GET:
    case THIRD_PARTY_PUT:
    case TEST_CLIENT_TO_SERVER:
    case TEST_SERVER_TO_CLIENT:
        return TransferFile( client, request );
    case COPY_REQUEST:
        return CopyFile(client, request );
    case CLOSE_CONNECTION_REQUEST:
        client->endRequest( NEST_SUCCESS );
        return CLOSE_CONNECTION;
    case NEST_SYNC:
        return NestSync( client, request );
    default:
        return SynchronousRequestHandler( client, request );
    }

}

// these quick simple requests are just handled immediately by the main server
int
nestRequestHandler::SynchronousRequestHandler(
	nestClient *client,
	nestRequest *request )
{

    int connection_state = PASSIVE_CONNECTION;
    nest_debug( NEST_DEBUG_SERVER, "SynchronousRequestHandler for req %d\n",
            request->getType() );
    NestReplyStatus status = NEST_SUCCESS;

        // respond appropriately to each request type
    switch( request->getType() ) {
    case PROXY_REGISTER:
        status = RegisterProxyClient( client, request );
        break;
    case PROXY_QUERY:
        status = QueryProxys( client, request );
        break;
    case PROXY_FLUSH:
        status = FlushProxyCache( client );
        break;
    case PROXY_INDEX_PUSH:
        status = ReceiveProxyIndex( client );
        break;
    case FORMAT_DISK_REQUEST:
        status = FormatDisk( client, request );
        break;
    case BAD_REQUEST:
        status = client->sendErrorMessage( __FILE__, __FUNCTION__,  
                request->getStatus() );
        break;
    case VERSION_REQUEST:
        status = SendVersion( client, request );
        break;
    case REQUEST_NEST_AD:
        status = SendNestAd( client, request );
        break;
    case GET_FILEHANDLE:
        status = GetFilehandle( client, request );
        break;
    case AVAILABLE_SPACE_QUERY:
        status = SendAvailableSpace( client, request );
        break;
    case TIME_REQUEST:
        status = client->sendTimestamp( );
        break;
    case CONNECTION_REQUEST:
    case CHANGE_USER_REQUEST:
        status = AuthenticateUser( client, request );
        break;
    case ADD_NEST_USER:
        status = AddUser( client, request );
        break;
    case ADD_NEST_GROUP:
        status = AddGroup( client, request );
        break;
    case DELETE_NEST_USER:
        status = DeleteUser( client, request );
        break;
    case DELETE_NEST_GROUP:
        status = DeleteGroup( client, request );
        break;
    case ADD_USER_TO_GROUP:
        status = AddUserToGroup( client, request );
        break;
    case DELETE_USER_FROM_GROUP:
        status = DeleteUserFromGroup( client, request );
        break;
    case REQUEST_USER_LOT:
        status = ReserveUserSpace( client, request );
        break;
    case REQUEST_GROUP_LOT:
        status = ReserveGroupSpace( client, request );
        break;
    case LIST_USER_LOTS:
        status = ListUserLots( client, request );
        break;
    case LIST_GROUP_LOTS:
        status = ListGroupLots( client, request );
        break;
    case GET_LOT_STAT:
        status = GetLotStat( client, request );
        break;
    case UPDATE_LOT:
        status = UpdateLot( client, request );
        break;
    case CHANGE_LOT_T:
        status = ChangeLotType( client, request );
        break;
    case LOT_MOVE_FILE:
        status = MoveFileLot( client, request );
        break;
    case ADD_USER_TO_LOT:
        status = AddUserToLot( client, request );
        break;
    case DELETE_USER_FROM_LOT:
        status = DeleteUserFromLot( client, request );
        break;
    case ATTACH_LOT_TO_FILE:
        status = AttachLotToFile( client, request );
        break;
    case DETACH_LOT_FROM_FILE:
        status = DetachLotFromFile( client, request );
        break;
    case TERMINATE_LOT:
        status = TerminateLot( client, request );
        break;
    case FILE_DELETE_REQUEST:
        status = DeleteFile( client, request );
        break;
    case FILE_RENAME_REQUEST:
        status = RenameFile( client, request );
        break;
    case BANDWIDTH_STATISTICS_REQUEST:
        status = SendBandwidthStatistics( client, request );
        break;
    case DIR_LIST_REQUEST:
        status = SendDirectoryList( client, request );
        break;
    case FILEINFO_REQUEST:
        status = SendFileinfo( client, request );
        break;
    case SETATTR_REQUEST:
        status = SetFileAttr( client, request );
        break;
    case FILESIZE_REQUEST:
        status = SendFilesize( client, request );
        break;
    /** This has been replaced by the copy function that uses I/O
        And so it appears along with the other transfer functions
    case COPY_REQUEST:
        status = CopyFile( client, request );
        break;
    */
    case LINK_REQUEST:
        status = MakeLink( client, request );
        break;
    case PWD_REQUEST:
        status = SendPwd( client, request );
        break;
    case MKDIR_REQUEST:
    case CWD_REQUEST:
    case RMDIR_REQUEST:
        status = DirOp( client, request );
        break;
    case PHYSNAME_REQUEST:
        nest_debug( NEST_DEBUG_SERVER, "PhysName request\n" );
        status = SendPhysName( client, request );
        break;
    case GET_ACL:
    case GET_GROUP_ACL:
        nest_debug( NEST_DEBUG_SERVER, "GetUserAcl request\n" );
        status = SendUserAcl( client, request );
        break;
    case SET_ACL:
    case SET_GROUP_ACL:
        nest_debug( NEST_DEBUG_SERVER, "SetUserAcl request\n" );
        status = SetUserAcl( client, request );
        break;
    case NAME_COLLECTION:
    case UNNAME_COLLECTION:
    case QUERY_COLLECTION:
	case SET_COLLECTION_MTIME:
	case SET_COLLECTION_VERSION:
	case SET_COLLECTION_COMMENT:
	case GET_COLLECTION_MTIME:
	case GET_COLLECTION_VERSION:
	case GET_COLLECTION_COMMENT:
		status = CollectionOp( client, request );
        nest_server->WriteAd();
        nest_server->RegularMaintenance();
        break;
    case SERVER_TERMINATE_REQUEST:
        if ( TerminateServer( client, request ) == NEST_SUCCESS ) {
            exit( 0 );
        }
        break;
    case DEBUG_REQUEST:
        status = DebugRequest( client, request );
        break;
    default:
        assert( 0 );
        return CLOSE_CONNECTION;
    }

    client->endRequest( status );

    if ( connection_state == PASSIVE_CONNECTION ) {
            // this means the server is responsible for listening to the client
        if ( ! client->persistent() ) {
                // client is a one time visitor (e.g. nfs or http)
            connection_state = CLOSE_CONNECTION;
        }
    }
    return connection_state;
}

bool
nestRequestHandler::RemoveClient(
	nestClient *client,
	fd_set *readfds )
{
    nest_debug( NEST_DEBUG_CLIENT, "Removing client %d\n", client->getId() );

        // if this client has a socket, stop listening to it 
    if ( client->hasSocket() ) {
        NestFd_clear( client->getConnection()->getSocket(), readfds);
    }

        // if this client registered as a proxy, remove them from proxy list
    if ( client->isProxy( ) ) {
        RemoveProxyClient( client );
    }

    if ( serverIsProxy() ) {
        removeClientFromProxy( client );
    }

    nest_server->getClients()->removeClient( client );
    delete client;
    return true;
}

NestReplyStatus
nestRequestHandler::TerminateServer(
	nestClient *client,
	nestRequest *request )
{

    NestReplyStatus status = NEST_NONPRIVELEDGED_USER;
    const char *tmp_admin, *admin;

        // check whether user is the registered admin user
    if ( ! GetConfigValue( Config_AdminUser, &tmp_admin ) ) {
        status = NEST_NONPRIVELEDGED_USER;
    } else {
        admin = NestGetModifiedIdentity(tmp_admin);
        if ( !strcmp( client->getUserIdentity(), admin ) ) {
            status = NEST_SUCCESS;
        } else {
            status = NEST_NONPRIVELEDGED_USER;
        }
    }

    if ( status == NEST_SUCCESS ) {
        client->sendAck();
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }

    return status;
}

NestReplyStatus
nestRequestHandler::DebugRequest(
	nestClient *client,
	nestRequest *req )
{
	const char	*string = req->getDebugString( );

    NestReplyStatus status = add_nest_debug_level( string );
    if ( status == NEST_SUCCESS ) {
        io_manager->debugLevel( get_nest_debug_level() );
        nest_debug( NEST_DEBUG_DEBUG, "nest_debug set to %s\n", string );
        client->sendAck();
    } else {
        nest_debug( NEST_DEBUG_DEBUG, "could not set nest_debug set %s\n",
            string );
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

NestReplyStatus 
nestRequestHandler::GetFilehandle(
	nestClient *client,
	nestRequest *request )
{
    NestFileInfo info;

    NestReplyStatus status = storage_manager->fileinfo( 
            &info, 
            client->getRequest()->getNestFile(),
            client->getUser( ),
            client->getCwd() );

    if ( status == NEST_SUCCESS ) {
        client->sendFilehandle( &info );
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

NestReplyStatus
nestRequestHandler::SendFileinfo(
	nestClient *client,
	nestRequest *request )
{

    NestFileInfo	info;

	info.next = NULL;
	info.classad = NULL;

    NestReplyStatus status = NEST_NOT_YET_INITIALIZED;
    if ( serverIsProxy() ) {
        status = proxyFileinfo(
            &info,
            client->getRequest()->getNestFile(),
            client->getUser(),
            client->getCwd() );
    } else {
        status = storage_manager->fileinfo(
            &info,
            client->getRequest()->getNestFile(),
            client->getUser(),
            client->getCwd() );

    }

    if ( status == NEST_SUCCESS ) {
        client->sendFileinfo( &info );
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

NestReplyStatus
nestRequestHandler::SetFileAttr(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus     retval;
    char                *phys_name;

    retval = storage_manager->get_physical_filename(
        &phys_name,
        request->getFile(),
        client->getUser(),
        client->getCwd() );
    if( retval != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
    }
   
    NestFileInfo info; 
    retval = storage_manager->fileinfo( 
            &info, 
            request->getFile(),
            client->getUser(),
            client->getCwd() );
    if( retval != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
    }
      
    client->setFileattr(&info, 
                        phys_name,
                        request->getFileAttr() );
    NestFree( phys_name );

    return retval;
}

NestReplyStatus
nestRequestHandler::SendFilesize(
	nestClient *client,
	nestRequest *request )
{

    nestFileSize filesize;

    NestReplyStatus status = NEST_NOT_YET_INITIALIZED;
    if ( serverIsProxy() ) {
        status = proxyFilelength(
            &filesize,
            client->getRequest()->getNestFile(),
            client->getUser(),
            client->getCwd() );
    } else {
        status = storage_manager->filelength(
            &filesize,
            client->getRequest()->getNestFile(),
            client->getUser(),
            client->getCwd() );
    }

    if ( status == NEST_SUCCESS ) {
        client->sendFilesize( filesize );
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

NestReplyStatus
nestRequestHandler::SendPhysName(
	nestClient *client,
	nestRequest *request )
{

    NestReplyStatus        retval;
    char                *phys_name;

    retval = storage_manager->get_physical_filename(
        &phys_name,
        client->getRequest()->getNestFile(),    
        client->getUser(),
        client->getCwd() );

    if ( retval == NEST_SUCCESS ) {
        client->sendPhysName( phys_name );
        NestFree( phys_name );
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
    }

    return retval;
}

NestReplyStatus
nestRequestHandler::SendUserAcl(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus				retval;

	// Get the user/group
	const nestUserOrGroup		*target_ug;
    if ( request->getType() == GET_ACL ) {
		target_ug = request->getUser( );
	} else {
		target_ug = request->getGroup( );
	}

    // Target user specified?
    nest_debug( NEST_DEBUG_SERVER,
                "SendUserAcl: User/Group=%p='%s' Group=%c, File='%s'\n",
				target_ug->getName(),
				( request->getType() == GET_ACL ) ? 'f' : 't',
                client->getRequest()->getNestFile() );

    // Check for all users?
	if ( NULL == target_ug ) {
        NestACL    *Acl;

        retval = storage_manager->get_acl(
                    &Acl,
                    client->getRequest()->getNestFile(),
                    client->getUser(),
                    client->getCwd() );
        nest_debug( NEST_DEBUG_SERVER,
                    "SendUserAcl[ACL]: retval=%d, Acl @@ %p\n", retval, Acl );
        if ( retval == NEST_SUCCESS ) {
            client->sendAcl( Acl );
        } else {
            client->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
        }
    } else {
        // Just a single (specified) user -or- group
        nest_access_rights_t    access;
		retval = storage_manager->get_perm(
			&access,
			target_ug,
			client->getRequest()->getNestFile(),
			client->getUser(),
			client->getCwd() );
        nest_debug( NEST_DEBUG_SERVER,
                    "SendUserAcl[User='%s']: Access=0x%02x\n", 
                    target_ug->getName(), access );
        if ( retval == NEST_SUCCESS ) {
            client->sendAcl( target_ug->getName(),
							 target_ug->isGroup(),
							 access );
        } else {
            client->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
        }
    }

    return retval;
}

NestReplyStatus
nestRequestHandler::SetUserAcl(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus				status;

	// Get the user/group
	const nestUserOrGroup		*target_ug;
    if ( request->getType() == GET_ACL ) {
		target_ug = request->getUser( );
	} else {
		target_ug = request->getGroup( );
	}

    // Get request parameters
    nest_debug( NEST_DEBUG_SERVER,
				"SetUserAcl: User='%s', IsGroup=%c, File='%s', ACL=0x%02x\n",
				target_ug->getName(),
				target_ug->isGroup() ? 't' : 'f',
                request->getFile(),
				request->getRights() );

    // Make the check
	status = storage_manager->set_perm( request->getRights(),
										target_ug,
										request->getFile(),
										client->getUser(),
										client->getCwd() );

    nest_debug( NEST_DEBUG_SERVER, "SetUserAcl: status=%d\n", status );

    // Done
    if ( status == NEST_SUCCESS ) client->sendAck( );
    else client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    return status;
}

NestReplyStatus
nestRequestHandler::SendVersion(
	nestClient *client,
	nestRequest *request )
{
    client->sendVersion( NEST_VERSION );
    return NEST_SUCCESS;
}


NestReplyStatus
nestRequestHandler::SendPwd(
	nestClient *client,
	nestRequest *request )
{
    string pwd;

    NestReplyStatus reply = storage_manager->expand_path( 
            pwd, client->getUser(),
            client->getCwd(), "" );

    if (reply != NEST_SUCCESS) {
        client->sendErrorMessage( __FILE__, __FUNCTION__, reply);
    } else {
        client->sendPwd( pwd.c_str() );
    }

    return reply;
}

NestReplyStatus
nestRequestHandler::CollectionOp(
	nestClient *client,
	nestRequest *request )
{
	string name = request->getCollectionName( );
	bool			sendack = false;

    NestReplyStatus status = NEST_SUCCESS;
    nest_obj_type_t filetype;

    switch( request->getType() ) {

    case NAME_COLLECTION:
        status = storage_manager->get_type( &filetype,
											request->getFile( ),
											client->getUser( ),
											client->getCwd() );

        if (status == NEST_SUCCESS ) {
            string pathname;
            storage_manager->expand_path( pathname,
										  client->getUser( ),
										  client->getCwd(),
										  request->getFile( ) );

            status = collection_manager->createAssociation( name, pathname );
        }
		sendack = true;
        break;

    case UNNAME_COLLECTION:
        status = collection_manager->removeAssociation( name );
		sendack = true;
        break;

	case SET_COLLECTION_MTIME:
	case SET_COLLECTION_VERSION:
	case SET_COLLECTION_COMMENT:
	case GET_COLLECTION_MTIME:
	case GET_COLLECTION_VERSION:
	case GET_COLLECTION_COMMENT:
	{
		sendack = true;
		nestCollection	*col = collection_manager->getCollection( name );
		if ( ! col ) {
			return NEST_LOCAL_FILE_NOT_FOUND;
		}

		switch( request->getType() ) {
		case SET_COLLECTION_MTIME: {
			col->setMtime( request->getCollectionMtime() );
			break;
		}
		case SET_COLLECTION_VERSION: {
			col->setVersion( request->getCollectionVersion( ) );
			break;
		}
		case SET_COLLECTION_COMMENT: {
			col->setComment( request->getCollectionComment( ) );
			break;
		}
		case GET_COLLECTION_MTIME:
		case GET_COLLECTION_VERSION:
		case GET_COLLECTION_COMMENT:
			break;
		}
		delete col;
		break;
	}

    case QUERY_COLLECTION:
    default:
        assert( 0 );
    }

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else if ( sendack ) {
		client->sendAck( );
	} else {
		assert( 0 );		// QUERY_COLLECTION not yet in chirp
    }

    return status;
}

NestReplyStatus
nestRequestHandler::DirOp(
	nestClient *client,
	nestRequest *request )
{
    const char *dir  = request->getDir();
    const nestUser *user = client->getUser();
    const char *cwd  = client->getCwd();

    NestReplyStatus status;

    switch( request->getType() ) {
    case RMDIR_REQUEST:
        status = storage_manager->rmDir( user, cwd, dir );
        break;

    case CWD_REQUEST:
	{
		char *new_cwd;
        status = storage_manager->chDir( &new_cwd, user, cwd, dir );
        if (status == NEST_SUCCESS) {
            status = client->setCwd( new_cwd );
            NestFree( new_cwd );
        }
        break;
	}

    case MKDIR_REQUEST:
        status = storage_manager->mkDir( user, cwd, dir, user->getUid() );
        break;

    default:
        assert( 0 );
    }

    if ( status == NEST_SUCCESS ) {
        client->sendAck( );
    }
    else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

void
nestRequestHandler::getAuthString(
	int types,
	string &str )
{
	static	struct { int	mask; const char *name; } table [] = {
		{ NEST_AUTH_NONE, "none" },
		{ NEST_AUTH_GSS, "GSS" },
		{ NEST_AUTH_IP, "IP" },
		{ 0, NULL }
	};

	// Walk through the table..
	int		count = 0;
	for ( int i = 0;  table[i].mask;  i++ ) {
		if ( types & table[i].mask ) {
			if ( count++ ) {
				str += ", ";
			}
			str += "\"";
			str += table[i].name;
			str += "\"";
		}
	}
}

NestReplyStatus
nestRequestHandler::AuthenticateUser(
	nestClient *client,
	nestRequest *request )
{
    char *username = NULL;
    char *identity = NULL;
    NestReplyStatus stat = NEST_SUCCESS;


	int		auth_type = request->getAuthType();
	string	auth_str;
	getAuthString( auth_type, auth_str );

	// Note: This is not-reentrant
	static string	allowed_str;
	static int		allowed_cache = -1;
	if ( allowed_cache != allowedAuthentication ) {
		allowed_str = "";
		allowed_cache = allowedAuthentication;
		getAuthString( allowedAuthentication, allowed_str );
	}

        // check whether requested authentication method is allowed
    if ( ! ( auth_type & allowedAuthentication) ) {
        nest_debug( NEST_DEBUG_AUTH,
					"Client %s requested unallowed authentication type:"
					" %d %s\n",
					request->getUser()->getName(),
					auth_type, auth_str.c_str() );
		nest_debug( NEST_DEBUG_AUTH, "Allowed types: %s\n",
					allowed_str.c_str() );
        stat = NEST_INVALID_AUTHENTICATION;
        client->sendErrorMessage( __FILE__, __FUNCTION__,  stat ); 
        return stat;

    } else if ( check_nest_debug_status( NEST_DEBUG_AUTH ) ) {
		const nestUser *user = request->getUser();
        nest_debug( NEST_DEBUG_AUTH,
					"Client %s (%s) requesting auth type %d %s\n",
					user->getName(), client->getHostIp(),
					auth_type, auth_str.c_str() );
    }

        // check requested authentication method
    switch( auth_type ) {

    case NEST_AUTH_GSS:
        stat = client->getConnection()->gssAccept( &username, &identity );
        if ( stat != NEST_SUCCESS ) {
            nest_debug( NEST_DEBUG_AUTH,
						"Client %s (%s) couldn't authenticate: %s\n",
						request->getUser(),
						client->getHostIp(),
						NestErrorString( stat ) );
            client->sendErrorMessage( __FILE__, __FUNCTION__,  stat );
        }
        break;

    case NEST_AUTH_NONE:
        username = NestStrdup( (char*)request->getUser() );
        identity = NestStrdup( username );
        break;

	case NEST_AUTH_IP:
		// TODO:  put in IP based authentication ala Kangaroo
        stat = NEST_INVALID_AUTHENTICATION;
		client->sendErrorMessage( __FILE__, __FUNCTION__,  stat );
		nest_warn( __FILE__, __FUNCTION__,
				   "Client %s (%s) requested IP authentication"
				   " which is not implmented\n",
				   request->getUser(), client->getHostIp() );
		break;

    default:
        nest_warn( __FILE__, __FUNCTION__, "Unknown auth type requested\n" );
        client->sendErrorMessage( __FILE__, __FUNCTION__,  
                NEST_INVALID_AUTHENTICATION );
        stat = NEST_INVALID_AUTHENTICATION;
    }

        // if we make it here, the authentication was successful
        // now go ahead and set the user name accordingly
    if ( stat == NEST_SUCCESS ) {
        assert( username != NULL );
        assert( identity != NULL || (allowedAuthentication & NEST_AUTH_NONE) );

		// Create a nest user object
		nestUser	*user = new nestUser( identity );
		user->setLocalName( username );

		int valid = storage_manager->validate_user( user );
		if ( valid ) {

            nest_debug( NEST_DEBUG_SERVER | NEST_DEBUG_CLIENT,
						"Client identity set to %s (user %s / uid %d)\n", 
						user->getIdentity(),
						user->getOsName(),
						user->getUid() );

			// Create user home directory...
			(void) storage_manager->add_home_dir( user );

			// Note: client->setUser() passes ownership of the user object
			client->setUser( user );
			user = NULL;
            client->sendAck( );
        } else {
            nest_warn( __FILE__, __FUNCTION__, 
                    "WARN: unable to change to user: %s\n", identity );
            client->sendErrorMessage(__FILE__, __FUNCTION__,
									 NEST_UNKNOWN_USER);
        }

        if ( username != NULL ) {
			NestFree( username );
		}
        if ( identity != NULL ) {
			NestFree( identity );
		}
        if ( user != NULL ) {
			delete user;
		}

            // proxy servers may need to do something special with
            // this client
        if ( serverIsProxy() ) {
            addNewClientToProxy( client ); 
        }

    }

    return stat;
}

NestReplyStatus
nestRequestHandler::AddUser(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus status;

	// Create a user object with the default group
	const nestGroup	*new_group = request->getGroup( );
	nestUser		new_user( request->getUser( ) );
	new_user.setGroup( new_group );

	status = storage_manager->add_user( client->getUser(), &new_user );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
nestRequestHandler::DeleteUser(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus status;

    status = storage_manager->delete_user( client->getUser(),
										   request->getUser() );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
nestRequestHandler::AddGroup(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus status;

    status = storage_manager->add_group( client->getUser(),
										 request->getGroup() );
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
nestRequestHandler::DeleteGroup(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus status;

    status = storage_manager->delete_group( client->getUser(),
											request->getGroup() );
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
nestRequestHandler::AddUserToGroup(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus status;

    status = storage_manager->add_user_to_group(
		client->getUser(),
		request->getUser(),
		request->getGroup() );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
nestRequestHandler::DeleteUserFromGroup(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus status;

    status = storage_manager->delete_user_from_group(
		client->getUser(), request->getUser(), request->getGroup() );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
nestRequestHandler::ReserveUserSpace(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus status = NEST_SUCCESS;
    NestLotHandle handle;

	// Create a "nest user" object
	const nestUser *user = request->getUser( );
	if( ! user ) {
		user = client->getUser( );
	}

	// And, one for the parent user / group
	status = lot_manager->request_user_lot(
		&handle,
		client->getUser( ),
		request->getLotSize( ),
		request->getLotDuration( ),
		user,
		request->getLotParentUG( ),
		request->getLotParentLot( ),
		request->getLotType( )
		);

	{
		rsv_id_t	tmpid;
		NestGet_id_from_handle( &handle, &tmpid );
		nest_debug( NEST_DEBUG_SERVER, "ReserveUserSpace: Handle=%p, id=%d\n",
					&handle, tmpid );
	}
                        
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
        client->sendLotHandle( &handle );
    }
    return status;
}

NestReplyStatus
nestRequestHandler::ReserveGroupSpace(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus status;
    NestLotHandle handle;

    status = lot_manager->request_group_lot(
                &handle,
                client->getUser(),
                request->getLotSize(),
                request->getLotDuration(),
                request->getGroup( ),
				request->getLotParentUG( ),
                request->getLotParentLot( ),            
                request->getLotType( )
             );
            
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
        client->sendLotHandle( &handle );
    }
    return status;
}

NestReplyStatus
nestRequestHandler::ListUserLots(
	nestClient *client,
	nestRequest *request )
{
    NestLotListInternal     *result  = NULL;
    NestReplyStatus status;

	const nestUser	*user = request->getUser( );
	if ( ! user ) {
		user = client->getUser( );
	}

	status = lot_manager->list_lots(
		&result,
		client->getUser(),
		user );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        while (result) {
            NestLotListInternal *tmp = result->next;
            delete result;
            result = tmp;
        }
        return status;
    } else {
        client->sendAck();
        while (result) {
            NestLotListInternal *tmp = result->next;
            client->sendNestLot( &result->stat );
            delete result;
            result = tmp;
        }
    }

    client->sendAck();
    return status;
}

NestReplyStatus
nestRequestHandler::ListGroupLots(
	nestClient *client,
	nestRequest *request )
{
    NestLotListInternal     *result = NULL;
    NestReplyStatus status;

	status = lot_manager->list_lots(
		&result,
		client->getUser( ),
		request->getGroup( ) );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        while (result) {
            NestLotListInternal *tmp = result->next;
            delete result;
            result = tmp;
        }
        return status;
    } else {
        client->sendAck();
        while (result) {
            NestLotListInternal *tmp = result->next;
            client->sendNestLot( &result->stat );
            delete result;
            result = tmp;
        }
    }

    client->sendAck();
    return status;
}

NestReplyStatus
nestRequestHandler::GetLotStat(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus ret_val;
    NestLotInternal rstat;

    const NestLotHandle   *handle = request->getLotHandle( );
    if ( ! handle ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,
								  NEST_BAD_PARAMETERS );
    }

    ret_val = lot_manager->stat_lot(
		handle,
		client->getUser( ),
		&rstat );

    if (ret_val != NEST_SUCCESS) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  ret_val );
    } else {
        client->sendAck();
        client->sendNestLot( &rstat );
    }
    return ret_val;
}

NestReplyStatus
nestRequestHandler::UpdateLot(
	nestClient *client,
	nestRequest *request )
{
	NestReplyStatus status;

	const NestLotHandle   *handle = request->getLotHandle( );
    if ( ! handle ) {
        status = NEST_BAD_PARAMETERS;
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        return status;
    }

    status = lot_manager->update_lot(
		handle,
		client->getUser( ),
		request->getLotSize( ),
		request->getLotDuration() );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
nestRequestHandler::ChangeLotType(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus status;

	const NestLotHandle   *handle = request->getLotHandle( );
     status = lot_manager->change_lot_type(
                handle,
                client->getUser( ),
                request->getLotType(),
                request->getLotDuration() );
    
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
nestRequestHandler::MoveFileLot(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus status;
    nestLotSize movebytes = request->getLotBytesToMove();
    const NestLotHandle *source_lot = request->getLotSourceHandle( );
	const NestLotHandle *dest_lot = request->getLotDestHandle( );

    // Extract the file name..

    // Check the Bytes to move
    if (movebytes < 0 || !source_lot || !dest_lot )
    {
        status = NEST_BAD_PARAMETERS; 
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        return status;
    }

    status = storage_manager->modify_lot_usage(
        client->getUser( ),
        request->getNestFile(),
        client->getCwd(),
        source_lot,
        movebytes,
        dest_lot );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
nestRequestHandler::TerminateLot(
	nestClient *client,
	nestRequest *request )
{
    const NestLotHandle *handle = request->getLotHandle( );

    NestReplyStatus status;

    if ( !handle ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,
								  NEST_BAD_PARAMETERS );
    }

    status = lot_manager->terminate_lot( handle,
										 client->getUser() );
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
nestRequestHandler::AddUserToLot(
	nestClient *client,
	nestRequest *request )
{
    const NestLotHandle *handle = request->getLotHandle( );
    NestReplyStatus status; 
        
    if ( !handle ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,
								  NEST_BAD_PARAMETERS );
    }

    status = lot_manager->add_user_to_lot( handle,
										   client->getUser( ),
										   request->getUser( ) );

    if(status != NEST_SUCCESS) {    
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
nestRequestHandler::DeleteUserFromLot(
	nestClient *client,
	nestRequest *request )
{
    const NestLotHandle *handle = request->getLotHandle( );
    NestReplyStatus status; 
        
    if ( !handle ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,
								  NEST_BAD_PARAMETERS );
    }

    status = lot_manager->delete_user_from_lot(
		handle,
		client->getUser( ),
		request->getUser() );

    if(status != NEST_SUCCESS) {    
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;    
}

NestReplyStatus
nestRequestHandler::AttachLotToFile(
	nestClient *client,
	nestRequest *request )
{
    const NestLotHandle *handle = request->getLotHandle( );

    NestReplyStatus status;

    if ( !handle ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,
								  NEST_BAD_PARAMETERS );
    }

    status = storage_manager->attach_lot_to_file( handle,
												  request->getFile( ),
												  client->getUser( ),
												  client->getCwd() );


    if(status != NEST_SUCCESS) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }

    return status;
}                                 
                                                    
NestReplyStatus
nestRequestHandler::DetachLotFromFile(
	nestClient *client,
	nestRequest *request )
{
    const NestLotHandle *handle = request->getLotHandle( );

    NestReplyStatus status;

    if ( !handle ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__, 
								  NEST_BAD_PARAMETERS );
    }

    status = storage_manager->detach_lot_from_file( handle,
													request->getFile( ),
													client->getUser( ),
													client->getCwd() );

    if(status != NEST_SUCCESS) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }

    return status;
}            

NestReplyStatus
nestRequestHandler::MakeLink(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus status = storage_manager->Link(
            client->getUser( ),
            request->getFile(),
            request->getDestFile(),
            request->getIsSymLink(),
            client->getCwd() );

    if ( status == NEST_SUCCESS ) {
        client->sendAck();
    }
    else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }

    return status;
}

NestReplyStatus
nestRequestHandler::CopyFile(
	nestClient *client,
	nestRequest *request )
{
	// This copy function in storage manager is no longer used

#if 0
    NestReplyStatus status =
        storage_manager->copyFile( client->getIdentity(),
                                      request->getNestFile(),
                                      request->getLinkName(),
                                      client->getCwd() );

    if( status == NEST_SUCCESS ) client->sendAck( );
    else client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    return status;
#endif

    NestReplyStatus status;

    nestProtocol *storage, *otherStorage;

    status = openStorageConnections( &storage, &otherStorage, client );
    if(status == NEST_SUCCESS){
        client->assignStorageConnection( storage );
        client->assignOtherStorageConnection( otherStorage );
        status = io_manager->addTask( client );
    }
        
    if ( status == NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_TRANSFER,
                "copy transfer added to IO manager\n" );
        client->dump( );
        return ACTIVE_CONNECTION;
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        client->endRequest( status );
        nest_debug( NEST_DEBUG_TRANSFER, "copy transfer failure: %s\n",
                NestErrorString( status ) );
        client->dump( );
        return PASSIVE_CONNECTION;
    }

}

NestReplyStatus
nestRequestHandler::DeleteFile(
	nestClient *client,
	nestRequest *request )
{

    const char *file = request->getNestFile();

    NestReplyStatus status = 
        storage_manager->deleteFile( client->getUser( ), 
									 file,
									 client->getCwd() );

    if ( status == NEST_SUCCESS ) client->sendAck( );
    else client->sendErrorMessage( __FILE__, __FUNCTION__,  status );

    return status;
}

NestReplyStatus
nestRequestHandler::RenameFile(
	nestClient *client,
	nestRequest *request )
{
    NestReplyStatus status =
        storage_manager->renameFile( client->getUser( ),
									 request->getFile(),
									 request->getDestFile(),
									 client->getCwd()
									 );
                                
    if ( status == NEST_SUCCESS ) client->sendAck( );
    else client->sendErrorMessage( __FILE__, __FUNCTION__,  status );

    return status;
    
}


NestReplyStatus
nestRequestHandler::openStorageConnections(
	nestProtocol **storage,
	nestProtocol **otherStorage,
	nestClient   *client)
{

    NestReplyStatus     reply;
    nestProtocol *storage_con, *other_storage_con;
    nestRequest *request = client->getProxyRequest();
    nestFileSize offset = request->getStart();
	int link_count = 0;
    
    nestFileSize filesize;
    reply = storage_manager->filelength(
		&filesize,
		request->getNestFile(),
		client->getUser( ),
		client->getCwd() );
    NEST_REPLY_CHECK(reply);
    request->setEnd(filesize);

	// Open the destination file
    reply = storage_manager->open( &storage_con,
								   client->getUser(),
								   request->getDestFile(),
								   WRITE_FLAGS,
								   offset,
								   filesize,
								   client->getCwd(),
								   link_count );
	*storage = ( reply == NEST_SUCCESS ) ? (nestProtocol *)storage_con : NULL;
    NEST_REPLY_CHECK(reply);

	// Open the source file
    reply = storage_manager->open( &other_storage_con,
								   client->getUser(),
								   request->getFile(),
								   READ_FLAGS,
								   offset,
								   filesize,
								   client->getCwd(),
								   link_count );
	*otherStorage =
		( reply == NEST_SUCCESS ) ? (nestProtocol *)other_storage_con : NULL;
    NEST_REPLY_CHECK(reply);

    nest_debug(NEST_DEBUG_TRANSFER, "Created copy transfer: %s->%s\n",
			   request->getNestFile(), request->getDestFile() );

    return reply;
}

NestReplyStatus
nestRequestHandler::openStorageConnection(
	nestProtocol **storage,
	nestClient *client,
	const nestRequest *request )
{

    NestReplyStatus     reply = NEST_SUCCESS;
    nestFileSize  offset;
    nestFileSize filesize;

	int link_count = 0;
	nestProtocol *tmp;

        // TODO: check with consistency manager
    switch( request->getType() ) {
    case CLIENT_SEND_FILE:
    case THIRD_PARTY_GET:
        offset = request->getStart();
        filesize = request->getEnd();
        reply = storage_manager->open( &tmp,
									   client->getUser(),
									   request->getFile(),
									   WRITE_FLAGS,
									   offset,
									   filesize,
									   client->getCwd(),
									   link_count );
		if ( reply == NEST_SUCCESS ) {
			*storage = tmp;
		} else {
			*storage = NULL;
		}
        break;
    case CLIENT_RECV_FILE:
    case THIRD_PARTY_PUT:
        offset = request->getStart();
        filesize = -1;
        reply = storage_manager->open( &tmp,
									   client->getUser(),
									   request->getFile(),
									   READ_FLAGS,
									   offset,
									   filesize,
									   client->getCwd(),
									   link_count );
		if ( reply == NEST_SUCCESS ) {
			*storage = tmp;
		} else {
			*storage = NULL;
		}
        break;
    default:
        *storage = NULL;
		nest_debug( NEST_DEBUG_ALWAYS,
					"Unknown stoarge request %d\n", request->getType() );
        break;
    }

    return reply;
}

    // this function is called by both the home server and proxy servers
    // so a pointer is passed to the proxyInit function
    //
    // when we run in proxy mode, we translate client requests into
    // requests for chunks.  however, we also will use transfer file
    // to send the proxy chunk once it arrives from the server
    // so we only want to translate in the first case, but not the second
int
nestRequestHandler::TransferFile(
	nestClient *client,
	nestRequest *request,
	bool already_translated )
{

    nestRequest *req = NULL;
	nestRequest *client_req = request;

	// Handle proxy requests, too...
    if ( serverIsProxy() ) {
        if ( ! already_translated ) {
            proxyTranslateRequest( client );
        }
        req = client->getProxyRequest();
    } else {
        req = request;
    }

    NestReplyStatus status = NEST_SUCCESS;

    int  type         = req->getType();
    nestFileSize end  = req->getEnd();
    const char *file  = req->getFile();

    nest_debug( NEST_DEBUG_TRANSFER,
				"TransferFile %s for %s (%s) (len %lld), type %d\n",
				file, client->getUserIdentity(),
				client->getHostIp(), end, type );

    switch( type ) {
    case CLIENT_SEND_FILE:
            // try to reserve the necessary space
            // status = MakeReservation( client );
        break;

    case CLIENT_RECV_FILE:
    case THIRD_PARTY_PUT:
        nestFileSize filelength;
            // get the filelength, if file not found, report this and return
        status = storage_manager->filelength( &filelength,
											  file,
											  client->getUser(),
											  client->getCwd() );
		if ( status != NEST_SUCCESS ) {
            break;
        }

            // check if user is requesting to read the entire file or to read
            // past eof, adjust that here
            // don't forget to adjust in both proxy and normal reqs
        if ( end == ENTIRE_FILE ) {
            end = filelength;
            req->setEnd( filelength );
            client_req->setEnd( filelength );
        } else if ( end > filelength ) {
            nest_debug( NEST_DEBUG_TRANSFER, 
                        "Req end %lld > filelength %lld. Fixing\n",
                        end, filelength );
            end = filelength;
            //long diff = req->getEnd() - filelength;
                // these next two lines are what proxys should do...
            //req->setEnd( req->getEnd() - diff );
            //request->setEnd(request->getEnd() - diff);
			nestFileSize new_end = filelength + req->getChunkSize();
			req->setEnd( new_end );
            client_req->setEnd( new_end );
			nest_debug( NEST_DEBUG_TRANSFER,
						"After fix: end=%lld, new_end=%lld, chunk=%lld\n",
						end, new_end, req->getChunkSize() );
        }
            // special case, user trying to read past eof
        if ( req->getStart() >= filelength + req->getChunkSize() ) {
                // make it a 0 byte transfer so they will go away
			req->setStart( req->getEnd() );
			client_req->setStart( client_req->getEnd() );
        }
        break;

    case THIRD_PARTY_GET:
            // nothing to do, except when we find out the size, maybe
            // we need to try to make a reservation . . .
        break;

    case TEST_CLIENT_TO_SERVER:
    case TEST_SERVER_TO_CLIENT:
            // nothing to do here
        break;

    default:
        status = NEST_UNKNOWN_REQUEST_TYPE;
        break;
    }

        // deal with proxy logic here to retrieve files we don't have
        // probably should never see this NEST_FLOCK error
    if ( serverIsProxy() &&
		 ( status==NEST_FILE_NOT_FOUND || status==NEST_FLOCK )  ){
        if ( status == NEST_FLOCK ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "Possibly mishandling an unexpected error at %s,%s: %s",
                    __FILE__, __FUNCTION__, NestErrorString( status ) );
        }
        assert( type == CLIENT_RECV_FILE ); 
        nest_debug( NEST_DEBUG_PROXY, "Will proxy request %s for %s.\n",
                file, client->getUserIdentity() );
        status = addProxyTransfer( client );
        if ( status == NEST_SUCCESS ) {
            return ACTIVE_CONNECTION;
        } else {
            nest_debug( NEST_DEBUG_PROXY, "Failed to proxy request %s: %s.\n",
                    file, NestErrorString( status ) ); 
                // fall through to sendErrorMessage
        }
    }

        // make sure parameters are OK
        // but don't overwrite a previously set error value
	nestFileSize	start = req->getStart( );
    if ( status == NEST_SUCCESS ) {
        if ( (start < 0) || ((end != ENTIRE_FILE) && (end < start)) ) {
            nest_warn( __FILE__,__FUNCTION__,
					   "bad params, %ld end, %ld start\n",
					   end, start );
            status = NEST_BAD_PARAMETERS;
        }
    }
    
        // if no errors yet, add transfer request into io manager
    if ( status == NEST_SUCCESS ) {
            // ioManager now responsible for client->endRequest
        nestProtocol *storage;
        status = openStorageConnection( &storage, client, client_req );

		// Ok?
		if ( status != NEST_SUCCESS ) {
			nest_debug( NEST_DEBUG_TRANSFER,
						"transfer failure: storage error: %s\n",
						NestErrorString( status ) );
		}
		else
		{
			// This is for transfers that don't send filesize info
			if( req->getEnd() == ENTIRE_FILE ) {
				status = storage_manager->increaseAllocation( client );
				if(status != NEST_SUCCESS) {
					storage_manager->cleanupGftpTransfer(client);
				}
			}

			if ( status == NEST_SUCCESS ) {
				client->assignStorageConnection( storage );
				status = io_manager->addTask( client );
			}
		}
    }

    if ( status == NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_TRANSFER,
                "transfer added to io manager\n" );
        client->dump( );
        return ACTIVE_CONNECTION;
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        client->endRequest( status );
        nest_debug( NEST_DEBUG_TRANSFER, "transfer failure: %s\n",
                NestErrorString( status ) );
        client->dump( );
        return PASSIVE_CONNECTION;
    }
}

NestReplyStatus
nestRequestHandler::SendAvailableSpace(
	nestClient *client,
	nestRequest *request )
{
    nestLotSize avail, total;
    NestReplyStatus status = storage_manager->bytesAvailable(IS_USER,
            &avail, &total );
    if ( status == NEST_SUCCESS ) {
        client->sendDiskspace( avail, total );
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

int
nestRequestHandler::NestSync(
	nestClient *client,
	nestRequest *request )
{
    int connection = PASSIVE_CONNECTION;
    struct timeval before, after;
    nest_debug( NEST_DEBUG_SERVER, "client %s requested a sync\n",
            client->getUserIdentity() );
    NestGettimeofday( &before );
    NestReplyStatus status = storage_manager->sync( client->getUser() );
    NestGettimeofday( &after );

    if( status == NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_SERVER, "sync successful, took %ld millis\n",
                NestTimediff( &before, &after, __FUNCTION__ ) );
        if ( serverIsProxy() ) {
            status = flushDirtyWrites( client ); 
                // client is now responsibility of proxy server 
            connection = ACTIVE_CONNECTION; 
        } else {
            client->sendAck();
        }
    } else {
        nest_warn( __FILE__, __FUNCTION__,
				   "sync failure: %s\n", NestErrorString( status ) );
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }

    if ( connection == PASSIVE_CONNECTION ) {
            // otherwise, the proxy code must eventually end the request
        client->endRequest( status );
    }
    return connection;
}


#define PROXY_ITERATOR list<char*>::iterator

// a client (or a proxy monitor) can query a home nest to know which proxy
// nodes have registered themselves for a particular client
NestReplyStatus
nestRequestHandler::QueryProxys(
	nestClient *client,
	nestRequest *request )
{
    list< char *> *proxys;

    if ( registered_proxys == NULL ) {
        proxys = NULL;
    } else {
        proxys = (list<char*>*)
            nest_hash_table_lookup( registered_proxys,
									client->getHostIdentity() );
    }

    client->getConnection()->sendProxyList( proxys );

    return NEST_SUCCESS;
    
}

NestReplyStatus
nestRequestHandler::RemoveProxyClient(
	nestClient *client )
{

    list< char *> *proxys;
    const char *proxy = client->proxyAddress();

    nest_debug( NEST_DEBUG_PROXY, "Removing proxy %s for client %s.\n",
                proxy, client->getHostIdentity() );

        // find the list of proxys registered for this client
    assert( registered_proxys != NULL );
    void *entry = nest_hash_table_lookup( registered_proxys,
										  client->getHostIdentity());
    proxys = (list<char *> *)entry;
    assert( proxys != NULL );

        // remove this particular proxy
    bool found = false;
    PROXY_ITERATOR itr = proxys->begin();
    while ( itr != proxys->end() ) {
        char *g = (char *)*itr;
        if ( ! strcmp( g, proxy ) ) {
            found = true;
            itr = proxys->erase( itr );
            NestFree( g );
            break;
        }
        itr++;
    }
    if ( ! found ) {
        nest_warn( __FILE__, __FUNCTION__,
				   "Tried to remove proxy %s for client %s, not found.\n",
				   proxy, client->getHostIdentity() );
        //assert( found );
    }

    return NEST_SUCCESS;
}

    
// there are two reasons why clients register with a nest:
// 1) nest is home node and client is a proxy server
// 2) nest is proxy coordinator node and client is a proxy server
NestReplyStatus
nestRequestHandler::RegisterProxyClient(
	nestClient *client,
	nestRequest *request )
{

    list< char *> *proxys;      // the list of proxys known for a client
    const char *proxy;          // the proxy location being added now

    proxy = request->getProxyLocation();
    client->setProxy( proxy );
    nest_debug(NEST_DEBUG_PROXY,
			   "Client %s (%d) from %s registering proxy %s\n",
			   client->getHostIdentity(), client->getId(),
			   client->getHostIp(), proxy);

        // init the main structure if there already isn't one
    if ( registered_proxys == NULL ) {
        registered_proxys = nest_hash_table_create( 1037, NULL );
    }

        // see if this client has previously registered any proxys
        // if not, create a list for them
    void *entry = nest_hash_table_lookup( registered_proxys,
										  client->getHostIdentity());
    if ( entry == NULL ) {
        proxys = new list< char *>;
        proxys->clear();
        nest_hash_table_insert( registered_proxys, 
                                client->getHostIdentity(), 
                                (void *)proxys );
    } else {
        proxys = (list<char *> *)entry;
    }

        // see whether we need to add this proxy to the list 
    bool found = false;
    for( PROXY_ITERATOR itr = proxys->begin(); itr != proxys->end(); itr++ ) {
        if ( ! strcmp( *itr, proxy ) ) {
            found = true;
            break;
        }
    }
        // important to do a push_back and not a push_front so that
        // new dynamicly larding pfs clients put proxys into same buckets
        // as old pfs dynamically larding clients
    if ( ! found ) proxys->push_back( NestStrdup( proxy ) );

        // if the proxy is coordinator needs to know about this
    if ( serverIsProxy() ) {
            // proxy will deal with sending ack or error reply
        return proxyRegisterProxyClient( client, (char*)proxy );
    }

    //return QueryProxys( client );
    client->sendAck();
    return NEST_SUCCESS;
}

NestReplyStatus
nestRequestHandler::FormatDisk(
	nestClient *client,
	nestRequest *request )
{
        // delete all entries and report available space
    NestReplyStatus status = 
        storage_manager->formatDisk( client->getUser() , NULL );
    SendAvailableSpace( client, request );
    return status;
}

NestReplyStatus
nestRequestHandler::SendNestAd(
	nestClient *client,
	nestRequest *request )
{

    NestReplyStatus status = NEST_SUCCESS;
        // should we update it first maybe?
    nest_server->WriteAd();

    nestProtocol *adfile = storage_manager->openMetaFile(AD_NAME,READ_FLAGS);

    if ( adfile == NULL ) {
        status = NEST_REMOTE_FILE_NOT_FOUND; 
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        int adlen = storage_manager->metaFilelength( AD_NAME );
        nestAd *ad = advertiser->readAd( adfile, adlen );
        if ( ad == NULL ) {
            status = NEST_REMOTE_FILE_NOT_FOUND; 
            client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        } else {
            client->sendAd( ad );
            delete( ad );
        }
        delete( adfile );
    }
    return status; 
}

NestReplyStatus
nestRequestHandler::SendDirectoryList(
	nestClient *client,
	nestRequest *request )
{

    NestFileInfo *list =
        storage_manager->directoryContents( client->getUser(),
											client->getCwd() );

    client->sendDirectoryList( list );
    while( list != NULL ) {
        NestFileInfo *tmp = list;
        list = list->next;
        NestFree( tmp );
    }
    return NEST_SUCCESS;
}

NestReplyStatus
nestRequestHandler::SendBandwidthStatistics(
	nestClient *client,
	nestRequest *request )
{
    client->sendBandwidthStatistics( );
    return NEST_SUCCESS;
}
