/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_debug.h"
#include "nest_pthread_util.h"
#include "nest_shared.h"

// Max # of args to parse
#define NEST_MAXARGS    	20

// Pthread support
#ifdef USE_PTHREADS
pthread_mutex_t nest_common_mutex;
pthread_mutex_t nest_malloc_mutex;
pthread_mutex_t nest_strdup_mutex;

void
initCommonMutex( ) {
    nest_pthread_mutex_init( &nest_common_mutex, NULL );
    nest_pthread_mutex_init( &nest_malloc_mutex, NULL );
    nest_pthread_mutex_init( &nest_strdup_mutex, NULL );
}

void
lockCommonMutex( pthread_mutex_t *mux, char *func, int line ) {
    nest_debug( NEST_DEBUG_CONCURRENCY, "%d acquiring 0x%x from %s:%d\n",
				NestGetpid(), mux, func, line );
    nest_pthread_mutex_lock( mux );
}

void
unlockCommonMutex( pthread_mutex_t *mux, char *func, int line ) {
    nest_debug( NEST_DEBUG_CONCURRENCY, "%d released 0x%x from %s:%d\n",
				NestGetpid(), mux, func, line );
    nest_pthread_mutex_unlock( mux );
}

#else

void initCommonMutex()                                      {}
void lockCommonMutex    (pthread_mutex_t *, char *, int )   {}
void unlockCommonMutex  (pthread_mutex_t *, char *, int )   {}

#endif


bool
NestIsDir( const char *filename )
{
    struct stat buf;

    if ( ( stat( filename, &buf ) ) < 0 || ! S_ISDIR( buf.st_mode ) ) {
        return false;
    }

    return true;
}

bool
NestIsFile( const char *filename )
{
    struct stat buf;

    if ( ( stat( filename, &buf ) ) < 0 || ! S_ISREG( buf.st_mode ) ) {
        return false;
    }

    return true;
}

bool
NestIsLink( const char *filename )
{
    struct stat buf;

    if ( ( lstat( filename, &buf ) ) < 0 || ! S_ISLNK( buf.st_mode ) ) {
        return false;
    }

    return true;
}

void
NestParseLine( char *line, char ***argv, int *argc )
{
    char tmp_arg[NEST_MAXPATHLEN];
    int i,j;
    *argc=0;
 
    *argv = (char**) NestCalloc(NEST_MAXARGS*sizeof(char*));
    if(!*argv) return;
    
    for(i=0; i<NEST_MAXARGS; i++)
        (*argv)[i] = 0;
 
    i=0;
    bool quoted = false;
    for( (*argc)=0; (*argc)<NEST_MAXARGS; (*argc)++ ) {
        j=0;
        while( line[i] != 0 && line[i] != '\n' && 
			   (line[i] == ' ' || line[i] == '\t')) {
            i++;
		}
        if(line[i] == '"'){
            i++;
            quoted = true;
        }
        if(quoted){
            while(line[i] != 0 && line[i] != '\n' && line[i] != '"'){
                tmp_arg[j++] = line[i++];
            }
            if(line[i] == '"')
                i++;
        } else {
            while(line [i] != 0 && line[i] != '\n' && 
                  line[i] != ' ' && line[i] != '\t'){
                tmp_arg[j] = line[i];
                i++;
                j++;
            }
        }
        tmp_arg[j]=0;
        quoted = false;
 
        if(!strcmp(tmp_arg,"")){
            (*argv)[*argc] = 0;
            break;
        }
        else{
            (*argv)[*argc] = (char*) NestMalloc(strlen(tmp_arg)+1);
            strcpy((*argv)[*argc],tmp_arg);
        }
    }
}

void
NestFreeArgv( char **argv, int argc )
{
    if(argv) {
        for( int i=0;  i<argc;  i++ ) {
            if( argv[i] ) {
                NestFree( argv[i] );
			}
		}
        NestFree( argv );
    }
}


size_t
NestStrlen( const char *s )
{
    return strlen( s );
}

int
NestSnprintf( char *buf, int len, const char *fmt, ... )
{

    va_list ap;
    int n;

    va_start(ap, fmt);
    n = vsnprintf (buf, len, fmt, ap);
    va_end(ap);

    /*
    nest_debug( NEST_DEBUG_SERVER,
                "NestSnprintf called - %d %s!\n",
                len,
                fmt );
    */

    if ((n < 0) || (n >= len)) {
        //nest_debug( NEST_DEBUG_SERVER, "NestSnprintf returns error!\n" );
        if (buf) {
            *buf = 0;
        }
        return -1;
    }

    /*
    nest_debug( NEST_DEBUG_SERVER,
                "NestSnprintf returns %s!\n",
                buf );
    */

    return n;
}

char*
NestMakeString(int initial_size, const char *fmt, ...)
{
    /* Guess we need no more than 100 bytes. */
    int n, size = initial_size;
    char *p;
    va_list ap;
    if ((p = (char *) NestMalloc (size)) == NULL) {
        return NULL;
    }

    while (1) {
        /* Try to print in the allocated space. */
        va_start(ap, fmt);
        n = vsnprintf (p, size, fmt, ap);
        va_end(ap);
        /* If that worked, return the string. */
        if (n > -1 && n < size) {
            return p;
        }
        /* Else try again with more space. */
        if (n > -1) {       /* glibc 2.1 */
            size = n+1;     /* precisely what is needed */
        } else    {           /* glibc 2.0 */
            size *= 2;      /* twice the old size */
        }
        if ((p = (char *) NestRealloc (p, size)) == NULL) {
            return NULL;
        }
    }
}

void *
NestMemset( void *dst, int c, size_t n )
{
    return memset( dst, c, n );
}

void *
NestMemcpy( void *dst, const void *src, size_t n )
{
    return memcpy( dst, src, n );
}

void
NestFreeNormal( void *ptr )
{
    assert( ptr != NULL );
    LOCK_MUTEX( &nest_malloc_mutex );
    free( ptr );
    UNLOCK_MUTEX( &nest_malloc_mutex );
}

void
NestFreeDebug( void *ptr, char *func, int line )
{
    nest_debug( NEST_DEBUG_MALLOC,
				"Freeing 0x%x at %s:%d\n", ptr, func, line );
    NestFreeNormal( ptr );
}

void *
NestCallocNormal(size_t size)
{
    void *retptr = NestMalloc(size);
    memset( retptr, 0, size );
    return retptr;
}                                                        

void *
NestCallocDebug(size_t size, char *func, int line )
{
    void *ptr = NestCallocNormal( size );
    nest_debug( NEST_DEBUG_MALLOC, 
            "Calloced %ld bytes at 0x%x from %s:%d.\n", size, ptr, func, line);
    return ptr; 
}

void *
NestMallocDebug( size_t size, char *func, int line ) {
    void *vptr = NestMallocNormal( size );
    nest_debug( NEST_DEBUG_MALLOC, 
        "Mallocced %ld at 0x%x from %s:%d.\n", size, vptr, func, line );
    return vptr;
}

void *
NestMallocNormal(size_t size)
{
  assert(size > 0);
  LOCK_MUTEX( &nest_malloc_mutex );
  void *retptr = malloc(size);
  UNLOCK_MUTEX( &nest_malloc_mutex );
  if(retptr == NULL){
    nest_die( "failed to malloc %d bytes: %s\n", size, strerror( errno ) );
  }
  return retptr;
}

void *
NestReallocNormal(void *ptr, size_t size)
{
    LOCK_MUTEX( &nest_malloc_mutex );
    void *retptr = realloc( ptr, size );
    UNLOCK_MUTEX( &nest_malloc_mutex );
    return retptr;
}

void *
NestReallocDebug( void *ptr, size_t size, char *func, int line )
{
    void *vptr = NestReallocNormal( ptr, size );
    nest_debug( NEST_DEBUG_MALLOC,
            "Reallocced 0x%x to grow 0x%x by %d %s:%d\n",
            vptr, ptr, size, func, line );
    return vptr;
}

char *
NestStrdupDebug( const char *src, char *func, int line )
{
    char *ret = NestStrdupNormal( src );
    nest_debug( NEST_DEBUG_MALLOC, 
        "Strduped %d at 0x%x from %s:%d.\n", 
            strlen(ret), (void*)ret, func, line );
    return ret;       
}

char *
NestStrdupNormal( const char *src )
{
    assert( src != NULL );
    LOCK_MUTEX( &nest_strdup_mutex );
    char *ret = strdup( src );
    UNLOCK_MUTEX( &nest_strdup_mutex );
    if ( ret == NULL ) {
        nest_die( "Couldn't strdup %s: %s\n", src, strerror( errno ) );
    }
    return ret;
}

// File descriptor handling
int
NestLogMessage( int fd, char * msg )
{

    struct timeval tv;
    char * now;
    
    NestGettimeofday( &tv );
    now = ctime( (time_t *) &tv.tv_sec );
    if ( NestWriten( fd, now, NestStrlen( now ) ) < 0 ) return ERROR_EXIT;
    if ( NestWriten( fd, "  ", 2 ) < 0 )             return ERROR_EXIT;
    if ( NestWriten( fd, msg, NestStrlen( msg ) ) < 0 ) return ERROR_EXIT;      

    return NEST_SUCCESS;
}

int
NestClose( int sockfd )
{
    int ret = close( sockfd );
    if ( ret < 0 ) {
        nest_warn( __FILE__, __FUNCTION__, "close of %d: %s\n", sockfd, strerror( errno ) );
    }
    nest_debug( NEST_DEBUG_FD, 
				"Closed %d: %s\n", sockfd, 
				( ( ret < 0 ) ? strerror( errno ) : "success" ) );
    return ret;
}

// returns length of file, or -1 if filename does not refer to a valid file
nestFileSize
NestFilelength( const char *filename )
{
    
    struct stat buf;

	// TODO
    if ( ( stat( filename, &buf ) ) < 0 || ! S_ISREG( buf.st_mode ) ) {
        return -1;
    }

    return buf.st_size;
}

int
NestHandle_to_string(const NestLotHandle *handle,
                     char *str,
                     int len)
{
    if (!handle || !str) {
        return -1;
    }

    rsv_id_t    tmp;

    if (NestGet_id_from_handle( handle, &tmp ) < 0) {
        return -1;
    }

    NestSnprintf( str, len, "%lu", tmp );
    str[len - 1] = 0;

    return 0;
}


int
NestString_to_handle(NestLotHandle *handle,
                     const char *str)
{
    if (!handle || !str) {
        return -1;
    }

    rsv_id_t    tmp;
    sscanf( str, "%lu", &tmp );

    if (NestAdd_id_to_handle( handle, tmp ) < 0) {
        return -1;
    }

    return 0;
}

int
NestAdd_id_to_handle(NestLotHandle   *handle,
                     rsv_id_t rsv_id)
{
    if (!handle) {
        return -1;
    }
    handle->rsv_id = rsv_id;
    return 0;
}

int
NestGet_id_from_handle(const NestLotHandle   *handle,
                       rsv_id_t *rsv_id)
{

    if (!handle || !rsv_id) {
        if (rsv_id) {
            *rsv_id = 0;
        }
        return -1;
    }

    *rsv_id = handle->rsv_id;
    return 0;
}

/*
returns -1 on error
*/
ssize_t
NestWriten( int fd, const void *vptr, ssize_t n )
{
    ssize_t nleft;
    ssize_t nwritten;
    const char *ptr;

    /*  this is old stupidness from when I did test sends and receives
        // if the passed descriptor is negative, then NestWriten is a no-op
    if ( fd < 0 ) return n;
    */

    ptr = (char *)vptr;
    nleft = n;

    while ( nleft > 0 ) {
        if ( ( nwritten = NestWrite( fd, ptr, nleft ) ) < 0 ) {
            return -1;
        }
        nleft -= nwritten;
        ptr += nwritten;
    }

    return n;
}

ssize_t
NestWrite( int fd, const void *vptr, ssize_t n )
{
    ssize_t nwritten;

    if ( ( nwritten = write( fd, vptr, n ) ) < 0 ) {
        if ( errno == EINTR ) {
            return 0;
        } else { 
            return -1;
        }
    } 

    return nwritten;
}


// returns -1 on error
ssize_t
NestReadn( int fd, void *vptr, ssize_t n )
{
    ssize_t nleft;
    ssize_t nread;
    char *ptr;

         // if the passed descriptor is negative, then NestReadn is a no-op
    if ( fd < 0 ) return n;

    ptr = (char *) vptr;
    nleft = n;
    while( nleft > 0 ) {
        if ( ( nread = NestRead( fd, ptr, nleft ) ) < 0 ) {
            return -1;
        } else if ( nread == 0 ) {
                // is this right?  isn't it possible just no bytes right now?
            break;        // EOF    
        }

        nleft -= nread;
        ptr   += nread;
    }

    return( n - nleft );
}

ssize_t
NestRead( int fd, void *vptr, ssize_t n )
{
    ssize_t nread;

    while ( ( nread = read( fd, vptr, n ) ) < 0 ) {
        if ( errno == EINTR ) {
            continue;    
        } else { 
            return -1;
        }
    }

    return nread;
} 

int
NestOpen( const char* pathname, int flags, mode_t mode )
{
    int fd = open( pathname, flags, mode );
    /*
    if ( fd < 0) {
        nest_warn( __FILE__, __FUNCTION__, "open %s: %s\n", pathname, strerror( errno ) );
    }
    */
    nest_debug( NEST_DEBUG_FD,
				"Opened %d for %s: %s\n", fd, pathname,
				( ( fd < 0 ) ? strerror( errno ) : "success" ) );
    return fd;
}

int 
NestDup( int filedes )
{
    int retval = dup( filedes );
    if ( retval < 0 ) {
        nest_perror( "dup" );
        assert( 0 );
        return ERROR_EXIT;
    }
    nest_debug( NEST_DEBUG_FD,
				"Duped %d from %d.\n", retval, filedes );
    return retval;
}

int 
NestDup2( int filedes, int filedes2 )
{
    int retval = dup2( filedes, filedes2 );
    if ( retval < 0 ) {
        nest_perror( "dup2" );
        assert( 0 );
        return ERROR_EXIT;
    }
    nest_debug( NEST_DEBUG_FD, 
				"Dup2ed %d from %d.\n", retval, filedes );
    return retval;
}

NestReplyStatus
NestConnect( int sockfd, struct sockaddr *serv_addr, int addrlen )
{
    if ( connect( sockfd, serv_addr, addrlen ) < 0 ) {
        nest_debug( NEST_DEBUG_PROXY,
					"Can't connect: %s.\n", strerror(errno) );
        return ErrnoToNest( errno );
    }
    return NEST_SUCCESS;
}

int
NestsetSocket( int domain, int type, int protocol, int port )
{

    static sockaddr_in sin;
    int sd;
    int on = 1;
    unsigned addr_sz = sizeof( sin );
    
    sd = NestSocket( domain, type, protocol );

    memset( &sin, 0, sizeof( sockaddr_in ) );
    sin.sin_family = domain; 
    sin.sin_addr.s_addr = htonl( INADDR_ANY );
    sin.sin_port = htons( (unsigned short) port );

        // maybe we need to set TCP buffer sizes ??
    NestSetsockopt( sd, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof( on ) );
    if ( NestBind( sd, (sockaddr *)&sin, sizeof( struct sockaddr_in ) ) != 0 ) {
        return -1;
    }
    NestGetsockname( sd, (sockaddr *)&sin, (socklen_t *) &addr_sz );
    NestListen( sd, BACKLOG );
    return sd;
}

void
NestListen( int sockfd, int backlog )
{
    if ( listen( sockfd, backlog ) < 0 ) {
        nest_perror( "listen" );
        exit( ERROR_EXIT );
    }
}

int
NestBind( int sockfd, const struct sockaddr *addr, socklen_t addrlen )
{
    if ( bind( sockfd, addr, addrlen ) < 0 ) { 
        nest_perror( "bind" );
        return -1;    
    }
    return 0;
}

void
NestSetsockopt( int sockfd, int level, int name, void *val, socklen_t len )
{
    if ( setsockopt( sockfd, level, name, (char *)val, len ) < 0 ) {
        nest_perror( "setsockopt" );
    }
}

int
NestSocket( int family, int type, int protocol )
{
    int sockfd;
    if ( ( sockfd = socket( family, type, protocol ) ) < 0 ) { 
        nest_perror( "socket" );
        return -1;    
    }
    return sockfd;
}

struct hostent*
NestGethostbyname( const char *hostname )
{
    struct hostent *host = NULL;
    struct hostent *tmp  = NULL;
    LOCK_MUTEX( &nest_common_mutex );
    tmp = gethostbyname( hostname );
    if ( tmp == NULL ) {
        nest_warn( __FILE__, __FUNCTION__, "gethostbyname %s: %s\n",
				   hostname, strerror( errno ) );
    } else {
        host = (struct hostent *)NestMalloc( sizeof( *tmp ) );
        NestMemcpy( host, tmp, sizeof ( *host ) );
    }
    UNLOCK_MUTEX( &nest_common_mutex );
    return host;
}

void
NestGetsockname( int sockfd, struct sockaddr *localaddr, socklen_t *addrlen )
{
    if ( getsockname( sockfd, localaddr, addrlen ) < 0 ) {
        nest_perror( "getsockname" );
        exit( ERROR_EXIT );
    }
}

char *
NestGetUserName( char *user )
{
    static char *result = NULL;
    static bool init = false;
    LOCK_MUTEX( &nest_common_mutex );
    if ( ! init ) {
        init = true;
        struct passwd *pass = getpwuid( geteuid() );
        if(pass) {
            result = NestStrdup( pass->pw_name );
        } else {
            result = 0;
        }
    }
    UNLOCK_MUTEX( &nest_common_mutex );
    if ( result ) {
        strcpy( user, result );
    }
    return result;
}

void
NestGettimespec( struct timespec *ts )
{
  struct timeval tv;
  NestGettimeofday( &tv );

  ts->tv_sec = tv.tv_sec;
  ts->tv_nsec = tv.tv_usec*1000;
}

void
NestGettimeofday( struct timeval *time ) {
    if ( gettimeofday( time, 0 ) < 0 ) {
        nest_perror( "gettimeofday" );
        assert( 0 );
    }
}

time_t
NestTime(time_t *tsecs) {
    time_t retval;
    if ((retval = time(tsecs)) == ((time_t)-1)) {
        nest_perror("time");
        assert( 0 );
    }
   return retval;
}

long
NestElapsedTime( const struct timeval *start )
{
    struct timeval now;
    NestGettimeofday( &now );
    return NestTimediff( start, &now, __FUNCTION__ );
}

/* returns microseconds between the interval */
long
NestTimediff( const struct timeval *start,
			  const struct timeval *end,
			  const char *function )
{

    long long micros = (long long)end->tv_sec * MILLION + end->tv_usec;
    micros -= ((long long)start->tv_sec * MILLION + start->tv_usec);
    /*
    long millis = micros / THOUSAND;

    if ( millis == 0 && micros != 0 ) {
        nest_warn( __FILE__, __FUNCTION__, 
        "Change NestTimediff to return micros:  %ld != %ld / 1000\n",
                millis, micros );
    }
    return ( millis );
    */
    long ret_micros = (long)micros;
    if ( ret_micros < 0 ) {
        nest_warn( __FILE__, __FUNCTION__, 
                "Returning negative timediff from %s: %ld\n"
                "\tstart %ld.%ld\n"
                "\tend   %ld.%ld\n", 
                function,
                ret_micros,
                start->tv_sec, start->tv_usec,
                end->tv_sec,   end->tv_usec );
    }
    return ret_micros; 
}

double
NestBandwidth( nestFileSize bytes, 
			   struct timeval *start,
			   struct timeval *end,
			   const char *function )
{
    return NestBandwidth( bytes, NestTimediff( start, end, function ) );
}

double
NestBandwidth( nestFileSize bytes,
			   long micros )
{

    if ( bytes <= 0 || micros <= 0 ) {
        if ( bytes < 0 || micros < 0 ) {
            nest_warn( __FILE__, __FUNCTION__,
					   "Asked to compute negative bw: %ld bytes, %ld micros\n",
					   bytes, micros );
            return -1;
        } else {
            return 0;
        }
    }

    /*
    float secs;
    float MBs;
    secs = (float) micros / ( THOUSAND * THOUSAND );
    MBs  = (float) bytes / ( 1024 * 1024 );

    return MBs / secs;  
    */
        // do one less division, compiler should do constant for us
    double constant = ( (double)THOUSAND * THOUSAND ) / ( 1024.0 * 1024.0 );
    double bw = ( (double)bytes / micros ) * constant;
    return bw;
}

// used by nest_get_put and by nest_client_exe
NestTransferInfo *
NestmakeTransferInfo( )
{
    NestTransferInfo *trans = (NestTransferInfo *)
            NestCalloc( sizeof ( NestTransferInfo ) );
    NestGettimeofday( &trans->starttime );
    memcpy(&trans->lastprogress, &trans->starttime, sizeof(struct timeval));
    return trans;
}

int
NestString_to_NestFileInfo(NestFileInfo *info,
                           const char *msg)
{
	int		num = 0;

    if (!info || !msg) {
        return -1;
    }

    char    tmp_buf[NEST_NAMELEN];

	// Scan out & copy the name
	int tmp;
    tmp = sscanf( msg,
				  "%s %s",
				  info->name,
				  tmp_buf );
	if ( tmp < 2 ) {
		return -1;
    }
	num++;

    int src		= NestStrlen(info->name) + NestStrlen(tmp_buf) + 1;
	int dest	= 0;

	// Copy the lot list
    while ((msg[src] != '\n') && (src < NEST_MAXPATHLEN)) {
        info->LotList[dest++] = msg[src++];
    }
	num++;

    info->LotList[dest] = 0;
	int		ftype;
	info->atime = info->ctime = info->mtime = (int) time( NULL );
	tmp = sscanf( tmp_buf,
				  "%ld:%d:%d:%d:%d",
				  &info->size,
				  &ftype,
				  &info->atime, &info->ctime, &info->mtime );
	if ( tmp < 1 ) {
		return -1;
	} else if ( tmp < 2 ) {
		info->type = 0;
	} else {
		info->type = (nest_obj_type_t) ftype;
	}
	num += tmp;
	
    return num;
}

/**
* takes a pathname and keeps moving forward through it until it finds the
* final '/' and returns a ptr to the filename
*/
char* 
NestExtractFilename( char *pathname )
{
    char *ptr, *tmp;

    ptr = pathname;
    tmp = strchr( pathname, '/' );

    while( tmp != NULL ) {
        ptr = ++tmp;
        tmp = strchr( ptr, '/' );
    }    

    return ptr;
}


// Parse a size string
NestReplyStatus
NestParseSizeString( const char *arg, long long *size, char default_mod )
{
	nestLotSize  bytes;
	double		 dbytes;
	char         mod = default_mod;
	bool         isint;

	if ( strchr(arg, '.' ) ) {
		sscanf( arg, "%lf%c", &dbytes, &mod );
		isint = false;
	} else {
		sscanf( arg, "%lld%c", &bytes, &mod );
		isint = true;
	}

	switch( toupper( mod ) ) {
	case '\0':
	case 'B':
		// Do nothing
		break;
	case 'K':
		if ( isint )
			bytes *= 1024LL;
		else
			dbytes *= 1024.0;
		break;
	case 'M':
		if ( isint )
			bytes *= ( 1024LL * 1024LL );
		else
			dbytes *= ( 1024.0 * 1024.0 );
		break;
	case 'G':
		if ( isint )
			bytes *= ( 1024LL * 1024LL * 1024LL );
		else
			dbytes *= ( 1024.0 * 1024.0 * 1024.0 );
		break;
	case 'T':
		if ( isint )
			bytes *= ( 1024LL * 1024LL * 1024LL * 1024LL );
		else
			dbytes *= ( 1024.0 * 1024.0 * 1024.0 * 1024.0 );
		break;
	default:
		return NEST_BAD_PARAMETERS;
	}

	// If they specified the floating value, convert it to integer..
	if ( !isint ) {
		bytes = (nestLotSize) dbytes;
	}

	*size = bytes;
	return NEST_SUCCESS;
}

NestReplyStatus
NestParseTimeString( const char *arg, int *seconds, char default_mod )
{
	int      sec;
	double   dsec;
	char     mod = default_mod;
	bool     isint;

	if ( strchr(arg, '.' ) ) {
		sscanf( arg, "%lf%c", &dsec, &mod );
		isint = false;
    } else if ( strchr(arg, ':' ) ) {
		int	a1 = 0, a2 = 0, a3 = 0, a4 = 0, a5 = 0, args;
		args = sscanf( arg, "%d:%d:%d:%d:%d", &a1, &a2, &a3, &a4, &a5 );
		switch( args ) {
		case 0:
			return NEST_BAD_PARAMETERS;
		case 1:
			sec = a1;
			break;
		case 2:
			sec = ( ( a1 * 60 ) +
					a2 );
			break;
		case 3:
			sec = ( ( a1 * 60 * 60 ) +
					( a2 * 60 ) +
					a3 );
			break;
		case 4:
			sec = ( ( a1 * 60 * 60 * 24 ) +
					( a2 * 60 * 60 ) +
					( a3 * 60 ) +
					a4 );
			break;
		case 5:
			sec = ( ( a1 * 60 * 60 * 60 * 24 * 365 ) +
					( a2 * 60 * 60 * 24 ) +
					( a3 * 60 * 60 ) +
					( a4 * 60 ) +
					a5 );
			break;
		}
		*seconds = sec;
		return NEST_SUCCESS;
	} else {
		sscanf( arg, "%d%c", &sec, &mod );
		isint = true;
	}

	switch( toupper( mod ) ) {
	case '\0':
	case 'S':
		// Do nothing
		break;
	case 'M':
		if ( isint )
			sec *= 60;
		else
			dsec *= 60.0;
		break;
	case 'H':
		if ( isint )
			sec *= ( 60 * 60 );
		else
			dsec *= ( 60.0 * 60.0 );
		break;
	case 'D':
		if ( isint )
			sec *= ( 24 * 60 * 60 );
		else
			dsec *= ( 24.0 * 60.0 * 60.0 );
		break;
	case 'Y':
		if ( isint )
			sec *= ( 365 * 24 * 60 * 60 );
		else
			dsec *= ( 365.0 * 24.0 * 60.0 * 60.0 );
		break;
	default:
		return NEST_BAD_PARAMETERS;
	}

	// If they specified the floating value, convert it to integer..
	if ( !isint ) {
		sec = (int) dsec;
	}

	*seconds = sec;
	return NEST_SUCCESS;
}

const char *
NestFormatSizeString( long long size, char *buf, int bufsize, bool raw )
{
	if ( raw ) {
		if ( size >= 1024LL * 1024LL * 1024LL * 1024LL ) {
			snprintf( buf, bufsize, "%.2f TB (%lld)",
					  size/(1024.0*1024.0*1024.0*1024.0), size );
		} else if ( size >= 1024LL * 1024LL * 1024LL ) {
			snprintf( buf, bufsize, "%.2f GB (%lld)",
					  size/(1024.0*1024.0*1024.0), size );
		} else if (size >= 1024LL * 1024LL ) {
			snprintf( buf, bufsize, "%.2f MB (%lld)",
					  size/(1024.0*1024.0), size );
		} else if (size >= 1024LL ) {
			snprintf( buf, bufsize, "%.2f KB (%lld)",
					  size/(1024.0), size );
		} else {
			snprintf( buf, bufsize, "%lld (%lld)", size, size );
		}
	} else {
		if ( size >= 1024LL * 1024LL * 1024LL * 1024LL ) {
			snprintf( buf, bufsize, "%.2f TB",
					  size/(1024.0*1024.0*1024.0*1024.0) );
		} else if ( size >= 1024LL * 1024LL * 1024LL ) {
			snprintf( buf, bufsize, "%.2f GB",
					  size/(1024.0*1024.0*1024.0) );
		} else if (size >= 1024LL * 1024LL ) {
			snprintf( buf, bufsize, "%.2f MB",
					  size/(1024.0*1024.0) );
		} else if (size >= 1024LL ) {
			snprintf( buf, bufsize, "%.2f KB",
					  size/(1024.0) );
		} else {
			snprintf( buf, bufsize, "%lld", size );
		}
	}
	return buf;
}

const char *
NestFormatTimeString( int seconds, char *buf, int bufsize, bool raw )
{
	int sec = seconds % 60;
	int	minutes = ( seconds / 60 ) % 60;
	int	hours = ( seconds / (60*60) ) % 24;
	int days = ( seconds / (24*60*60) ) % 365;
	int years = seconds / (24*60*60*365);
	if ( raw ) {
		if ( years ) {
			snprintf( buf, bufsize, "%d years, %d days, %02d:%02d:%02d",
					  years, days, hours, minutes, sec );
		} else if ( days ) {
			snprintf( buf, bufsize, "%d days, %02d:%02d:%02d",
					  days, hours, minutes, sec );
		} else if ( hours ) {
			snprintf( buf, bufsize, "%02d:%02d:%02d",
					  hours, minutes, sec );
		} else if ( minutes ) {
			snprintf( buf, bufsize, "%02d:%02d",
					  minutes, sec );
		} else {
			snprintf( buf, bufsize, "%ds",
					  sec);
		}
	} else {
		if ( years ) {
			snprintf( buf, bufsize, "%d years, %d days, %02d:%02d:%02d (%ds)",
					  years, days, hours, minutes, sec, seconds );
		} else if ( days ) {
			snprintf( buf, bufsize, "%d days, %02d:%02d:%02d (%ds)",
					  days, hours, minutes, sec, seconds );
		} else if ( hours ) {
			snprintf( buf, bufsize, "%02d:%02d:%02d (%d)",
					  hours, minutes, sec, seconds );
		} else if ( minutes ) {
			snprintf( buf, bufsize, "%02d:%02d (%d)",
					  minutes, sec, seconds );
		} else {
			snprintf( buf, bufsize, "%ds",
					  seconds );
		}
	}
	return buf;
}
