/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_storage_manager.h"
#include "nest_protocol_manager.h"
#include "nest_protocol_filesystem.h"
#include "nest_config.h"
#include "nest_pwent.h"


NestReplyStatus
parseStorageSpace( const char *arg, nestLotSize *space )
{
	nestLotSize  bytes;
	double		 dbytes;
	char         mod = '\0';
	bool         isint;

	if ( strchr(arg, '.' ) ) {
		sscanf( arg, "%lf%c", &dbytes, &mod );
		isint = false;
	} else {
		sscanf( arg, "%lld%c", &bytes, &mod );
		isint = true;
	}

	switch( toupper( mod ) ) {
	case 'B':
		// Do nothing
		break;
	case 'K':
		if ( isint )
			bytes *= 1024LL;
		else
			dbytes *= 1024.0;
		break;
	case '\0':
	case 'M':
		if ( isint )
			bytes *= ( 1024LL * 1024LL );
		else
			dbytes *= ( 1024.0 * 1024.0 );
		break;
	case 'G':
		if ( isint )
			bytes *= ( 1024LL * 1024LL * 1024LL );
		else
			dbytes *= ( 1024.0 * 1024.0 * 1024.0 );
		break;
	case 'T':
		if ( isint )
			bytes *= ( 1024LL * 1024LL * 1024LL * 1024LL );
		else
			dbytes *= ( 1024.0 * 1024.0 * 1024.0 * 1024.0 );
		break;
	default:
		return NEST_BAD_PARAMETERS;
	}

	// If they specified the floating value, convert it to integer..
	if ( !isint ) {
		bytes = (nestLotSize) dbytes;
	}

	*space = bytes;
	return NEST_SUCCESS;
}


nestStorageManager::nestStorageManager( void )
{
	storage_space = -1;
	file_create_mode = FILE_MODE;

	userlist = "";
	denylist = "";

	Resources = NULL;
	rootDir = NULL;

}

nestStorageManager::~nestStorageManager( void )
{
}

int
nestStorageManager::init( void )
{

    bool recovery = true;

    const char *StorageDir = NULL;
    rootDir = NULL;

        // needed for NFS
    fh_hash = nest_hash_table_create( 1037, NULL );
    assert( fh_hash != NULL );

    /*
        Must have all these inits that initializes various global
        variables before any of the ->initialize() s are called!
    */

    const char *value;

    if ( GetConfigValue(Config_StorageDir, &value) && *value != 0) {
        StorageDir = value;
    } else {
        nest_die( "Must have StorageDir defined.\n" );
    }

    if ( GetConfigValue(Config_DisableLots, NULL)) {
        NestEnableLots = false;
    }

    if ( GetConfigValue(Config_StorageSpace, &value) && *value != 0) {
		parseStorageSpace( value, &storage_space );
    }

    if ( GetConfigValue(Config_UserList, &value) && *value != 0) {
        const char *identity_list = NestGetModifiedIdentity(value);
        userlist = NestStrdup(identity_list);
    }
    if ( GetConfigValue(Config_DenyList, &value) && *value != 0) {
        const char *identity_list = NestGetModifiedIdentity(value);
        denylist = NestStrdup(identity_list);
    }

    if ( GetConfigValue(Config_AdminUser, &value) && *value != 0) {
        NestAdminUser = NestGetModifiedIdentity(value);
    }

    if ( GetConfigValue(Config_SystemUser, &value) && *value != 0) {
        NestSystemUser = value;
    }

    if ( GetConfigValue(Config_AnonymousUser, &value) && *value != 0) {
        NestAnonymousUser = value;
    }

    if ( GetConfigValue(Config_PasswordFile, &value) && *value != 0) {
        SystemPwdFile = value;
    }

    if ( GetConfigValue(Config_GroupFile, &value) && *value != 0) {
        SystemGrpFile = value;
    }

    if ( GetConfigValue(Config_GridMapFile, &value) && *value != 0) {
        GridMapFile = value;
    }

    if ( GetConfigValue(Config_MetaDir, &value) && *value != 0) {
        MetaDir = value;
    }

    if ( GetConfigValue(Config_ActualGroup, &value) && *value != 0) {
        ActualGroup = value;
    }

    if ( GetConfigValue(Config_FileCreateMode, &value) && *value != 0) {
        file_create_mode = (unsigned)strtol( value, NULL, 8 );
    }

    AdminUid = NestGetuid();
    if (strcmp(NestAdminUser, "")) {
		nestPwent	pwent;
		if ( ! pwent.findent( AdminUid, SystemPwdFile ) ) {
            AdminUid = pwent.getent_uid( );
        }
    }

    ActualGid = NestGetgid();
    if (strcmp(ActualGroup, "")) {
        group_ent   ge;
        if (Nestgetgrpnam( ActualGroup, &ge, SystemGrpFile ) >= 0) {
            ActualGid = ge.grp_gid;
        }
    }

    // make and move into the storage (root) dir
    nest_debug( NEST_DEBUG_STORAGE, "StorageDir is %s\n", StorageDir );
    if ( NestMkdir( StorageDir, DIR_MODE ) < 0 ) {
        if ( errno != EEXIST ) {
            nest_die( "mkdir %s: %s\n", StorageDir, strerror( errno ) );
        }
    }
    NestChmod( StorageDir, 0777 );
    NestChown( StorageDir, AdminUid, ActualGid );
    NestChdir( StorageDir );

        // remember what the rootDir is
    char buffer[NEST_MAXPATHLEN];
    if ( ( NestGetcwd( buffer, NEST_MAXPATHLEN ) ) != NEST_SUCCESS ) {
        nest_die( "getcwd failed" );
    }
    rootDir = NestStrdup( buffer );
    nest_debug( NEST_DEBUG_ALWAYS, "Storage Manager's cwd is %s\n", rootDir );

        // create our meta data dir
    nest_debug( NEST_DEBUG_STORAGE, "MetaDir is %s\n", MetaDir );
    if ( NestMkdir( MetaDir, DIR_MODE ) < 0 ) {
        if ( errno != EEXIST ) {
            nest_perror( "mkdir" );
            nest_die( "mkdir failed" );
        }
    }
    NestChmod( MetaDir, DIR_MODE );
    NestChown( MetaDir, AdminUid, ActualGid );

    //  lots won't work unless we are run as root
    /*  This is not true anymore! 
    if ( NestGetuid() != 0 ) {
        NestEnableLots = false;
    }
    */

    //Resources = new ClassAdCollectionServer();
    ClassAdCollection *r = new ClassAdCollection();
    Resources = r; 
    if ( Resources == NULL ) {
        nest_die("Couldn't instantiate a ClassAdCollection!\n");
    }
    query.Bind(Resources);

    string  log_file = "";
    log_file = log_file + MetaDir + "/" + CLASSAD_LOGFILE;

    nest_stat   tmp_stat_buf;

    if (NestStat( log_file.c_str(), &tmp_stat_buf ) < 0) {
        recovery = false;
    }

    // see whether required system-files are present!

    if (NestStat( SystemPwdFile, &tmp_stat_buf ) < 0) {
        nest_die("Couldn't find the password file at %s\n",
                 SystemPwdFile);
    }

    if (NestStat( SystemGrpFile, &tmp_stat_buf ) < 0) {
        nest_die("Couldn't find the group file at %s\n",
                 SystemGrpFile);
    }

    if (NestStat( GridMapFile, &tmp_stat_buf ) < 0) {
        nest_die("Couldn't find the grid-mapfile at %s\n",
                 GridMapFile);
    }

    if (!Resources->InitializeFromLog(log_file, "", "" )) {
		nest_debug( NEST_DEBUG_ALWAYS, "ClassAd error: %s\n",
					CondorErrMsg.c_str() );
        nest_die("Couldn't recover ClassAdCollection from log!\n");
    }
    NestChmod( log_file.c_str(), FILE_MODE );
    NestChown( log_file.c_str(), AdminUid, ActualGid );

    NestReplyStatus reply;


	// Setup & initialize the identity manager
	identity_manager->set_userlist( userlist );
	identity_manager->set_denylist( denylist );

    if (recovery) {
        reply = identity_manager->recover(Resources);
        if ( reply != NEST_SUCCESS ) {
            nest_die( "IdentityManager 'recover' failed! - %s\n",
                        NestErrorString(reply) );
        }
    } else {
        reply = identity_manager->initialize(Resources);
        if ( reply != NEST_SUCCESS ) {
            nest_debug( NEST_DEBUG_ALWAYS,
                        "IdentityManager 'initialize' failed! - %s\n",
                        NestErrorString(reply) );
            nest_die( "IdentityManager 'initialize' failed!\n" );
        }
    }

	// Setup the lot manager
    lot_manager->setup( rootDir, storage_space );

    /*
        dont bother about lot_manager::init or recovery is its disabled!
    */
    if (NestEnableLots) {
        lot_manager->init( );
        if (recovery) {
            if ( (reply = lot_manager->recover(Resources)) != NEST_SUCCESS ) {
                nest_debug( NEST_DEBUG_ALWAYS,
                            "LotManager 'recover' failed - %s\n",
                            NestErrorString(reply) );
                nest_warn( __FILE__, __FUNCTION__,
                        "No LotManager => disabling lots\n" );
                if (reply == NEST_NO_QUOTA_ENFORCEMENT) {
                    nest_warn( __FILE__, __FUNCTION__,
                            "The quota system has been disabled.\n"
                              "Enable quotas using \"$/sbin/quotaon -a\" "
                              "to enable lots!\n" );
                }
                nest_debug( NEST_DEBUG_ALWAYS,
                            "Error returned is %s\n", NestErrorString(reply));
                NestEnableLots = false;
            }
        } else {
            if ( (reply = lot_manager->initialize(Resources))
                        != NEST_SUCCESS ) {
                nest_debug( NEST_DEBUG_ALWAYS,
                            "LotManager 'initialize' failed - %s\n",
                            NestErrorString(reply) );
                nest_warn( __FILE__, __FUNCTION__,
                        "No LotManager => disabling lots\n" );
                if (reply == NEST_NO_QUOTA_ENFORCEMENT) {
                    nest_warn( __FILE__, __FUNCTION__,
                            "The quota system has been disabled.\n"
                            "Enable quotas using \"$/sbin/quotaon -a\" "
                            "to enable lots!\n" );
                }
                NestEnableLots = false;
            }
        }
    }

	// Initialize the file catalog
    if (recovery) {
        reply = file_catalog->recover(Resources, rootDir);
        if ( reply != NEST_SUCCESS ) {
            nest_debug( NEST_DEBUG_ALWAYS,
                        "FileCatalog 'recover' failed - %s\n",
                        NestErrorString(reply) );
            nest_die( "FileCatalog 'recover' failed!\n" );
        }
    } else {
        reply = file_catalog->initialize(Resources, rootDir);
        if ( reply != NEST_SUCCESS ) {
            nest_debug( NEST_DEBUG_ALWAYS,
                        "FileCatalog 'initialize' failed - %s\n",
                        NestErrorString(reply) );
            nest_die( "FileCatalog 'initialize' failed!\n" );
        }
    }

	// Create & initialize the collection manager.
    if (recovery) {
        reply = collection_manager->recover(Resources);
        if ( reply != NEST_SUCCESS ) {
            nest_die( "CollectionManager 'recover' failed! - %s\n",
                        NestErrorString(reply) );
        }
    } else {
        reply = collection_manager->initialize(Resources);
        if ( reply != NEST_SUCCESS ) {
            nest_debug( NEST_DEBUG_ALWAYS,
                        "CollectionManager 'initialize' failed! - %s\n",
                        NestErrorString(reply) );
            nest_die( "CollectionManager 'initialize' failed!\n" );
        }
    }

	// And, initialize the rest of the storage manager
    reply = initialize();
    if ( reply != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "StorageManager 'initialize' failed - %s\n",
                     NestErrorString(reply));
        nest_die( "StorageManager 'initialize' failed!\n" );
    }

    return 0;
}

NestReplyStatus
nestStorageManager::initialize()
{

    int     sync_status = synchronize( );
    if ( sync_status < 0 ) {
        nest_debug( NEST_DEBUG_ALWAYS, "Synchronization error\n" );
        return sync_status;
    }

    //  Got to take care of aborted trasfers in progress

    string  constraint = "other.TransferInProgress == true ";
    ExprTree *constraint_tree = parser.ParseExpression( constraint );
    if (!constraint_tree) {
        return NEST_CLASSAD_ERROR;
    }
    ViewName    file_view = NEST_FILE_CATALOG_VIEW;

    query.Query( file_view, constraint_tree );
	delete constraint_tree;
    query.ToFirst();

    string key;
    if (!query.Current(key)) {
        return NEST_SUCCESS;
    }

    ClassAd             *file_ad;
    nestProtocol        *storage_con;
    string              str_user;
    string              str_group;
    string              l_fname;		// Logical file name
    char                filename[NEST_MAXPATHLEN];
    bool                result;
    int                 count = 0;

    do {
        file_ad = Resources->GetClassAd( key );

        if (!file_ad) {
            continue;
        }

        result  = file_ad->EvaluateAttrString( "OwnerUser", str_user );
        result &= file_ad->EvaluateAttrString( "OwnerGroup", str_group );
        result &= file_ad->EvaluateAttrString( "LogicalName", l_fname );

        if (!result) {
            continue;
        }

        file_catalog->get_reverse_logical_name( filename, l_fname.c_str() );

        nest_debug( NEST_DEBUG_STORAGE,
                    "Transfer to file %s owned by %s was in progress during "
                    "the previous crash\n"
                    "Fixing it ...\n",
                    filename,
                    str_user.c_str() );
		nestGroup   tmp_group( &str_group );
		nestUser	tmp_user( &str_user );
		tmp_user.setUid( AdminUid );
		tmp_user.setGroup( &tmp_group );

        NestReplyStatus status;
        status = open( &storage_con,
					   &tmp_user,
                       filename,
                       O_WRONLY,
                       0,
                       -1,
                       "~/",
                       0 );
        if ( NEST_SUCCESS == status ) {
            close ( storage_con );

            delete storage_con;
        }
        storage_con = NULL;

        count ++;

    } while (query.Next(key));

    nest_debug ( NEST_DEBUG_STORAGE,
                 "%d transfers were in progress during the previous crash!\n",
                 count );
    return NEST_SUCCESS;
}

nestProtocol *
nestStorageManager::openMetaFile( char *file,
                                  int flags ) {

        // move into the meta dir
    if ( chMetaDir() < 0 ) return NULL;

        // open the file
    int fd = NestOpen( file, flags );
    NestChmod( file, FILE_MODE );
    NestChown( file, AdminUid, ActualGid );

        // move back to the rootDir
    if ( chRootDir() < 0 ) {
        nest_perror( "chdir" );
    }

        // wrap the descriptor in a protocol object
    return ( fd < 0 ) ? NULL : new nestProtocolFilesystem( fd );
}

void
nestStorageManager::writeAd( nestProtocol *dest )
{
    char buffer[NEST_MESSAGESIZE];

    long long int capacity, total;
    if ( bytesAvailable(IS_USER, &capacity, &total ) 
            == NEST_SUCCESS )
    {
        NestSnprintf( buffer, NEST_MESSAGESIZE, "Disk = %lld\nFree = %lld\n", 
                total, capacity);
        dest->Sendn( buffer, NestStrlen( buffer ) );
    }

        // now lots?
    if(nestStorageManager::is_lots_enabled())
        NestSnprintf( buffer, NEST_MESSAGESIZE, "LotsEnabled = true\n");
    else
        NestSnprintf( buffer, NEST_MESSAGESIZE, "LotsEnabled = false\n");
    dest->Sendn( buffer, NestStrlen( buffer ) );
}

NestReplyStatus
nestStorageManager::renameMetaFile( char *oldpath,
                                    char *newpath)
{

    if (!oldpath || !newpath) {
        return NEST_BAD_PARAMETERS;
    }

    if ( chMetaDir() < 0) {
        return ErrnoToNest( errno );
    }

    NestRename( oldpath, newpath );

    if (chRootDir() < 0) {
        nest_perror( "chdir" );
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::statMetaFile( char *file,
                                  nest_stat *st)
{

    if (!file) {
        return NEST_BAD_PARAMETERS;
    }

    if ( chMetaDir() < 0) {
        return NEST_UNKNOWN_ERROR;
    }

    NestLstat( file, st );

    if (chRootDir() < 0) {
        nest_perror( "chdir" );
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::unlinkMetaFile( char *file ) {

    if (!file) {
        return NEST_BAD_PARAMETERS;
    }

    if ( chMetaDir() < 0) {
        return NEST_UNKNOWN_ERROR;
    }

    NestUnlink( (const char *)file );

    if (chRootDir() < 0) {
        nest_perror( "chdir" );
    }

    return NEST_SUCCESS;
}


nestFileSize
nestStorageManager::metaFilelength( char *file ) {
        // move into meta dir
        // get the length, move back to rootDir, return length
    if ( chMetaDir() < 0 ) return -1;
    nestFileSize len = NestFilelength( file );
    if ( chRootDir() < 0 ) {
        nest_perror( "chdir" );
    }
    return len;
}

char *
nestStorageManager::metaFilepath( char *file ) {
    static char path[NEST_MESSAGESIZE];
    memset( path, 0, NEST_MESSAGESIZE );

        // make sure file exists
    if ( metaFilelength( file ) < 0 ) return NULL;

        // move into meta dir and read cwd
    if ( chMetaDir() < 0 ) return NULL;
    NestReplyStatus status = NestGetcwd( path, NEST_MESSAGESIZE );
    if ( status != NEST_SUCCESS ) {
        chRootDir();
        return NULL;
    }

        // move back to root dir
    if ( chRootDir() < 0 ) {
        nest_perror( "chdir" );
    }

        // add a slash and the filename
    NestSnprintf( &(path[NestStrlen(path)]), 
            NEST_MESSAGESIZE - NestStrlen(path),
            "/%s", file );
    return path;
}

char *
nestStorageManager::getInitialClientDir() {
    return "~/";
}

// Stupid function to generate an FDkey -- NOT re-entrant
const char *
gen_fdkey( const char *lfn, const nestUser *user, int flags )
{
	( void ) flags;
	static char fdkey[NEST_MAXPATHLEN * 2];
	int rval = NestSnprintf( fdkey,
							 NEST_MAXPATHLEN * 2,
							 "%s.%s",
							 user->getOsName( ),
							 lfn );
	if ( rval < 0 ) {
		return NULL;
	} else {
		return fdkey;
	}
}


/*
    For file management, access control is done here

    For user management, access control is done here

    For group management, access control is done inside the identity_manager.

    For Lot management, access control is done part here and part
                        inside the lot_manager
*/

NestReplyStatus
nestStorageManager::open( nestProtocol **connection,
						  const nestUser *user,
                          const char *file,
                          int flags,
                          off_t offset,
                          ssize_t filesize,
                          const char *cwd,
                          int link_count )
{
        // other variables
    NestReplyStatus reply;
    char         l_fname[NEST_MAXPATHLEN];
    bool         new_file_created = false;
    string       filename;
    bool         system_file = false;
    
    nest_debug( NEST_DEBUG_STORAGE,
                "opening '%s' - user '%s' - flags %o\n",
                file,
                user->getName(),
                flags );

    reply = expand_path(filename, user, cwd, file);
    NEST_REPLY_CHECK(reply);

    /*
        ignore return value - as while creating new files
        this would return NEST_FILE_NOT_FOUND
    */
    reply = file_catalog->get_logical_name( l_fname, filename.c_str(), user );
    system_file = file_catalog->is_system_file( l_fname );
    nest_debug( NEST_DEBUG_STORAGE, "'%s': system file = %d\n",
				l_fname, system_file );

        // see if we have this fd already cached
        // make sure to expand_path and get_logical_name first
        // if the fd is cached and we skip down, these must be set
    int        fd = -1;

	nestProtocolFilesystem *conn = NULL;
    char    p_fname[NEST_MAXPATHLEN];
	if ( fd < 0 ) {
		bool	file_exists = file_catalog->is_file_present ( l_fname );

        // it should be "O_RDONLY ==" and not "O_RDONLY &"- stupid
        // O_RDONLY is defined as '0' took me 3 hours of debugging to
        // fix this ://
		if (flags == O_RDONLY) {
			nest_debug( NEST_DEBUG_STORAGE, "RDONLY\n" );
			if ( file_exists ) {
				if (!file_catalog->access( l_fname, user, NEST_READ_ACCESS)) {
					return NEST_NONPRIVELEDGED_USER;
				}
			} else {
				if (file_catalog->paccess( l_fname, user, NEST_LIST_ACCESS)) {
					return NEST_FILE_NOT_FOUND;
				}
				return NEST_NONPRIVELEDGED_USER;
			}
		}

		if (flags & O_CREAT) {
			nest_debug( NEST_DEBUG_STORAGE, "CREAT\n" );
			if (!file_catalog->paccess( l_fname, user, NEST_INSERT_ACCESS)) {
				return NEST_NONPRIVELEDGED_USER;
			}

			if(!system_file) {
				reply = file_catalog->add_entry(l_fname, user, NEST_FILE);

				if (reply != NEST_SUCCESS && reply != NEST_FILE_EXISTS) {
					return reply;
				}
				new_file_created =  (reply == NEST_SUCCESS);
			} else {
				new_file_created = true;
			}
			file_exists = true;	// Ok, it exists now...
		}

		if (flags & O_WRONLY) {
			nest_debug( NEST_DEBUG_STORAGE, "WRONLY\n" );
			if ( file_exists ) {
				// File exists
				if (!file_catalog->access( l_fname, user, NEST_WRITE_ACCESS)) {
					return NEST_NONPRIVELEDGED_USER;
				}
			}
			else {
				// File does not exist
				//  -- it probably *should* have been created by O_CREAT above
				if (file_catalog->paccess( l_fname, user, NEST_WRITE_ACCESS)) {
					return NEST_FILE_NOT_FOUND;
				}
				return NEST_NONPRIVELEDGED_USER;
			}
		}

		if (flags & O_RDWR)  {
			nest_debug( NEST_DEBUG_STORAGE, "RDWR\n" );
			if ( file_exists ) {
				if (! file_catalog->access( l_fname, user, NEST_RW_ACCESS )) {
					return NEST_NONPRIVELEDGED_USER;
				}
			}
			else {
				if (file_catalog->paccess( l_fname, user, NEST_RW_ACCESS)) {
					return NEST_FILE_NOT_FOUND;
				}
				return NEST_NONPRIVELEDGED_USER;
			}
		}

		reply = file_catalog->get_physical_name(p_fname, l_fname);

		if ( reply != NEST_SUCCESS ) {
			nest_debug( NEST_DEBUG_STORAGE,
						"Error while translating "
						"logical name to physical name!\n" );
			return reply;
		}

		nest_obj_type_t type;
		reply = file_catalog->get_type( &type, l_fname );

		if (reply != NEST_SUCCESS) {
			goto delete_and_out;
		}

		if ( (type == NEST_LINK) && !system_file) {
			/*
			  sym-link ==> need to act accordingly
			*/

			if (link_count++ > NEST_LINK_MAXLOOP) {
				nest_debug( NEST_DEBUG_STORAGE,
							"Too many levels of symbolic links!\n" );
				return NEST_TOO_MANY_LINKS;
			}

			return nestStorageManager::open(connection,
											user,
											p_fname,
											flags,
											offset,
											filesize,
											cwd,
											link_count );
		}

		if ( (type != NEST_FILE) && (type != NEST_LINK) ) {
			nest_debug( NEST_DEBUG_STORAGE,
						"Trying to open a non-file object!\n" );
			reply = NEST_BAD_PARAMETERS;
			goto delete_and_out;
		}

		fd = NestOpen( p_fname, flags, nestStorageManager::file_create_mode );
		if ( fd < 0 ) {
			reply = ErrnoToNest(errno);
			goto delete_and_out;
		}
	}
	
	// Create the connection
	conn = new nestProtocolFilesystem( fd, (off_t)offset );
	*connection = conn;

    if (! conn ) {
        nest_debug( NEST_DEBUG_STORAGE,
            "Couldn't instantiate a nestProtocolFilesystem in open!\n" );
        reply = NEST_OUT_OF_RESOURCES;
        goto close_and_out;
    }

    //  adding the open_flags
    conn->set_open_flags( flags );

    if (flags == O_RDONLY) {
        return NEST_SUCCESS;
    }

    if ( new_file_created ) {
        NestChmod( p_fname, nestStorageManager::file_create_mode );
        NestChown( p_fname, user->getUid( ), ActualGid );
    }

    //  adding the l_fname & p_fname to the *result
    conn->set_filenames( l_fname, p_fname );

    //  adding the open_filesize
    nestFileSize filesize_bytes;
    reply = file_catalog->get_size( &filesize_bytes, l_fname );
    if (reply != NEST_SUCCESS) {
        goto free_close_and_out;
    }

	// Compute how much the file has grown
	nestLotSize  more_bytes;
	if ( filesize < 0 ) {
		more_bytes = -1;
	} else if ( filesize > offset ) {
		more_bytes = filesize - filesize_bytes;
	} else {
		more_bytes = offset - filesize_bytes;
	}

    nestLotSize avail_bytes;
    if (NestEnableLots && !system_file) {
        /*          
            We have to do this for file transfers that overwrite a file 
            but do not transfer the size of the file 
        */
        if(filesize == ENTIRE_FILE) {
            reply = resetAllocation(file, user, cwd, offset);
            if (reply != NEST_SUCCESS) {
                goto free_close_and_out;
            }
        }

        avail_bytes = lot_manager->available_lot_space(user, l_fname);

        nest_debug( NEST_DEBUG_STORAGE,
                    "Avail user_lot_space(%s) = %lld & filesize = %ld\n",
                    user->getName(),
                    avail_bytes,
                    more_bytes );
	
        if ( (avail_bytes == 0) && (more_bytes > 0) ) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Available Lot space for user %s is 0 bytes!\n",
                        user->getName() );
            reply = NEST_LOT_NOT_ENOUGH;
            goto free_close_and_out;
        }

        if ( (more_bytes > 0) && (avail_bytes < more_bytes) ) {
            nest_debug (NEST_DEBUG_STORAGE,
                        "Not enough space available in the user's Lots!\n");
            reply = NEST_LOT_NOT_ENOUGH;
            goto free_close_and_out;
        }
	
		// Zero byte files are not attached to any lot and hence do
		// not need a lot.
        if (more_bytes > 0) {
            reply = lot_manager->update_lot_usage( l_fname, more_bytes);
            if(reply != NEST_SUCCESS) {
                goto free_close_and_out;
            }

            filesize_bytes += more_bytes;
            file_catalog->set_size( l_fname, filesize_bytes );
        }
    } else {
        nestLotSize total_bytes;
		reply = bytesAvailable( IS_USER, &avail_bytes, &total_bytes );
        if ( reply != NEST_SUCCESS ) 
        {
            goto free_close_and_out; 
        }

        nest_debug( NEST_DEBUG_STORAGE,
                    "Avail device_space = %lld & filesize = %ld\n",
                    avail_bytes,
                    more_bytes );

        if ((avail_bytes == 0) && (more_bytes != 0)) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Available disk space is 0 bytes!\n" );
            reply = NEST_DISK_FULL;
            goto free_close_and_out;
        }

        if ((more_bytes > 0) && (avail_bytes < more_bytes)) {
            nest_debug (NEST_DEBUG_STORAGE,
                        "Not enough space available in the device!\n");
            reply = NEST_DISK_FULL;
            goto free_close_and_out;
        }
    }

    nest_debug( NEST_DEBUG_STORAGE,
                "Setting open_filesize %ld in nestStorageManager::open()\n",
                filesize_bytes );
    conn->set_open_filesize(filesize_bytes);

    if (!NestEnableLots) {
        return NEST_SUCCESS;
    }


    //  adding user and group IDs
    nest_debug( NEST_DEBUG_STORAGE,
                "Setting ids in nestStorageManager::open() UID %d GID %d\n",
                user->getUid(), ActualGid );
    conn->set_ids( user->getUid( ), ActualGid );

    /*
        Its a PUT -
        Lots are enabled -
        (*conn) is all set -

        Turn on the "TransferInProgress" flag
    */

    if( !system_file ) {
        reply = file_catalog->set_transfer_flag( l_fname, true );
        if (reply != NEST_SUCCESS) {
            goto free_close_and_out;
        }
    }

    return NEST_SUCCESS;

free_close_and_out:
    delete (*connection);
    conn = NULL;
    *connection = NULL;

close_and_out:
	if ( fd >= 0 ) {
		NestClose( fd );
	}

delete_and_out:
    if (new_file_created) {
        if( !system_file ) {
            file_catalog->delete_entry( l_fname );
		}
        NestUnlink(p_fname);
    }

    return reply;
}

char *
nestStorageManager::filehandleToPath( NestFilehandle *uniquifier )
{
    char *path = 
		(char *) nest_hash_table_lookup( fh_hash, (char*)uniquifier ); 
    nest_debug( NEST_DEBUG_STORAGE, "found hash of %d to %s\n",
				*uniquifier, path );
    return path;
}

NestReplyStatus
nestStorageManager::pathToFilehandle(   const nestUser *user,
                                        const char *file,
                                        const char *cwd,
                                        NestFilehandle    *uniquifier,
                                        int        link_count ) 
{

	// this is no longer used -- look in CVS to see what used to be here...
    assert( 0 );
    return NEST_NOT_YET_IMPLEMENTED;
}

void
nestStorageManager::cache_uniquifier(   string &filename,
                                        NestFilehandle *uniquifier,
                                        nest_obj_type_t type ) 
{

        // if the file is a dir, add a slash onto the end
        // this won't work for symbolic links . . .
    if ( type == NEST_DIR ) {
        nest_debug( NEST_DEBUG_STORAGE,
					"adding slash to end of lookup dir\n" );
        filename += '/';
    }
    if (nest_hash_table_lookup( fh_hash, (char*)uniquifier) == NULL) {
        nest_hash_table_insert( fh_hash, (char*)uniquifier, 
                NestStrdup(filename.c_str()) );
        nest_debug( NEST_DEBUG_STORAGE,
					"adding hash of %d to %s (ht size %d)\n", 
            *uniquifier, filename.c_str(), nest_hash_table_size( fh_hash ) );
    } else {
        nest_debug( NEST_DEBUG_STORAGE,
					"not inserting a duplicate hash for %d\n", 
            *uniquifier, filename.c_str() );
    }
}

NestReplyStatus
nestStorageManager::closeLotFile( nestClient *client,
								  const nestRequest *req,
								  const nestUser *owner,
								  const char * l_fname,
								  nestFileSize open_filesize,
								  nestFileSize close_filesize,
								  bool &delete_file )
{
	nestUser        user;
	nestLotSize     lot_space = 0;
	NestReplyStatus error_return;

	// Figure out how much lot space the user has...
	lot_space = lot_manager->available_lot_space( owner, l_fname);

	if( req->getEnd() != ENTIRE_FILE ) {
    
		// If the file size is known before hand, at the end of
		// transfer, we check the filesize again to make sure nothing
		// has gone wrong.  If more data than the file size is sent,
		// the lot is updated if space is available or the file is
		// removed.

		nestLotSize more_space = ( close_filesize - open_filesize );
		if ( lot_space >= more_space) {

			nest_debug( NEST_DEBUG_STORAGE,
						"Lot_space %f >= filesize %f\n",
						lot_space, more_space );

			error_return = lot_manager->update_lot_usage( l_fname, more_space);

			if ( error_return == NEST_SUCCESS) {

				// return value purposefully ignored
				// - doesn't really matter or really can't do anything at all
				// - in all other cases the file will get deleted, so
				//   dont bother!

				file_catalog->set_transfer_flag( l_fname, false );
			} else {
				nest_debug( NEST_DEBUG_STORAGE,
							"File-Lot attachment failed!\n" );
				delete_file = true;
				return error_return;
			}
		}
		else {
			// Here: ( lot_space < more_space )
			nest_debug( NEST_DEBUG_STORAGE,
						"Not enough space available in user's Lots\n" );
			delete_file = true;
			return NEST_LOT_NOT_ENOUGH;
		}
		return NEST_SUCCESS;
	}


	// If we've made it here, it's not the entire file

	if( !client || !req ) {
		delete_file = true;
		return NEST_BAD_PARAMETERS;
	}

	// If the file size is not known before hand, we cannot use
	// the open_filesize value as it will be zero. Instead
	// we compare close_filesize with the allocated lot
	// space and allocate more if needed or delete the file as in
	// previous case.
			
	// hopefully the allocated lot space is larger than actual
	// filesize - cos lot space is allocated in MBs when dynamically
	// allocated.
                
	nestLotSize          allocated_space = 0;

	error_return = fileusage( &allocated_space, 
							  req->getNestFile(),
							  client->getUser(),
							  client->getCwd() );

	if(error_return == NEST_SUCCESS) {
		nestLotSize more_space = allocated_space - close_filesize;

		// if more_space is positive:
		//    We have over allocated. Reduce the lot allcation to file
		// if more_space is negative:
		//    We have under allocated.  Increase the lot allocation
		//    to file if lot space is available,
		// else:
		//    return error.

		if( more_space > 0 ) {
			lot_manager->update_lot_usage( l_fname, 0 - more_space );
			file_catalog->set_transfer_flag( l_fname, false );
		} else {
			if ( lot_space >= more_space) {
				nest_debug( NEST_DEBUG_STORAGE,
							"Lot_space %f >= filesize %f\n",
							lot_space,
							more_space );

				error_return = lot_manager->update_lot_usage(
					l_fname, 0 - more_space);

				if ( error_return == NEST_SUCCESS) {
					file_catalog->set_transfer_flag( l_fname, false );
					return NEST_SUCCESS;
				}
				else {
					nest_debug( NEST_DEBUG_STORAGE,
								"File-Lot attachment failed!\n" );
					delete_file = true;
					return error_return;
				}

			}
			else
			{
				nest_debug( NEST_DEBUG_STORAGE,
							"Not enough space available in user's Lots\n");
				delete_file = true;
				return NEST_LOT_NOT_ENOUGH;
			}
		}
	}

	return error_return;
}

// contract
// storageConnection cannot be destroyed until _after_ calling this function
NestReplyStatus
nestStorageManager::close( nestProtocol *connection,
                           nestClient *client ) 
{
	nestProtocolStorage *storageConnection = NULL;

    if (!connection || !client) {
        return NEST_BAD_PARAMETERS;
    }

	if ( connection->getType() != STORAGE_PROTOCOL ) {
		nest_die( "Weirdness in nestStorageManager\n" );
	} else {
		storageConnection = (nestProtocolStorage *)connection;
	}

    nestRequest *req = client->getProxyRequest();
    if(!req) {
         return NEST_BAD_PARAMETERS;
    }
    
    int open_flags;
    int fd = storageConnection->getSocket();

	NestReplyStatus error_return = NEST_SUCCESS;
	bool delete_file = false;
    storageConnection->get_open_flags( &open_flags );

	// If the file was open for writing, we have more work to do
    if (  (open_flags & O_WRONLY) || (open_flags & O_RDWR)  ) {
        char           l_fname[NEST_MAXPATHLEN];
        char           p_fname[NEST_MAXPATHLEN];
		nestUser       owner;
		nestFileSize   close_filesize = 0;

        storageConnection->get_filenames(l_fname, p_fname);

		error_return = file_catalog->get_owner_user( l_fname, &owner );
		const nestUser *ownerp = &owner;

		if ( error_return ) {
			nest_debug( NEST_DEBUG_ALWAYS,
						"%s:%d: Error finding owner of '%s': %d '%s'\n",
						__FUNCTION__, __LINE__, l_fname,
						error_return, NestErrorString( error_return ) );
			ownerp = client->getUser( );
			goto CLEANUP;
		}

		// If it's not a system file, proceed on..
        if ( file_catalog->is_system_file( l_fname )) {
			goto DONE;
		}

		nest_stat   stat_buf;
		if ( NestFstat(fd, &stat_buf) < 0 ) {
			nest_debug( NEST_DEBUG_STORAGE, "Next to impossible!\n" );
		} else {
			close_filesize  = stat_buf.st_size;
		}
		file_catalog->set_size( l_fname, close_filesize );

		nestFileSize   open_filesize;
		storageConnection->get_open_filesize( &open_filesize );

		// If lots are enabled, extra logic is required
		if (NestEnableLots) {
			error_return = closeLotFile( client,
										 req,
										 &owner,
										 l_fname,
										 open_filesize,
										 close_filesize,
										 delete_file );
		}

		// Clean up
	  CLEANUP:
		if ( delete_file ) {
			char      file[NEST_MAXPATHLEN];
			file_catalog->get_reverse_logical_name( file, l_fname );

			// as the file is going to be deleted and the open_filesize
			// would have been already accounted to relevant Lots, set the
			// filesize to open_filesize and call deleteFile()
			file_catalog->set_size( l_fname, open_filesize );
			deleteFile( ownerp, file, "" );
			file_catalog->set_transfer_flag( l_fname, false );
		}
	}

	// Close the FD it it wasn't cached..
  DONE:
	NestClose( fd );

	return error_return;
}


/*
    Only NEST_LINKS - deletes for ~system links dont come here
*/
NestReplyStatus
nestStorageManager::deleteLink( const char *logical_link )
{
    return file_catalog->delete_entry(logical_link);
}

NestReplyStatus
nestStorageManager::renameFile(const nestUser *user,
                               const char *oldpath,
                               const char *newpath,
                               const char *cwd,
                               bool update_lot )
{

    nest_debug( NEST_DEBUG_STORAGE, 
                "rename called - '%s', '%s' -> '%'s\n",
                user->getName(), oldpath, newpath );
    
    if ( !user || !oldpath || !newpath ) {
        return NEST_BAD_PARAMETERS;
    }

    string oldfn;
	string newfn;

    NestReplyStatus reply = expand_path( oldfn, user, cwd, oldpath );
    NEST_REPLY_CHECK(reply);

    reply = expand_path( newfn, user, cwd, newpath );
    NEST_REPLY_CHECK(reply);

    char    l_oldfn[NEST_MAXPATHLEN];		// Logical old file name
    char    l_newfn[NEST_MAXPATHLEN];		// Logical new file name

    reply = file_catalog->get_logical_name( l_oldfn, oldfn.c_str(), user);
    NEST_REPLY_CHECK(reply); 

    /*
        ignoring return value explicitly as the file entry would
        not be created yet
     */
    reply = file_catalog->get_logical_name( l_newfn, newfn.c_str(), user);

    bool old_is_sysf = file_catalog->is_system_file(l_oldfn);
    bool new_is_sysf = file_catalog->is_system_file(l_newfn);

    if(old_is_sysf && !new_is_sysf || !old_is_sysf && new_is_sysf ) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

    if (file_catalog->is_file_present (l_oldfn)) {
        if (!file_catalog->access(l_oldfn, user, NEST_READ_ACCESS)) {
            return NEST_NONPRIVELEDGED_USER;
        }
    } else {
        if (file_catalog->paccess(l_oldfn, user, NEST_LIST_ACCESS)) {
            return NEST_FILE_NOT_FOUND;
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!file_catalog->is_file_present (l_newfn)) {
        if (!file_catalog->access(l_oldfn, user, NEST_WRITE_ACCESS)) {
            return NEST_NONPRIVELEDGED_USER;
        }
    } else {
        if (file_catalog->paccess(l_oldfn, user, NEST_LIST_ACCESS)) {
            return NEST_FILE_EXISTS;
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    nest_obj_type_t type;
    reply = file_catalog->get_type(&type, l_oldfn);
    NEST_REPLY_CHECK(reply);

    if (type != NEST_FILE) {
        return NEST_NOT_YET_IMPLEMENTED;
    }
  
    nestFileSize filesize_bytes;
    reply = file_catalog->get_size(&filesize_bytes, l_oldfn); 
    NEST_REPLY_CHECK(reply);
    
    /* 
        The following operations must be atomic. And I don't know
        the correct way to do it. For now, I am ignoring errors.
        The operations:
            1) Update file catalog entries
            2) Rename file
            3) Update lot usage
     */

    // Add entry in file catalog
    reply = file_catalog->add_entry(l_newfn, user, NEST_FILE);
    NEST_REPLY_CHECK(reply);
    reply = file_catalog->set_size(l_newfn, filesize_bytes );
    if(reply != NEST_SUCCESS) {
        goto delete_file_entry_and_return;
    }

    // Do actual rename
    char    p_oldfn[NEST_MAXPATHLEN];
    char    p_newfn[NEST_MAXPATHLEN];
    file_catalog->get_physical_name( p_oldfn, l_oldfn);
    file_catalog->get_physical_name( p_newfn, l_newfn);

    if( rename(p_oldfn, p_newfn) < 0 ) {
        goto delete_file_entry_and_return;
    }
    
    NestChmod( p_newfn, nestStorageManager::file_create_mode );
    NestChown( p_newfn, user->getUid( ), ActualGid );

    // Update lot usage
    if (NestEnableLots && update_lot && !old_is_sysf) {
        reply = lot_manager->update_lot_usage( l_oldfn, 0 - filesize_bytes );
        if(reply != NEST_SUCCESS) {
            goto undo_rename_and_return;
        }
        reply = lot_manager->update_lot_usage(l_newfn, filesize_bytes );
        if(reply != NEST_SUCCESS) {
            goto redo_lot_usage_and_return;
        }
    }
   
    // Delete entry in file catalog 
    reply = file_catalog->delete_entry( l_oldfn );
    if(reply != NEST_SUCCESS) {
        goto undo_lot_usage_and_return;
    }

#if 0
    //    FIX ME: Correct this part when moving across NeST and 
	//   ~system file systems are implemented

    nestLotSize avail_bytes;
    if (NestEnableLots) {

        avail_bytes = lot_manager->available_user_lot_space(user);

        nest_debug( NEST_DEBUG_STORAGE,
                    "Avail user_lot_space(%s) = %lld & filesize = %ld\n",
                    user->getName(),
                    avail_bytes,
                    filesize_bytes );
	
        if ((avail_bytes == 0) && (filesize_bytes > 0)) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Available Lot space for user %s is 0 bytes!\n",
                        user->getName() );
            reply = NEST_LOT_NOT_ENOUGH;
            goto delete_file_entry_and_out;
        }

        if ((filesize_bytes > 0) && (avail_bytes < filesize_bytes)) {
            nest_debug (NEST_DEBUG_STORAGE,
                        "Not enough space available in the user's Lots!\n");
            reply = NEST_LOT_NOT_ENOUGH;
            goto delete_file_entry_and_out;
        }
	
		// Zero byte files are not attached to any lot and hence do
		// not need a lot.
	 
        if (filesize_bytes > 0) {
            nestLotSize  more_bytes = filesize_bytes;

            reply = lot_manager->update_lot_usage(l_newfn,
                                                  more_bytes);
            if(reply != NEST_SUCCESS) {
                goto delete_file_entry_and_out;
            }

            reply = file_catalog->set_size( l_newfn,
                                            filesize );
            if(reply != NEST_SUCCESS)
                goto delete_file_entry_and_out;

        }
    } else {
        nestLotSize total_bytes;

		reply = bytesAvailable( IS_USER, &avail_bytes, &total_bytes );
        if ( reply != NEST_SUCCESS ) 
        {
            goto delete_file_entry_and_out; 
        }

        nest_debug( NEST_DEBUG_STORAGE,
                    "Avail device_space = %lld & filesize = %ld\n",
                    avail_bytes,
                    filesize_bytes );

        if ((avail_bytes == 0) && (filesize_bytes != 0)) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Available disk space is 0 bytes!\n" );
            reply = NEST_DISK_FULL;
            goto delete_file_entry_and_out;
        }

        if ((filesize_bytes > 0) && (avail_bytes < filesize_bytes)) {
            nest_debug (NEST_DEBUG_STORAGE,
                        "Not enough space available in the device!\n");
            reply = NEST_DISK_FULL;
            goto delete_file_entry_and_out;
        }
        
        reply = file_catalog->set_size( l_newfn,
                                        filesize_bytes );
        if(reply != NEST_SUCCESS)
            goto delete_file_entry_and_out;
        
    }
#endif

    return NEST_SUCCESS;

undo_lot_usage_and_return:
    lot_manager->update_lot_usage(l_newfn,
                                  0 - filesize_bytes );

redo_lot_usage_and_return:
    lot_manager->update_lot_usage(l_oldfn,
                                  filesize_bytes ); 
    
undo_rename_and_return:
    rename(p_newfn, p_oldfn);


delete_file_entry_and_return:
    file_catalog->delete_entry( l_newfn );

    return reply;
}

NestReplyStatus
nestStorageManager::deleteFile( const nestUser *user,
                                const char *file,
                                const char *cwd,
                                bool update_lot)
{

    if ( !user || !file ) {
        return NEST_BAD_PARAMETERS;
    }

    string filename;
    NestReplyStatus reply = expand_path( filename, user, cwd, file );
    NEST_REPLY_CHECK(reply);

    char    l_fname[NEST_MAXPATHLEN];
    bool    delete_access_present = false;

    reply = file_catalog->get_logical_name( l_fname, filename.c_str(), user);

    delete_access_present =
		file_catalog->paccess( l_fname, user, NEST_DELETE_ACCESS );

    if (reply != NEST_SUCCESS) {
        if (delete_access_present) {
            return reply;
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!delete_access_present) {
        return NEST_NONPRIVELEDGED_USER;
    }

    nest_obj_type_t type;
    char    	p_fname[NEST_MAXPATHLEN];
	nestUser	owner;

    reply = file_catalog->get_type( &type, l_fname );

    NEST_REPLY_CHECK(reply);

    if (type == NEST_DIR) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Cannot delete dir with the SM::deleteFile() !\n" );
        return NEST_BAD_PARAMETERS;
    }

    if (file_catalog->is_system_file(l_fname)) {

        reply = file_catalog->get_physical_name( p_fname, l_fname );

        NEST_REPLY_CHECK(reply);

        /*
            system-links can never point to nest-files
            in this version
        */

        if (type == NEST_LINK) {
            if (NestUnlink(p_fname) < 0) {
                return ErrnoToNest(errno);
            }
            return NEST_SUCCESS;
        }

        /*
            Its a system-file - OK
            See, if its a physical_file of some logical nest-file
        */

        reply = file_catalog->get_logical_name( l_fname, p_fname );

        if (reply != NEST_SUCCESS) {

            /*
                if somebody is deleting a non-NeST file
                dont care about the Lot status!

                use TruckUnlink anyway!
            */

            if (NestTruncUnlink(p_fname) < 0) {
                return ErrnoToNest(errno);
            }

            return NEST_SUCCESS;
        }
    }

    /*
        only system-files, nest-files or nest-links
        might come here!
    */

    if (type == NEST_LINK) {
        return deleteLink( l_fname );
    }

    /*
        only NEST_FILES come here
    */

    reply = file_catalog->get_physical_name( p_fname, l_fname );
    if (reply != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Error while translating %s to physical name!\n",
                    l_fname );
        nest_debug( NEST_DEBUG_STORAGE,
                    "Probably, accessing a non-NeST file!\n" );
        return NEST_FILE_NOT_FOUND;
    }

    reply = file_catalog->get_owner_user( l_fname, &owner );

    if (reply != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Error while getting owner name for %s \n",
                    l_fname );
        return NEST_FILE_NOT_FOUND;
    }

    nestFileSize   filesize_bytes;
    reply = file_catalog->get_size( &filesize_bytes, l_fname );
    NEST_REPLY_CHECK(reply);

    /*
        I really dont know the best way to do this

        Basically,
            the lot_usage should be (conditionally) updated,
            the file_catalog entry should be deleted and
            the physical file should be Unlinked

        I don't know how to make these changes atomic

        I'm ignoring errors
    */

    if (NestEnableLots && update_lot) {
        reply = lot_manager->update_lot_usage(l_fname, 0 - filesize_bytes );
        NEST_REPLY_CHECK(reply);
    }

    int retval = NestTruncUnlink(p_fname);
    if (retval < 0) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "NestTruncUnlink() failed - %s!\n",
                    strerror(errno) );
    }

    reply = file_catalog->delete_entry(l_fname);
    if (reply != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "FileCatalog->delete_entry(%s) failed: %s"
                    " - IGNORING ERROR\n",
                    l_fname,
                    NestErrorString( reply ) );
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::get_type( nest_obj_type_t *result,
                              const char *file,
                              const nestUser *user,
                              const char *cwd )
{

    string filename;
    NestReplyStatus reply = expand_path( filename, user, cwd, file );
    NEST_REPLY_CHECK(reply);

    char l_fname[NEST_MAXPATHLEN];
    reply = file_catalog->get_logical_name( l_fname, filename.c_str(), user);

    if (reply != NEST_SUCCESS) {
        if (file_catalog->paccess( l_fname, user, NEST_LIST_ACCESS )) {
            return reply;
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!file_catalog->access( l_fname, user, NEST_READ_ACCESS )) {
        return NEST_NONPRIVELEDGED_USER;
    }

    reply = file_catalog->get_type( result, l_fname);
    if (reply != NEST_FILE_NOT_FOUND) {
        return reply;
    }

    /*
        Hmm .. could be a NEST_DIR and the l_fname
        is missing a '/' at the end - oh boy!
    */

    int len = NestStrlen( l_fname );
    if ( l_fname[len - 1] == '/' ) {
        return NEST_FILE_NOT_FOUND;
    }

    l_fname[len] = '/';
    l_fname[len+1] = 0;

    return file_catalog->get_type( result, l_fname );
}

NestReplyStatus
nestStorageManager::get_physical_filename(char **physical,
                                          const char *file,
                                          const nestUser *user,
                                          const char *cwd,
                                          int link_count)
{
    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    char    logical_name[NEST_MAXPATHLEN];
    char    physical_name[NEST_MAXPATHLEN];

    reply = file_catalog->get_logical_name( logical_name,
                                            filename.c_str(),
                                            user );

    if (reply != NEST_SUCCESS) {
        if (file_catalog->paccess( logical_name, user, NEST_LIST_ACCESS )) {
            return reply;
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!file_catalog->access(logical_name, user, NEST_READ_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    reply = file_catalog->get_physical_name( physical_name,
                                             logical_name );
    NEST_REPLY_CHECK(reply);

    if (!file_catalog->is_system_file( logical_name )) {
        nest_obj_type_t     filetype;

        reply = file_catalog->get_type( &filetype,
                                        logical_name );
        NEST_REPLY_CHECK(reply);

        if (filetype == NEST_LINK) {

            if (link_count++ > NEST_LINK_MAXLOOP) {
                nest_debug( NEST_DEBUG_STORAGE,
                            "Too many levels of symbolic links!\n" );
                return NEST_TOO_MANY_LINKS;
            }

            return get_physical_filename( physical,
                                          physical_name,
                                          user,
                                          cwd,
                                          link_count );
        }
    }

    *physical = NestStrdup( physical_name );
    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::Link(const nestUser *user,
                         const char *oldpath,
                         const char *newpath,
                         int val,
                         const char *cwd)
{

    if (!user || !oldpath || !newpath) {
        return NEST_BAD_PARAMETERS;
    }

    if (!val) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

    string  oldpath_file,
            newpath_file;

    /*
        ignoring return value intentionally
    */

    NestReplyStatus reply = expand_path( oldpath_file,
                                         user,
                                         cwd,
                                         oldpath );
    /*
        ignoring return value intentionally
    */

    reply = expand_path( newpath_file,
                         user,
                         cwd,
                         newpath );

    char    old_logical[NEST_MAXPATHLEN],
            new_logical[NEST_MAXPATHLEN];

    /*
        ignoring return value intentionally
    */
    reply = file_catalog->get_logical_name( old_logical,
                                            oldpath_file.c_str(),
                                            user );

    /*
        ignoring return value intentionally
    */
    reply = file_catalog->get_logical_name( new_logical,
                                            newpath_file.c_str(),
                                            user );

    nest_obj_type_t     type;

    reply = file_catalog->get_type( &type,
                                    old_logical );
    
    NEST_REPLY_CHECK(reply);
    
    if (type == NEST_DIR) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

    if (!file_catalog->paccess( new_logical, user, NEST_INSERT_ACCESS )) {
        return NEST_NONPRIVELEDGED_USER;
    }

#if 0
    if (!file_catalog->access( old_logical, user, NEST_READ_ACCESS )) {
        return NEST_NONPRIVELEDGED_USER;
    }
#endif


    if (file_catalog->is_file_present( new_logical )) {
        return NEST_FILE_EXISTS;
    }

    char    old_filename[NEST_MAXPATHLEN];
    reply = file_catalog->get_reverse_logical_name(old_filename,
                                                   old_logical);
    NEST_REPLY_CHECK(reply);

    reply = file_catalog->add_entry(new_logical,
                                    old_filename,
                                    user,
                                    NEST_LINK);

    if (reply != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Couldn't add-entry into FileCatalog for NEST_LINK! %s\n",
                    NestErrorString(reply));
        return reply;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::sync(const nestUser *user)
{
    NestSync();
    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::synchronize()
{

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::maintenance()
{
    nest_debug( NEST_DEBUG_STORAGE,
                "Executing Storage manager maintenance - "
                "Truncating ClassAd log.\n" );
    Resources->TruncateLog();
    identity_manager->synchronize();
    synchronize();
    file_catalog->synchronize();
    if (NestEnableLots) {
        //lot_manager->synchronize();
    }
    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::shutdown()
{
    identity_manager->shutdown();
    file_catalog->shutdown();
    if (NestEnableLots) {
        lot_manager->shutdown();
    }
    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::get_acl(NestACL **result,
							const char *file,
							const nestUser *user,
							const char *cwd )
{

    if (!result || !file || !user) {
        return NEST_BAD_PARAMETERS;
    }

    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    char    l_fname[NEST_MAXPATHLEN];
    reply = file_catalog->get_logical_name( l_fname, filename.c_str(), user );

    if (!file_catalog->paccess( l_fname, user, NEST_LIST_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    NEST_REPLY_CHECK(reply);

    *result = NULL;

    reply = file_catalog->get_acl(result, l_fname, IS_USER );
    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_acl(result, l_fname, IS_GROUP );

    return reply;
}

NestReplyStatus
nestStorageManager::get_perm(nest_access_rights_t	*result,
							 const nestUserOrGroup	*for_ug,
							 const char				*file,
							 const nestUser			*user,
							 const char				*cwd )
{

    if (!result || !for_ug || !file || !user) {
        return NEST_BAD_PARAMETERS;
    }

    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    char    l_fname[NEST_MAXPATHLEN];
    reply = file_catalog->get_logical_name( l_fname, filename.c_str(), user );

    if (!file_catalog->paccess( l_fname, user, NEST_LIST_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    reply = file_catalog->get_perm(l_fname, for_ug, result);

    return reply;
}

NestReplyStatus
nestStorageManager::set_perm(nest_access_rights_t arg,
							 const nestUserOrGroup *to_ug,
							 const char *file,
							 const nestUser *user,
							 const char *cwd)
{
    if (!to_ug || !file || !user) {
        return NEST_BAD_PARAMETERS;
    }

    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    char    l_fname[NEST_MAXPATHLEN];
    reply = file_catalog->get_logical_name( l_fname, filename.c_str(), user );

    if (reply != NEST_SUCCESS) {
        if (file_catalog->paccess(l_fname, user, NEST_MODIFYAC_ACCESS)) {
            return reply;
        }

        return NEST_NONPRIVELEDGED_USER;
    }

	nestUser    owner;
	reply = file_catalog->get_owner_user( l_fname, &owner );
	NEST_REPLY_CHECK(reply);

    if (!file_catalog->access( l_fname, user, NEST_MODIFYAC_ACCESS) ) {
		if ( user == &owner ) {
			return NEST_NONPRIVELEDGED_USER;
		}
    }

    reply = file_catalog->set_perm(l_fname, to_ug, arg);

    return reply;
}

NestReplyStatus
nestStorageManager::frame_full_path( string &result,
                                     const nestUser *user,
                                     const char *file )
{

    result = "";

    if (!file || !file[0]) {
        //result = result + "~" + user + "/";
        result = result + "~" + "/";
        return NEST_SUCCESS;
    }

    if (!strncmp( file, "~/", 2 )) {
        //result = result + "~" + user + (file + 1);
        result = result + "~" + (file + 1);
    } else {
        result = file;
    }

    return NEST_SUCCESS;
}


NestReplyStatus
nestStorageManager::expand_path ( string &result,
                                  const nestUser *user,
                                  const char *cwd,
                                  const char *file )
{
   
    NestReplyStatus reply;
    char    final_path[NEST_MAXPATHLEN];
    nest_debug( NEST_DEBUG_STORAGE, "expand_path: cwd='%s'\n", cwd );

    if (file[0] && (file[0] == '~')) {
        reply = file_catalog->get_dest_file( final_path,
                                             file, user );
        NEST_REPLY_CHECK(reply);

        return frame_full_path( result, user, final_path );
    }

    /*
        the filename is not absolute anymore

        cwd needn't be collapsed as it would never have .s or ..s
    */

    reply = frame_full_path( result, user, cwd);

    NEST_REPLY_CHECK(reply);

    result = result + file;
    nest_debug( NEST_DEBUG_STORAGE,
				"expand_path: result='%s'\n", result.c_str() );

    /*
        this needs to be collapsed!!
    */
    reply = file_catalog->get_dest_file( final_path,
                                         result.c_str(),
										 user );

    NEST_REPLY_CHECK(reply);
    nest_debug( NEST_DEBUG_STORAGE, "expand_path: final='%s'\n", final_path );

    result = final_path;
    return NEST_SUCCESS;
}


NestReplyStatus
nestStorageManager::get_dest_dir( char *dest,
                                  const char *cwd,
                                  const char *dir,
                                  const nestUser *user )
{

    if (!dest || !cwd || !dir) {
        return NEST_BAD_PARAMETERS;
    }

    if (!strcmp(dir, "~")) {
        strcpy( dest, "~/" );
        return NEST_SUCCESS;
    }

    char    tmp_dest[NEST_MAXPATHLEN];

    if (dir[0] == '~') {
        if (NestSnprintf( tmp_dest, NEST_MAXPATHLEN, "%s", dir ) < 0) {
            return NEST_FILENAME_TOO_LONG;
        }
    }
    else {
        if (cwd[NestStrlen(cwd)-1] != '/') {
            if (NestSnprintf( tmp_dest,
                              NEST_MAXPATHLEN,
                              "%s/%s",
                              cwd,
                              dir ) < 0) {
                return NEST_FILENAME_TOO_LONG;
            }
        }
        else {
            if (NestSnprintf( tmp_dest,
                              NEST_MAXPATHLEN,
                              "%s%s",
                              cwd,
                              dir ) < 0) {
                return NEST_FILENAME_TOO_LONG;
            }
        }
    }

    /*
        takes care of ../ ./ and //
    */

    NestReplyStatus reply = file_catalog->get_dest_file( dest,
														 tmp_dest,
														 user );
    NEST_REPLY_CHECK(reply);

    int last_byte = NestStrlen(dest) - 1;

    if ((last_byte < 0) || (last_byte >= NEST_MAXPATHLEN)) {
        return NEST_BAD_PARAMETERS;
    }

    if (dest[last_byte] != '/') {
        dest[last_byte + 1] = '/';
        dest[last_byte + 2] = 0;
    }

    return NEST_SUCCESS;
}


NestReplyStatus
nestStorageManager::chDir(char **new_cwd,
                          const nestUser *user,
                          const char *cwd,
                          const char *dir)
{

    char    dest[NEST_MAXPATHLEN];
    char    logical_dest[NEST_MAXPATHLEN];

    nest_debug( NEST_DEBUG_STORAGE,
                "chdir called - %s,%s\n", user->getName(), dir );

    NestReplyStatus reply = get_dest_dir( dest,
                                          cwd,
                                          dir,
                                          user );

    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_logical_name( logical_dest,
                                            dest,
                                            user );

    if (reply != NEST_SUCCESS) {
        if (file_catalog->paccess(logical_dest,
                                                      user,
                                                      NEST_LIST_ACCESS)) {
            return reply;
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!file_catalog->access(logical_dest,
                                         user,
                                         NEST_LIST_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    *new_cwd = NestStrdup( dest );

    nest_debug( NEST_DEBUG_STORAGE, "chDir to %s\n", *new_cwd );

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::mkDir( const nestUser *user,
                           const char *cwd,
                           const char *dir,
                           user_id_t     uid )
{

    char    dest[NEST_MAXPATHLEN];
    char    logical_dest[NEST_MAXPATHLEN];

    nest_debug( NEST_DEBUG_STORAGE,
                "mkdir called - %s,%s\n", user->getName( ), dir );

    NestReplyStatus reply = get_dest_dir( dest,
                                          cwd,
                                          dir,
                                          user );
    NEST_REPLY_CHECK(reply);

    nest_debug( NEST_DEBUG_STORAGE,
                "mkdir %s [%d] - user '%s'\n",
                dest, NEST_MAXPATHLEN, user->getName() );

    string  tmp_full_path,
            tmp_userhome;

    reply = frame_full_path(tmp_userhome,
                            user,
                            "");
    NEST_REPLY_CHECK(reply);

    reply = frame_full_path(tmp_full_path,
                            user,
                            dest);
    NEST_REPLY_CHECK(reply);

    if (tmp_userhome == tmp_full_path) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Disallowing userhome_dir %s creation!\n",
                    tmp_userhome.c_str() );
        return NEST_BAD_PARAMETERS;
    }

    /*
        ignoring return value intentionally!
        similar to the one in ::open()
    */

    reply = file_catalog->get_logical_name( logical_dest,
                                            dest,
                                            user );

    bool is_sysf = file_catalog->is_system_file( logical_dest );
        
    if (!file_catalog->paccess( logical_dest,
                                                    user,
                                                    NEST_INSERT_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    if( !is_sysf ) {
        reply = file_catalog->add_entry( logical_dest,
                                         user,
                                         NEST_DIR);
        NEST_REPLY_CHECK(reply);
    }

    char    p_fname[NEST_MAXPATHLEN];

    reply = file_catalog->get_physical_name(p_fname, logical_dest);
    NEST_REPLY_CHECK(reply);

    if ( NestMkdir( p_fname, DIR_MODE ) < 0 ) {
        if (errno != EEXIST) {
            nest_warn( __FILE__, __FUNCTION__, "mkdir of '%s' failed: %s\n",
                       p_fname,
                       strerror( errno ) );

            if( !is_sysf ) {
                file_catalog->delete_entry( logical_dest );
                return ErrnoToNest( errno );
            }
        }
    }
    NestChmod( p_fname, DIR_MODE );
    NestChown( p_fname, uid, ActualGid );

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::deleteDirContents( const nestUser *user,
                                       const char *logical_dir)
{

    string constraint = "";
    constraint = constraint +
                    "other.DirName == \"" + logical_dir + "\"";

    ViewName    file_view = NEST_FILE_CATALOG_VIEW;
    ExprTree    *constraint_tree = parser.ParseExpression(constraint);
    query.Query(file_view, constraint_tree);
    query.ToFirst();

    string query_key;

    if (!query.Current(query_key)) {
        return NEST_SUCCESS;
    }

    ClassAd         *file_ad;

    do {
        file_ad = Resources->GetClassAd( query_key );

        if (!file_ad) {
            continue;
        }

        string  l_fname;
        if (!file_ad->EvaluateAttrString( "LogicalName", l_fname )) {
            continue;
        }
        string  type;
        if (!file_ad->EvaluateAttrString( "FileType", type )) {
            continue;
        }

        if (type == "NEST_DIR") {
            deleteDirContents( user, l_fname.c_str());
        }
        else {
            deleteFile(user, l_fname.c_str(), "");
        }

    } while (query.Next(query_key));

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::rmDir( const nestUser *user,
                           const char *cwd,
                           const char *dir,
                           bool recursive )
{

    char    dest[NEST_MAXPATHLEN];
    char    logical_dest[NEST_MAXPATHLEN],
            logical_cwd[NEST_MAXPATHLEN];

    NestReplyStatus reply = get_dest_dir( dest,
                                          cwd,
                                          dir,
										  user );

    if (!strcmp(dest, "~/")) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Trying to delete user's homedir!\n" );
        return NEST_BAD_PARAMETERS;
    }

    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_logical_name( logical_dest,
                                            dest,
                                            user );

    if (file_catalog->is_system_file( logical_dest )) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

    if (!file_catalog->paccess(logical_dest,
                                                   user,
                                                   NEST_DELETE_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_logical_name( logical_cwd,
                                            cwd,
                                            user );

    NEST_REPLY_CHECK(reply);

    /*
    if (!strcmp(logical_dest, logical_cwd)) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Trying to delete the current working dir!\n" );
        return NEST_BAD_PARAMETERS;
    }
    */

    if (recursive) {
        reply = deleteDirContents( user,
                                   logical_dest);

        NEST_REPLY_CHECK(reply);
    }

    char    p_fname[NEST_MAXPATHLEN];

    reply = file_catalog->get_physical_name(p_fname, logical_dest);
    NEST_REPLY_CHECK(reply);

    if ( NestRmdir( p_fname ) < 0 ) {
        reply = ErrnoToNest( errno );
    }

    if (file_catalog->delete_entry(logical_dest) != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Error while deleting NEST_DIR entry from FileCatalog!\n");
    }

    return reply;
}

char *
nestStorageManager::getRootDir( )
{
    return rootDir;
}

NestFileInfo *
nestStorageManager::directoryContents( const nestUser *user,
                                       const char *dir)
{

    char    dest_dir[NEST_MAXPATHLEN],
            logical_dir[NEST_MAXPATHLEN];
    NestReplyStatus reply;

    reply = get_dest_dir( dest_dir,
                          dest_dir,
                          dir,
                          user );
    if (reply != NEST_SUCCESS) {
        return NULL;
    }

    reply = file_catalog->get_logical_name(logical_dir,
                                           dest_dir,
                                           user);
    
    if (reply != NEST_SUCCESS) {
        return NULL;
    }
 
    if (!file_catalog->access(logical_dir,
                                         user,
                                         NEST_LIST_ACCESS)) {
        return NULL;
    }

    NestFileInfo *list = NULL;

    reply = file_catalog->dir_list( &list,
                                    logical_dir );

    if (reply != NEST_SUCCESS) {
        return NULL;
    }

    return list;
}

NestReplyStatus
nestStorageManager::add_user(const nestUser *cur_user,
							 const nestUser *new_user )
{

    if (!identity_manager->is_user_access_allowed( cur_user )) {
        return NEST_NONPRIVELEDGED_USER;
    }

    NestReplyStatus reply = identity_manager->add_user( new_user );
    NEST_REPLY_CHECK(reply);
	
    return add_home_dir( new_user );
}

NestReplyStatus
nestStorageManager::delete_user(const nestUser *cur_user,
                                const nestUser *user_to_delete )
{

    if (!identity_manager->is_user_access_allowed( cur_user )) {
        return NEST_NONPRIVELEDGED_USER;
    }

    NestReplyStatus reply = delete_home_dir( user_to_delete );

    NEST_REPLY_CHECK(reply);

    return identity_manager->delete_user( user_to_delete );
}

NestReplyStatus
nestStorageManager::increaseAllocation(nestClient *client)
{
    nestRequest         *req;
    nestLotSize         avl;
    nestLotSize         allocation;
    NestReplyStatus     reply;
    
    /* the allocated space is increased in multiples of 1 MB */
   
    if(!client) {
        return NEST_BAD_PARAMETERS;
    }

    if( !(req = client->getProxyRequest() ) ) {
        return NEST_BAD_PARAMETERS;
    }

    if (!NestEnableLots || lot_manager->kernel_quota_enabled()) {
        /*  
         *  allocate a very large chunk 1GB - no need to care of lots :-)
         *  Yet to proceed with the transfer the allocation parameter
         *  has to be set and can be set to arbitrarily large values to 
         *  minimize the number of times increaseAllocation is called.
         */  
        allocation = 1024 * 1024 * 1024;
        client->setAllocation( allocation );
        nest_debug(NEST_DEBUG_LOTS,
                   "Allocating %lld lot space for gftp-type transfers.\n",
                   allocation);  
        return NEST_SUCCESS;
    }

    string filename;
    char   l_fname[NEST_MAXPATHLEN];

    reply = expand_path(filename,
                        client->getUser(),
                        client->getCwd(),
                        req->getNestFile() );
    NEST_REPLY_CHECK(reply);
    /*
        ignore return value - as while creating new files
        this would return NEST_FILE_NOT_FOUND
    */
    reply = file_catalog->
		get_logical_name(l_fname, filename.c_str(),client->getUser() );
    
    avl = lot_manager->available_lot_space(client->getUser(), l_fname);

    if( avl <= 0 )
        return NEST_INSUFFICIENT_SPACE;
    if(avl < 1024*1024 )
        allocation = avl;
    else
        allocation = 1024*1024;

    nest_debug(NEST_DEBUG_LOTS,
               "Allocating %lld lot space for gftp-type transfers.\n",
               allocation);  
    client->setAllocation(allocation);
       
    reply = lot_manager->update_lot_usage(l_fname, allocation);
    return reply;
}

NestReplyStatus
nestStorageManager::resetAllocation(const char *file,
                                    const nestUser *user,
                                    const char *cwd,
                                    nestFileSize offset)
{
    NestReplyStatus reply;
    nestFileSize    size;
    
    reply = fileusage(&size, file, user, cwd);
    if(reply == NEST_FILE_NOT_FOUND){
        return NEST_SUCCESS;
    }

    string filename;
    char   l_fname[NEST_MAXPATHLEN];

    reply = expand_path(filename, user, cwd, file);
    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_logical_name(l_fname, filename.c_str(), user);
    NEST_REPLY_CHECK(reply);

    if(offset <  size) {
        reply = lot_manager->update_lot_usage(l_fname, offset - size);
    } 
    
    return reply;
}

NestReplyStatus
nestStorageManager::cleanupGftpTransfer(
                            nestClient *client)
{
    const char      *file, *cwd;
	const nestUser	*user;
    NestReplyStatus reply;
    char            l_fname[NEST_MAXPATHLEN];
    char            p_fname[NEST_MAXPATHLEN];
    string          filename;

    file = client->getRequest()->getNestFile();
    user = client->getUser();
    cwd  = client->getCwd();

    reply = expand_path( filename, user, cwd, file ); 
    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_logical_name(l_fname, filename.c_str(), user);
    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_physical_name(p_fname, l_fname);
    NEST_REPLY_CHECK(reply);

    reply = file_catalog->delete_entry( l_fname );
    NEST_REPLY_CHECK(reply);

    if (NestUnlink(p_fname) < 0) {
        return ErrnoToNest(errno);
    }
    return NEST_SUCCESS;
    
}
        

NestReplyStatus
nestStorageManager::add_group( const nestUser *cur_user,
							   const nestGroup *group )
{
    return identity_manager->add_group(cur_user, group);
}

NestReplyStatus
nestStorageManager::delete_group(const nestUser *cur_user,
								 const nestGroup *group )
{
    return identity_manager->delete_group(cur_user, group);
}

NestReplyStatus
nestStorageManager::add_user_to_group(const nestUser *cur_user,
									  const nestUser *user,
                                      const nestGroup *group )
{
    return identity_manager->add_user_in_group(cur_user, user, group );
}

NestReplyStatus
nestStorageManager::delete_user_from_group(const nestUser *cur_user,
										   const nestUser *user,
                                           const nestGroup *group )
{
    return identity_manager->delete_user_in_group(cur_user,
												  user, group );

}

NestReplyStatus
nestStorageManager::modify_lot_usage(const nestUser *user,
                                     const char *file,
                                     const char *cwd,
                                     const NestLotHandle  *from_lot,
                                     nestLotSize from_share,
                                     const NestLotHandle  *to_lot)
{
    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    char    l_fname[NEST_MAXPATHLEN];
    reply = file_catalog->get_logical_name( l_fname, filename.c_str(), user );

    if (reply == NEST_SUCCESS) {
        if (!file_catalog->access(l_fname, user, NEST_WRITE_ACCESS)) {
            return NEST_NONPRIVELEDGED_USER;
        }
    }
    else {
        if (file_catalog->paccess(l_fname, user, NEST_WRITE_ACCESS)) {
            return NEST_FILE_NOT_FOUND;
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    return lot_manager->modify_lot_usage(
		user, l_fname, from_lot, from_share, to_lot);
}

NestReplyStatus
nestStorageManager::attach_lot_to_file(const NestLotHandle *handle,
                                       const char *file,
                                       const nestUser *user,
                                       const char *cwd)
{
    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    NestReplyStatus reply;
    string          filename;
    char            l_fname[NEST_MAXPATHLEN];
    bool            new_file_created = false;

    reply = expand_path( filename, user, cwd, file );
    NEST_REPLY_CHECK(reply);
    
    reply = file_catalog->get_logical_name(l_fname, filename.c_str(), user);

    if( file_catalog->is_system_file(l_fname) ) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

    /*
     * Check if the user has the rights to create the file
     * if it is not already present.
     */
    if (!file_catalog->paccess(l_fname, user, NEST_INSERT_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }
    reply = file_catalog->add_entry(l_fname, user, NEST_FILE );
    if (reply != NEST_SUCCESS && 
        reply != NEST_FILE_EXISTS) {
        return reply;
    }
    new_file_created =  (reply == NEST_SUCCESS);

    /* Check if the user has the right to attach a lot to the file.
     * If the file entry was just created, we know that the right is
     * present.
     * Since NEST_MODIFYAC_ACCESS gives the access permissions to 
     * modify acls, this right is checked instead of adding a new
     * acl, say, NEST_ATTACH_LOT_ACCESS
     */
    if(!new_file_created) {
        if(!file_catalog->access(l_fname, user, NEST_MODIFYAC_ACCESS)) {
            return NEST_NONPRIVELEDGED_USER;
        }
    }
    
    /*
     * Check if the user has the rights to the lot 
     * so that it can be attached to the file.
     */
    NestLotInternal lot;
    reply = lot_manager->stat_lot( handle, user, &lot );
    if(reply != NEST_SUCCESS) {
        if(new_file_created) {
            file_catalog->delete_entry(l_fname);
        }
        return NEST_NONPRIVELEDGED_USER;
    }

    /*
     * Attach the lot to the file.
     */
    reply = file_catalog->attach_lot_to_file( l_fname, handle );
    if(reply != NEST_SUCCESS) {
        if(new_file_created) {
            file_catalog->delete_entry(l_fname);
        }
        return reply;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::detach_lot_from_file(const NestLotHandle *handle,
                                         const char *file,
                                         const nestUser *user,
                                         const char *cwd)
{
    if(!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    NestReplyStatus reply;
    string          filename;
    char            l_fname[NEST_MAXPATHLEN];

    reply = expand_path( filename, user, cwd, file );
    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_logical_name(l_fname, filename.c_str(), user);
    NEST_REPLY_CHECK(reply);
   
    if( file_catalog->is_system_file(l_fname) ) {
        return NEST_NOT_YET_IMPLEMENTED;
    }
    
    /*
     * Check if the user has the right to detach a lot from the file.
     */
    if(!file_catalog->access(l_fname, user, NEST_MODIFYAC_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }
    
    /*
     * Check if the user has the rights to the lot 
     * so that it can be attached to the file.
     */
    NestLotInternal lot;
    reply = lot_manager->stat_lot( handle, user, &lot);
    if(reply != NEST_SUCCESS && reply != NEST_LOT_NOT_FOUND)
        return reply;

    /*
     * Detach the lot from the file.
     */
    return file_catalog->detach_lot_from_file( l_fname, handle );
}

/*
    Returns the MAX(free_space) among all the devices available.
*/
NestReplyStatus
nestStorageManager::bytesAvailable(
	int			is_user,
	nestLotSize *free, 
	nestLotSize	*total )
{

    if (NestEnableLots) {
        *free = lot_manager->free_space(is_user);
        *total = lot_manager->total_space(is_user);
        return NEST_SUCCESS;
    } else {
        nest_fsstat fs_buf;

        if (NestStatFS(rootDir, &fs_buf) >= 0) {
            *free  = (nestLotSize) fs_buf.f_bfree  * fs_buf.f_bsize;
            *total = (nestLotSize) fs_buf.f_blocks * fs_buf.f_bsize; 
            return NEST_SUCCESS;
        } else {
            nest_debug( NEST_DEBUG_STORAGE,
                    "statfs failed in bytesAvailable()!\n");
            nest_perror( "statfs" );
            return ErrnoToNest( errno );
        }
    }
}


NestReplyStatus
nestStorageManager::filelength(   nestFileSize *filesize, 
                                  const char *file,
                                  const nestUser *user,
                                  const char *cwd )
{
    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    char    l_fname[NEST_MAXPATHLEN];
    char    p_fname[NEST_MAXPATHLEN];

    reply = file_catalog->get_logical_name(l_fname, filename.c_str(),  user);
    NEST_REPLY_CHECK(reply);

    if( file_catalog->is_system_file(l_fname) ) {

        if (!file_catalog->paccess(l_fname, user, NEST_LIST_ACCESS)) {
            return NEST_NONPRIVELEDGED_USER;
        }
        NEST_REPLY_CHECK(reply);

        reply = file_catalog->get_physical_name(p_fname, l_fname);
        NEST_REPLY_CHECK(reply);

        nest_stat   stat_buf;
        if (NestStat(p_fname, &stat_buf) < 0) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Next to impossible!\n" );
            *filesize = 0;
            return NEST_UNKNOWN_ERROR; 
        }
   
        *filesize = stat_buf.st_size; 
        return NEST_SUCCESS;
    }
    else {
        NestFileInfo info;
        reply = fileinfo( &info, file, user, cwd );
        if( reply == NEST_SUCCESS ) {
            *filesize = info.size;
        }
        return reply;
    }
}

NestReplyStatus
nestStorageManager::fileusage( nestFileSize *size,
                               const char *file,
                               const nestUser *user,
                               const char *cwd )
{
    NestFileInfo    info;
    NestReplyStatus status;
    *size = 0;

    status = fileinfo( &info,
                       file,
                       user,
                       cwd );
    NEST_REPLY_CHECK(status);

    ClassAd         *file_ad=info.classad;
    ExprTree *lot_list = file_ad->Lookup( "LotHandle" );
    if (!lot_list) {
        NestFree(info.classad);
        NestFree(&info);
        return NEST_CLASSAD_ERROR;
    }

    string str_lot_list = "";
    unparser.Unparse( str_lot_list, lot_list );
    *size = NestStrLotList_to_LotSize( str_lot_list.c_str() );

    return NEST_SUCCESS;
}

NestReplyStatus
nestStorageManager::fileinfo(   NestFileInfo *info, 
                                const char *file,
                                const nestUser *user,
                                const char *cwd,
                                int link_count ) 
{


    string filename;
    NestReplyStatus reply = expand_path( filename,
                                         user,
                                         cwd,
                                         file );
    NEST_REPLY_CHECK(reply);

    nest_debug( NEST_DEBUG_STORAGE,
                "file info: file='%s' user='%s' cwd='%s filename='%s'\n",
                file, user->getName(), cwd, filename.c_str() );

    if (!info) {
        return NEST_BAD_PARAMETERS;
    }

    char    l_fname[NEST_MAXPATHLEN];
    char    p_fname[NEST_MAXPATHLEN];


    reply = file_catalog->get_logical_name(l_fname, filename.c_str(), user);
    NEST_REPLY_CHECK(reply);
	strncpy( info->name, l_fname, NEST_MAXPATHLEN );
	info->name[NEST_MAXPATHLEN - 1] = '\0';

    if (!file_catalog->paccess(l_fname, user, NEST_LIST_ACCESS)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    NEST_REPLY_CHECK(reply);

    reply = file_catalog->get_physical_name(p_fname, l_fname);
    NEST_REPLY_CHECK(reply);

    nest_obj_type_t     filetype;

    reply = file_catalog->get_type( &filetype, l_fname );
    NEST_REPLY_CHECK(reply);
    info->type = filetype;

    if ((filetype == NEST_LINK) &&
        !file_catalog->is_system_file(l_fname)) 
    {

        if (link_count++ > NEST_LINK_MAXLOOP) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "Too many levels of symbolic links!\n" );
            return NEST_TOO_MANY_LINKS;
        }

            // recursion
        return fileinfo( info, p_fname, user, cwd, link_count );
    }

        // set classad and uniquifier
    if(!file_catalog->is_system_file(l_fname)) {
        reply = file_catalog->get_file_classad(l_fname, &(info->classad) );
        NEST_REPLY_CHECK(reply);
    }
    reply = file_catalog->get_file_uniquifier( p_fname, &(info->uniquifier) );
    NEST_REPLY_CHECK(reply);

    info->size = NestFilelength( p_fname );

        // if we make it here, we succeeded in the lookup, cache the result
    cache_uniquifier( filename, &(info->uniquifier), info->type ); 

		// Stuff we don't know
	info->LotList[0] = '\0';

    return NEST_SUCCESS;
}

bool
nestStorageManager::validate_user( nestUser *user )
{
	return identity_manager->is_user_present(user);
}

NestReplyStatus
nestStorageManager::formatDisk( const nestUser *user,
                                const nestGroup *group )
{

    return NEST_NOT_YET_IMPLEMENTED;

#if 0
    char    logical_home[NEST_MAXPATHLEN];
    NestReplyStatus reply = file_catalog->get_logical_name(logical_home,
                                                           "/",
                                                           user);

    NEST_REPLY_CHECK(reply);

    //  TODO: fix this

    return rmDir ( user,
                   logical_home,
                   "..",
                   true );
#endif
}

int
nestStorageManager::chRootDir( )
{
    int ret = chdir( rootDir );
    if ( ret < 0 ) {
        nest_warn( __FILE__, __FUNCTION__,
				   "Couldn't chdir to rootdir %s: %s\n", rootDir,
				   strerror( errno ) );
    }
    return ret;
}

int
nestStorageManager::chMetaDir( )
{
    if ( chRootDir() < 0 || chdir( MetaDir ) < 0 ) {
        nest_warn( __FILE__, __FUNCTION__,
				   "chdir to MetaDir %s: %s\n", MetaDir, strerror(errno) );
        return -1;
    }
    return 0;
}

NestReplyStatus
nestStorageManager::get_home_dir( string &result,
                                  const nestUser *user )
{

    /*
        File Catalog manager - depends heavily on this
        format for the homedir!

        If you change this - make sure nestFileCatalog
        is aware of this and is happy
    */

    result = "";
	const char *local = user->getLocalName( );
	if ( local ) {
		result = result + rootDir + "/" + local + ':';
	} else {
		result = result + rootDir + "/" + user->getIdentityClean() + ':';
	}

    return NEST_SUCCESS;
}


NestReplyStatus
nestStorageManager::add_home_dir(const nestUser *user )
{

    char			    logical_home[NEST_MAXPATHLEN];
	NestReplyStatus		reply;

	nest_debug( NEST_DEBUG_ALWAYS, "Adding home for '%s'\n",
				user->getIdentity() );

	// Ignoring return value intentionally
    (void) file_catalog->get_logical_name( logical_home, "/", user );

    string  physical_home;
    reply = get_home_dir( physical_home, user );
	if ( NEST_SUCCESS != reply ) {
		nest_debug( NEST_DEBUG_ALWAYS,
					"Failed to get logical home dir for user '%s'\n",
					user->getIdentity() );
		return reply;
	}

    if ( NestMkdir( physical_home.c_str(), DIR_MODE ) < 0 ) {
        if (errno != EEXIST) {
            nest_warn( __FILE__, __FUNCTION__, "mkdir of %s failed: %s\n",
                       physical_home.c_str(),
                       strerror( errno ) );
            return ErrnoToNest( errno );
        }
    }

    NestChmod( physical_home.c_str(), DIR_MODE );
    NestChown( physical_home.c_str(), user->getUid( ), ActualGid );

    reply = file_catalog->add_entry( logical_home,
                                     physical_home.c_str(),
                                     "/",
                                     user,
                                     NEST_DIR);

    if (reply == NEST_FILE_EXISTS) {
        reply = NEST_SUCCESS;
    }

    return reply;
}

NestReplyStatus
nestStorageManager::delete_home_dir(const nestUser *user )
{

    char    logical_home[NEST_MAXPATHLEN];
    /*
        ignoring return value intentionally
    */

    NestReplyStatus reply = file_catalog->get_logical_name(logical_home,
                                                           "/",
                                                           user);

    reply = file_catalog->delete_entry( logical_home );

    if (reply != NEST_SUCCESS &&
        reply != NEST_FILE_NOT_FOUND) {
        return reply;
    }

    string  physical_home;
    get_home_dir( physical_home, user );

    if ( NestRmdir( physical_home.c_str() ) < 0 ) {
        if (errno != ENOENT) {
            nest_warn( __FILE__, __FUNCTION__, "rmdir of %s failed: %s\n",
                       physical_home.c_str(),
                       strerror( errno ) );
            return ErrnoToNest( errno );
        }
    }

    return NEST_SUCCESS;
}
    
NestReplyStatus  
nestStorageManager::rememberLength( const nestUser *user,
									const char *cwd, 
									const char *file,
									nestFileSize *length ) 
{
	// this is for proxys to predict the length of previously fetched files.
	// not necessary on a "real" dedicated nest
    return NEST_NOT_YET_IMPLEMENTED;
}

char *
nestStorageManager::type( )
{
    return "NeST";
}

bool
nestStorageManager::is_lots_enabled( )
{
    return NestEnableLots;
}
